#include "bgpcommon.h"
#include <stdarg.h>

namespace BGPHist {


IPAddress::IPAddress():aType(atNone){}
IPAddress::IPAddress(unsigned a):aType(atIPv4){memcpy(&data,&a,4);}
IPAddress::IPAddress(const IPAddress & x){
    operator=(x);
}
IPAddress & IPAddress::operator=(const IPAddress & x){
    aType=x.aType;
    memcpy(&data,&x.data,sizeof(data));
    return *this;
}
IPAddress & IPAddress::assignIPv4(const void *x){
    aType=atIPv4;
    if(x) memcpy(&data,x,4); else memset(&data,0,4);
    return *this;
}
IPAddress & IPAddress::assignIPv4RD(const void *x){
    aType=atIPv4RD;
    if(x) memcpy(&data,x,12); else memset(&data,0,12);
    return *this;
}
IPAddress & IPAddress::assignIPv6(const void *x){
    aType=atIPv6;
    if(x) memcpy(&data,x,16); else memset(&data,0,16);
    return *this;
}
IPAddress & IPAddress::assignIPv6RD(const void *x){
    aType=atIPv6RD;
    if(x) memcpy(&data,x,24); else memset(&data,0,24);
    return *this;
}
IPAddress & IPAddress::assignL2(const void *x){
    aType=atL2;
    if(x) memcpy(&data,x,14); else memset(&data,0,14);
    return *this;
}
void IPAddress::clear(){
    aType=atNone;
}
size_t IPAddress::datalen() const{
   switch(aType){
    case atIPv4:return 4;
    case atIPv4RD:return 12;
    case atIPv6:return 16;
    case atIPv6RD:return 24;
    case atL2:return 14;
    }
   return 0;
}
IPAddress & IPAddress::assign(const void *x,size_t s){
    if (s==4) return assignIPv4(x);
    if (s==12) return assignIPv4RD(x);
    if (s==16) return assignIPv6(x);
    if (s==14) return assignL2(x);
    throw BGPException(Sprintf("Unknown address length %u",s));
}
IPAddress & IPAddress::setRD(const IPAddress & x){
    if (aType==atIPv4){
        memcpy(&data[8],&data[0],4);
        aType=atIPv4RD;
    }
    if (aType==atIPv6){
        memcpy(&data[8],&data[0],16);
        aType=atIPv6RD;
    }
    if ((aType!=atIPv4RD) && (aType!=atIPv6RD)) return *this;
    if ((x.aType!=atIPv4RD) && (x.aType!=atIPv6RD)) return *this;
    memcpy(&data[0],&x.data[0],8);
    return *this;
}
unsigned IPAddress::getIPv4Addr() const{
    switch(aType){
        case atIPv4:{
            return be32(&data[0]);
        }
        case atIPv4RD:{
            return be32(&data[8]);
        }
    }
    return 0;
}
uint64_t IPAddress::getIPv6AddrH() const{
    switch(aType){
        case atIPv6:{
            return be64(&data[0]);
        }
        case atIPv6RD:{
            return be64(&data[8]);
        }
    }
    return 0;
}
uint64_t IPAddress::getIPv6AddrL() const{
    switch(aType){
        case atIPv6:{
            return be64(&data[8]);
        }
        case atIPv6RD:{
            return be64(&data[16]);
        }
    }
    return 0;
}
unsigned IPAddress::getRDH() const{
    if ((aType!=atIPv4RD) && (aType!=atIPv6RD) && (aType!=atL2)) return 0;
    return be32(&data[0]);
}
unsigned IPAddress::getRDL() const{
    if ((aType!=atIPv4RD) && (aType!=atIPv6RD) && (aType!=atL2)) return 0;
    return be32(&data[4]);
}
unsigned IPAddress::getL2Site() const{
    if (aType!=atL2) return 0;
    return be16(&data[8]);
}
unsigned IPAddress::getL2Offset() const{
    if (aType!=atL2) return 0;
    return be16(&data[10]);
}
unsigned IPAddress::getL2Range() const{
    if (aType!=atL2) return 0;
    return be16(&data[12]);
}
IPAddrType IPAddress::getType() const {
    return aType;
}

std::string ipv6str(const IPV6 *p6){
    std::string ret;
    unsigned i,j,k,l;
    if (p6->u64[0]==0 && p6->u64[1]==0) return "::";
    if (p6->u64[0]==0 && p6->u16[5]==0xffff){
      //ipv4 aliased
      return Sprintf("::ffff:%u.%u.%u.%u",(unsigned)p6->c[12],(unsigned)p6->c[13],(unsigned)p6->c[14],(unsigned)p6->c[15]);
    }
    j=k=10;l=0;
    for(i=0;i<8;i++)if(p6->u16[i]==0){
      if (j>=8) j=i;
      if((i-j)>l) {k=j;l=i-j;}
    } else j=10;
    if ((l>1) && (k<7)){
      ret="";
      for(i=0;i<k;i++) ret+=Sprintf("%x:",be16(&p6->u16[i]));
      if ((k+l+1)>=8) ret+=":"; else {
        for(i=k+l+1;i<8;i++) ret+=Sprintf(":%x",be16(&p6->u16[i]));
      }
      //ret+=Sprintf("<%x:%x:%x:%x:%x:%x:%x:%x>",be16(&data[0]),be16(&data[2]),be16(&data[4]),be16(&data[6]),be16(&data[8]),be16(&data[10]),be16(&data[12]),be16(&data[14]));
    } else 
     ret=Sprintf("%x:%x:%x:%x:%x:%x:%x:%x",be16(&p6->u16[0]),be16(&p6->u16[1]),be16(&p6->u16[2]),be16(&p6->u16[3]),be16(&p6->u16[4]),be16(&p6->u16[5]),be16(&p6->u16[6]),be16(&p6->u16[7]));
    return ret;
}

std::string IPAddress::str_rd() const{
    if (aType!=atIPv6RD && aType!=atIPv4RD && (aType!=atL2)) return "";
    if (data[0]==0 && data[1]==1){
        return Sprintf("<%u.%u.%u.%u:%u>",(unsigned)data[2],(unsigned)data[3],(unsigned)data[4],(unsigned)data[5],(unsigned)be16(&data[6]));
    }
    return Sprintf("<%u:%u>",be32(&data[0]),be32(&data[4]));
}

std::string IPAddress::str() const{
    std::string ret;
    switch(aType){
        case atNone:ret="<none>";break;
        case atIPv4:ret=Sprintf("%u.%u.%u.%u",(unsigned)data[0],(unsigned)data[1],(unsigned)data[2],(unsigned)data[3]);break;
        case atIPv4RD:ret=str_rd()+Sprintf("%u.%u.%u.%u",(unsigned)data[8],(unsigned)data[9],(unsigned)data[10],(unsigned)data[11]);break;
        case atIPv6:{
            IPV6 *p6=(IPV6 *)&data;
            ret=ipv6str(p6);
            break;
        }
        case atIPv6RD:{
            IPV6 *p6=(IPV6 *)&data[8];
            ret=str_rd()+ipv6str(p6);
            break;
        }
        case atL2:{
            ret=str_rd()+Sprintf("%u:%u-%u",be16(&data[8]),be16(&data[10]),be16(&data[12]));
            break;
        }
    }
    return ret;
}
bool IPAddress::operator==(const IPAddress &x) const{
    if (aType!=x.aType) return false;
    return (memcmp(data,x.data,datalen())==0);
}
bool IPAddress::operator!=(const IPAddress &x) const{
    if (aType==x.aType) return false;
    return (memcmp(data,x.data,datalen())!=0);
}


IPNet::IPNet():preflen(0){}
IPNet::IPNet(const IPNet &x):IPAddress(x),preflen(x.preflen){}
IPNet::IPNet(const IPAddress &x,uint8_t l):IPAddress(x),preflen(l){}
IPNet & IPNet::operator=(const IPAddress & x){
    IPAddress::operator=(x);
    switch(aType){
        case atIPv4:preflen=32;break;
        case atIPv4RD:preflen=32;break;
        case atIPv6:preflen=128;break;
        case atIPv6RD:preflen=128;break;
        case atL2:preflen=32;break;
    }
    return *this;
}
unsigned IPNet::getPrefixLen() const{
    return preflen;
}
IPNet & IPNet::operator=(const IPNet & x){
    IPAddress::operator=(x);
    preflen=x.preflen;
    return *this;
}
IPNet & IPNet::assignIPv4(const void *v,uint8_t l){
    IPAddress::assignIPv4(v);
    preflen=l;
    return *this;
}
IPNet & IPNet::assignIPv6(const void *v,uint8_t l){
    IPAddress::assignIPv6(v);
    preflen=l;
    return *this;
}

IPNet & IPNet::assignNLRI4(uint8_t bits,const void *v){
    uint32_t u32=0;
    if (!bits) return assignIPv4(&u32,0);
    if (bits>32) throw BGPException(Sprintf("Incorrect NLRI length %u",bits));
    memcpy(&u32,v,(bits+7)/8);
    u32=be32(&u32);
    if (bits<32) u32&=((-1) ^ (1l << (32-bits))-1);
    u32=be32(&u32);
    return assignIPv4(&u32,bits);
}
IPNet & IPNet::assignNLRI6(uint8_t bits,const void *v){
    IPV6 addr;
    uint64_t u64;
    if (!bits) return assignIPv6(0,0);
    if (bits>128) throw BGPException(Sprintf("Incorrect NLRI length %u",bits));
    memset(&addr,0,sizeof(addr));
    memcpy(&addr,v,(bits+7)/8);
    if (bits<64){
        u64=be64(&addr.u64[0]);
        u64&=((-1) ^ (1ll << (64-bits))-1);
        addr.u64[0]=be64(&u64);
    } else
    if(bits==64){
        addr.u64[1]=0;
    } else {
        u64=addr.u64[1];
        u64&=((-1) ^ (1ll << (128-bits))-1);
        addr.u64[1]=be64(&u64);
    }
    return assignIPv6(&addr,bits);
}

std::string IPNet::str() const{
    std::string ret=IPAddress::str();
    ret+=Sprintf("/%u",(unsigned)preflen);
    return ret;    
}
bool IPNet::operator<(const IPNet &x) const{
    if (aType!=x.aType) {
        if(datalen()<x.datalen()) return true;
        return false;
    }
    if (preflen<x.preflen) return true;
    if (preflen>x.preflen) return false;
    if (memcmp(data,x.data,datalen())<0) return true;
    return false;
}

IPTagNet::IPTagNet(){}
IPTagNet::IPTagNet(const IPTagNet &x):labels(x.labels),IPNet(x){}
IPTagNet::IPTagNet(const IPNet &x):IPNet(x){}
IPTagNet::IPTagNet(const IPAddress &x,uint8_t l):IPNet(x,l){}
IPTagNet & IPTagNet::operator=(const IPAddress & x){
    IPNet::operator=(x);
    return *this;
}
IPTagNet & IPTagNet::operator=(const IPNet & x){
    IPNet::operator=(x);
    return *this;
}
IPTagNet & IPTagNet::operator=(const IPTagNet & x){
    IPNet::operator=(x);
    labels=x.labels;
    return *this;
}
/*
IPTagNet & IPTagNet::assignNLRI(const MPLSLabels &l,uint8_t bits,const void *v){
    IPNet::assignNLRI(bits,v);
    labels=l;
    return *this;
}
IPTagNet & IPTagNet::assignNLRI6(const MPLSLabels &l,uint8_t bits,const void *v){
    IPNet::assignNLRI6(bits,v);
    labels=l;
    return *this;
}
*/
uint8_t IPTagNet::extractLabels(uint8_t bits,const void *vbf){
    unsigned char *buf;
    buf=(unsigned char *)vbf;
    labels.clear();
    unsigned m,nl,lbl;
    m=0;
    nl=bits;
    for(;;){
     lbl=(buf[m] << 12) | (buf[m+1] << 4) | (buf[m+2] >> 4);
     m+=3;nl-=24;
     labels.push_back(lbl);
     if (lbl==slvWithdraw || lbl==slvExplicitNull || lbl==slvExplicitNull6 || lbl==slvImplicitNull) break;//last label
     if ((buf[m-1] & 1)!=0) break;//end of label stack
    }
    return m;
}
IPTagNet & IPTagNet::assignNLRI4(uint8_t bits,const void *vbf){
    if (!vbf || !bits) {labels.clear();IPNet::assignNLRI4(0,0);return *this;}
    unsigned char *buf;
    buf=(unsigned char *)vbf;
    uint8_t lbs=extractLabels(bits,vbf);
    IPNet::assignNLRI4(bits-lbs*8,&buf[lbs]);
    return *this;
}
IPTagNet & IPTagNet::assignNLRIRD4(uint8_t bits,const void *vbf){
    if (!vbf || !bits) {labels.clear();IPNet::assignNLRI4(0,0);return *this;}
    unsigned char *buf;
    buf=(unsigned char *)vbf;
    uint8_t lbs=extractLabels(bits,vbf);
    IPAddress aRd;
    aRd.assignIPv4RD(&buf[lbs]);
    IPNet::assignNLRI4(bits-lbs*8-64,&buf[lbs+8]);
    setRD(aRd);
    return *this;
}

IPTagNet & IPTagNet::assignNLRI6(uint8_t bits,const void *vbf){
    if (!vbf || !bits) {labels.clear();IPNet::assignNLRI6(0,0);return *this;}
    unsigned char *buf;
    buf=(unsigned char *)vbf;
    uint8_t lbs=extractLabels(bits,vbf);
    IPNet::assignNLRI6(bits-lbs*8,&buf[lbs]);
    return *this;
}

IPTagNet & IPTagNet::assignNLRIRD6(uint8_t bits,const void *vbf){
    if (!vbf || !bits) {labels.clear();IPNet::assignNLRI6(0,0);return *this;}
    unsigned char *buf;
    buf=(unsigned char *)vbf;
    uint8_t lbs=extractLabels(bits,vbf);
    IPAddress aRd;
    aRd.assignIPv6RD(&buf[lbs]);
    IPNet::assignNLRI6(bits-lbs*8-64,&buf[lbs+8]);
    setRD(aRd);
    return *this;
}
IPTagNet & IPTagNet::assignNLRIL2(uint8_t bits,const void *vbf){
    if (!vbf || !bits) {labels.clear();IPNet::assignNLRI4(0,0);return *this;}
    preflen=32;
    unsigned char *buf;
    buf=(unsigned char *)vbf;
    aType=atL2;
    memcpy(data,buf,14);//rdH,rdL,site,offset,range
    uint8_t lbs=extractLabels(24,buf+14);
    return *this;
}
std::string IPTagNet::str() const{
    if (labels.empty()) return IPNet::str();
    std::string ret="(";
    MPLSLabels::const_iterator i;
    for(i=labels.begin();i!=labels.end();i++) ret+=Sprintf("%s%u",(i==labels.begin() ? "":","),(unsigned)*i);
    ret+=")"+IPNet::str();
    return ret;
}

BGPException::~BGPException() throw(){}

BGPException::BGPException(const BGPException &_x) throw(){
    try{
        msg=_x.msg;
        code=_x.code;
    } catch(...){}
}

BGPException::BGPException(const char *_msg):msg(_msg){}
BGPException::BGPException(const std::string & _msg):msg(_msg){}
BGPException::BGPException(unsigned _code):code(_code){
    char cBuf[80];
    sprintf(cBuf,"code %u",code);
    msg=cBuf;
}
BGPException::BGPException(unsigned _code,const char *_msg):code(_code),msg(_msg){
}

const char* BGPException::what() const throw(){
    try{
        return msg.c_str();
    } catch(...){}
    return "BGPException";
}

BGPTimeout::BGPTimeout(){}
const char* BGPTimeout::what() const throw(){
    return "BGPTimeout";
}

std::string Sprintf(const char *fmt,...){
        std::string ret;
    size_t sz=strlen(fmt)*4;
    va_list ap;
    va_start(ap, fmt);
    if (sz<1024) sz=1024;
    ret.resize(sz);
    vsnprintf((char *)ret.c_str(),sz,fmt,ap);
    ret.resize(strlen(ret.c_str()));
    va_end(ap);
    return ret;
}

unsigned short be16(const void *p){
    unsigned short r;
    const unsigned char *c=(const unsigned char *)p;
    r=c[1] | (((unsigned short)c[0]) << 8);
    return r;
}
unsigned be24(const void *p){
    unsigned r;
    const unsigned char *c=(const unsigned char *)p;
    r=c[2] | (((unsigned)c[1]) << 8) | (((unsigned) c[0]) << 16);
    return r;
}
unsigned be32(const void *p){
    unsigned r;
    const unsigned char *c=(const unsigned char *)p;
    r=c[3] | (((unsigned)c[2]) << 8) | (((unsigned)c[1]) << 16) | (((unsigned)c[0]) << 24);
    return r;
}
uint64_t be64(const void *p){
    uint64_t r;
    const unsigned char *s=(const unsigned char *)p;
    unsigned char *d=(unsigned char *)&r;
    d[0]=s[7];
    d[1]=s[6];
    d[2]=s[5];
    d[3]=s[4];
    d[4]=s[3];
    d[5]=s[2];
    d[6]=s[1];
    d[7]=s[0];
    return r;
}



}
