#ifndef __bgpcommon_h_
#define __bgpcommon_h_
#include "config.h"
extern "C"{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
}
#include <string>
#include <vector>
#include <exception>

namespace BGPHist {

typedef unsigned BGPASN;

typedef std::vector<BGPASN> BGPASNs;

typedef enum {atNone=0,atIPv4,atIPv6,atIPv4RD,atIPv6RD,atL2} IPAddrType;

typedef enum {
    slvExplicitNull=0,
    slvRouterAlert=1,
    slvExplicitNull6=2,
    slvImplicitNull=3,
    slvWithdraw=524288
} SpecialLabelValue;

#pragma pack(push,1)
typedef union tagIPV6 {
    uint16_t u16[8];
    uint32_t u32[4];
    uint64_t u64[2];
    unsigned char c[16];
} IPV6;
#pragma pack(pop)
typedef uint32_t MPLSLabel;

typedef std::vector<MPLSLabel> MPLSLabels;

class IPAddress{
protected:
    IPAddrType aType;
    uint8_t data[24];
    size_t datalen() const;
    std::string str_rd() const;
public:
    IPAddress();
    IPAddress(unsigned );
    IPAddress(const IPAddress & );
    void clear();
    IPAddress & operator=(const IPAddress & );
    IPAddress & assignIPv4(const void *);
    IPAddress & assignIPv4RD(const void *);
    IPAddress & assignIPv6RD(const void *);
    IPAddress & assignIPv6(const void *);
    IPAddress & assignL2(const void *);
    IPAddress & setRD(const IPAddress & x);
    IPAddress & assign(const void *,size_t );
    unsigned getIPv4Addr() const;
    uint64_t getIPv6AddrH() const;
    uint64_t getIPv6AddrL() const;
    unsigned getRDH() const;
    unsigned getRDL() const;
    unsigned getL2Site() const;
    unsigned getL2Offset() const;
    unsigned getL2Range() const;
    std::string str() const;
    IPAddrType getType() const;
    bool operator==(const IPAddress &) const;
    bool operator!=(const IPAddress &) const;
};
typedef std::vector<IPAddress> IPAddresses;
class IPNet: public IPAddress{
protected:
		uint8_t preflen;
public:
		IPNet();
		IPNet(const IPNet &);
		IPNet(const IPAddress &,uint8_t);
    IPNet & operator=(const IPAddress & );
    IPNet & operator=(const IPNet & );
    IPNet & assignIPv4(const void *,uint8_t );
    IPNet & assignIPv6(const void *,uint8_t );
    IPNet & assignNLRI4(uint8_t bits,const void *);
    IPNet & assignNLRI6(uint8_t bits,const void *);
    unsigned getPrefixLen() const;
    std::string str() const;
    bool operator<(const IPNet &) const;
};
typedef std::vector<IPNet> IPNets;
class IPTagNet: public IPNet {
    uint8_t extractLabels(uint8_t bits,const void *);
public:
		MPLSLabels labels;
		IPTagNet();
		IPTagNet(const IPTagNet &);
		IPTagNet(const IPNet &);
		IPTagNet(const IPAddress &,uint8_t);
    IPTagNet & operator=(const IPAddress & );
    IPTagNet & operator=(const IPNet & );
    IPTagNet & operator=(const IPTagNet & );
    //IPTagNet & assignNLRI(const MPLSLabels &,uint8_t bits,const void *);
    //IPTagNet & assignNLRI6(const MPLSLabels &,uint8_t bits,const void *);
    IPTagNet & assignNLRI4(uint8_t bits,const void *);
    IPTagNet & assignNLRIRD4(uint8_t bits,const void *);
    IPTagNet & assignNLRI6(uint8_t bits,const void *);
    IPTagNet & assignNLRIRD6(uint8_t bits,const void *);
    IPTagNet & assignNLRIL2(uint8_t bits,const void *);
    std::string str() const;
};
typedef std::vector<IPTagNet> IPTagNets;

class BGPException: public std::exception {
public:
    std::string msg;
    unsigned code;
    BGPException(const char *_msg);
    BGPException(const std::string & _msg);
    BGPException(unsigned _code);
    BGPException(unsigned _code,const char *_msg);
    BGPException(const BGPException &_x) throw();
    virtual const char* what() const throw();
    virtual ~BGPException() throw();
};

class BGPTimeout: public std::exception {
public:
    BGPTimeout();
    virtual const char* what() const throw();
};

std::string Sprintf(const char *fmt,...);
unsigned short be16(const void *p);
unsigned be24(const void *p);
unsigned be32(const void *p);
uint64_t be64(const void *p);

};

#endif 

