#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include "config.h"
#include "bgppeer.h"
#include "bgplogcon.h"
#include "bgplogproxy.h"
#ifdef USE_MYSQL
#include "bgplogmysql.h"
#endif
#ifdef USE_OCI
#include "bgplogoci.h"
#endif
using namespace BGPHist;
BGPLogProxy * proxy;

void showproxystat(int signum){
    std::string s=proxy->get_stats();
    fprintf(stderr,"Stats: %s\n",s.c_str());
    signal(signum,showproxystat);
}

void usage(const char *s){
    fprintf(stderr,"%s: -p <bgp peer ip> -a <ASN> -d <database type>[:<database connection params>] -w <async wait msec> -f play capture file instead of bgp router connect\n",s);
    fprintf(stderr," Databases:\n  text\n");
#ifdef USE_OCI
    fprintf(stderr,"  oci:<user>/<password>@<tnsname>\n");
#endif    
#ifdef USE_MYSQL
    fprintf(stderr,"  mysql:<mysql host>:<mysqldb>:<user>:<pwd>\n");
#endif    
}
int main (int argc, char **argv)
{
    std::string sPeer,sDbT,sDb,s,sPF;
    std::vector<std::string> sMys;
    unsigned uASN,uAsync;
    int i,iPF;
    uASN=0;uAsync=0;iPF=-1;
    while((i=getopt(argc,argv,"p:a:d:hw:f:"))>0){
        switch(i){
            case 'h':
                usage(argv[0]);
                return 0;
            case 'p':
                sPeer=optarg;
                break;
            case 'a':
                sscanf(optarg,"%u",&uASN);
                break;
            case 'w':
                sscanf(optarg,"%u",&uAsync);
                break;
            case 'd':
                sDb=optarg;
                break;
            case 'f':
                sPF=optarg;
                break;
        }
    }
    if (!uASN || sDb.empty()){
        usage(argv[0]);
        return 1;
    }
    BGPLog *log;
    s="";
    sDbT="";
    for(i=0;i<sDb.size();i++){
        if (sDb[i]==':') {i++;break;}
        sDbT+=sDb[i];
    }
#ifdef USE_OCI
    if(sDbT=="oci"){
      for(;i<sDb.size();i++)if((sDb[i]=='/') || (sDb[i]=='@')){
        sMys.push_back(s);
        s="";
      } else s+=sDb[i];
      sMys.push_back(s);
      if (sMys.size()!=3) {
        usage(argv[0]);
        return 1;
      }
      log=new BGPLogOCI(sMys[2].c_str(),sMys[0].c_str(),sMys[1].c_str(),sPeer.c_str(),uASN);
    } else
#endif        
#ifdef USE_MYSQL
    if(sDbT=="mysql"){
      for(;i<sDb.size();i++)if(sDb[i]==':'){
        sMys.push_back(s);
        s="";
      } else s+=sDb[i];
      sMys.push_back(s);
      if (sMys.size()!=4) {
        usage(argv[0]);
        return 1;
      }
      log=new BGPLogMy(sMys[0].c_str(),sMys[1].c_str(),sMys[2].c_str(),sMys[3].c_str(),sPeer.c_str(),uASN);
    } else
#endif
    if (sDbT=="text"){
        log=new BGPLogCon();
    } else {
        fprintf(stderr,"Unknown or unsupported database type: '%s'\n",sDb.c_str());
    }
    signal(SIGPIPE,SIG_IGN);
    BGPLogProxy cache(log,uAsync);
    proxy=&cache;
    signal(SIGUSR1,showproxystat);
    //BGPPeer::connectTo(&cache,sPeer.c_str(),uASN);
    if (sPF.size()>0){
        iPF=open(sPF.c_str(),O_RDONLY);
        if (iPF<0){
            i=errno;
            fprintf(stderr,"Unable to open file '%s' error %u\n",sPF.c_str(),i);
            return 2;
        }
    }
    BGPPeer * peer=new BGPPeer(&cache);
    try {
        peer->asnPeer=uASN;
        peer->asnMy=uASN;
        peer->sPeer=sPeer.c_str();
        if (iPF>=0){
            peer->playfile(iPF);
            peer->lifecycle();
            close(iPF);
        } else {
         if (peer->connect())
            peer->lifecycle();
        }
    }
    catch (const BGPException & e){
        log->exception(0,&e);
        fprintf(stderr,"caught BGPException '%s'",e.what());
    }
    catch (const SocketException & e) {
     log->txt(0,e.what());
     fprintf(stderr,"Connecting to %s caught sockerr '%s'",sPeer.c_str(),e.what());
    }
    catch(const std::exception & e){
     fprintf(stderr,"caught exception '%s'",e.what());
    }
    delete peer;
    delete log;
    return 0;
}
