#include "bgplog.h"

namespace BGPHist{

std::string BGPExtCommunity::str() const{
    IPAddress ipa;
    uint32_t addr;
    addr=((uint32_t)a << 16) | (b >> 16);
    ipa.assignIPv4(&addr);
    switch(excType){
        case 2:{
            return Sprintf("{rt:%u:%u}",(unsigned)a,(unsigned)b);
        }
        case 258:{
            return std::string("{rt:")+ipa.str()+Sprintf(":%u}",(unsigned)(b & 0xffff));
        }
        case 266:{
            return std::string("{rt-import:")+ipa.str()+Sprintf(":%u}",(unsigned)(b & 0xffff));
        }
        case 32768:{
            return std::string("{route-type-vendor:")+ipa.str()+Sprintf(":%u:%u}",(unsigned)((b && 0xff00) >> 8),(unsigned)(b & 0xff));
        }
        case 32769:{
            return std::string("{route-id-vendor:")+ipa.str()+Sprintf(":%u:%u}",(unsigned)((b && 0xff00) >> 8),(unsigned)(b & 0xff));
        }
        case 774:{
            return std::string("{rte-type:")+ipa.str()+Sprintf(":%u:%u}",(unsigned)((b && 0xff00) >> 8),(unsigned)(b & 0xff));
        }
        case 3:{
            return Sprintf("{origin:%u:%u}",(unsigned)a,(unsigned)b);
        }
        case 5:{
            return Sprintf("{domain-id:%u:%u}",(unsigned)a,(unsigned)b);
        }
        case 9:{
            return Sprintf("{src-as:%u:%u}",(unsigned)a,(unsigned)b);
        }
    }
    return Sprintf("{%u:%u:%u}",(unsigned)excType,(unsigned)a,(unsigned)b);
}
bool BGPExtCommunity::operator!=(const BGPExtCommunity &x) const{
    return (excType!=x.excType) || (a!=x.a) || (b!=x.b);
}
bool BGPExtCommunity::operator==(const BGPExtCommunity &x) const{
    return (excType==x.excType) && (a==x.a) && (b==x.b);
}



BGPLog::~BGPLog(){}

bgplogstream & operator << (bgplogstream & sm,const IPAddress &a){return sm << a.str();}
bgplogstream & operator << (bgplogstream & sm,const IPNet &a){return sm << a.str();}
bgplogstream & operator << (bgplogstream & sm,const IPTagNet &a){return sm << a.str();}
bgplogstream & operator << (bgplogstream & sm,const IPAddresses &v){
	sm << "["<<v.size()<<"]:(";
	IPAddresses::const_iterator i;
	size_t j;
	for(j=1,i=v.begin();i!=v.end();i++,j++) sm << (j>1 ? ",":"") << (*i);
	return sm << ")";
}
bgplogstream & operator << (bgplogstream & sm,const IPNets &v){
	sm << "["<<v.size()<<"]:(";
	IPNets::const_iterator i;
	size_t j;
	for(j=1,i=v.begin();i!=v.end();i++,j++) sm << (j>1 ? ",":"") << (*i);
	return sm << ")";
}
bgplogstream & operator << (bgplogstream & sm,const MPLSLabels &v){
	sm << "["<<v.size()<<"]:(";
	MPLSLabels::const_iterator i;
	size_t j;
	for(j=1,i=v.begin();i!=v.end();i++,j++) sm << (j>1 ? ",":"") << (*i);
	return sm << ")";
}
bgplogstream & operator << (bgplogstream &sm,const IPTagNets &v){
	sm << "["<<v.size()<<"]:(";
	IPTagNets::const_iterator i;
	size_t j;
	for(j=1,i=v.begin();i!=v.end();i++,j++) sm << (j>1 ? ",":"") << (*i);
	return sm << ")";
}


bgplogstream & operator << (bgplogstream &sm,const MSG_Open &msg){
	sm << "Open(version=" << (unsigned)msg.version << " asn16=" << (unsigned)msg.asn16 << " holdtimer=" << msg.holdtimer << " routerid=" << msg.aRouterId.str();
	sm << ")";
	return sm;
}
bgplogstream & operator << (bgplogstream & sm,const MSG_UpdateAttr &msg){
	return sm << "UpdateAttr()";
}

void MSG_UpdateAttr::clear(){
    origin=0;
    vASPath.clear();
    aNextHop.clear();
    med=uEmpty;
    localpref=uEmpty;
    aAtomicAggregate.clear();
    aAggregatorAS=0;
    aAggregatorIP.clear();
    aOriginator.clear();
    vCom.clear();
    vExtCom.clear();
    vaClusterList.clear();
}

bool MSG_UpdateAttr::operator==(const MSG_UpdateAttr &x) const{
    if ((origin!=x.origin) || (localpref!=x.localpref) || (med!=x.med)) return false;
    if (aNextHop!=x.aNextHop) return false;
    if (aAtomicAggregate!=x.aAtomicAggregate) return false;
    if (aAggregatorIP!=x.aAggregatorIP) return false;
    if (aOriginator!=x.aOriginator) return false;
    if (aAggregatorAS!=x.aAggregatorAS) return false;
    if (vASPath!=x.vASPath) return false;
    if (vCom!=x.vCom) return false;
    if (vExtCom!=x.vExtCom) return false;
    if (vaClusterList!=x.vaClusterList) return false;
    return true;
}


}
