#ifndef __bgplog_h_
#define __bgplog_h_
#include "config.h"
#include <string>
#include <vector>
#include <map>
#include <iostream>
#include "bgppeer.h"


namespace BGPHist {
const uint32_t uEmpty = 0xffffffff;
typedef std::basic_ostream<char,std::char_traits<char> > bgplogstream;
bgplogstream & operator << (bgplogstream &,const IPAddress &);
bgplogstream & operator << (bgplogstream &,const IPAddresses &);
bgplogstream & operator << (bgplogstream &,const IPNet &);
bgplogstream & operator << (bgplogstream &,const IPTagNet &);
bgplogstream & operator << (bgplogstream &,const IPNets &);
bgplogstream & operator << (bgplogstream &,const MPLSLabels &);
bgplogstream & operator << (bgplogstream &,const IPTagNets &);

struct BGPCapabiliy {
    uint8_t code,length;
    uint8_t data[14];
};

typedef std::vector<BGPCapabiliy> BGPCapabilities;

struct BGPExtCommunity {
    uint16_t excType;
    uint16_t a;
    uint32_t b;
    std::string str() const;
    bool operator!=(const BGPExtCommunity &) const;
    bool operator==(const BGPExtCommunity &) const;
};
typedef std::vector<BGPExtCommunity> BGPExtCommunities;

typedef uint32_t BGPCommunity;
typedef std::vector<BGPCommunity> BGPCommunities;

struct MSG_Open{
        uint8_t version;
        uint16_t asn16;
        uint16_t holdtimer;
        IPAddress aRouterId;
        BGPCapabilities caps;
};
bgplogstream & operator << (bgplogstream &,const MSG_Open &);

struct MSG_Notify {
    uint8_t code;
};

struct MSG_UpdateAttr {
    uint8_t origin;
    BGPASNs vASPath;
    IPAddress aNextHop;
    uint32_t med,localpref;
    IPAddress aAtomicAggregate;
    BGPASN aAggregatorAS;
    IPAddress aAggregatorIP;
    IPAddress aOriginator;
    BGPCommunities vCom;
    BGPExtCommunities vExtCom;
    IPAddresses vaClusterList;
    void clear();
    bool operator==(const MSG_UpdateAttr &) const;
};
bgplogstream & operator << (bgplogstream &,const MSG_UpdateAttr &);

class BGPLog {
public:
    BGPLog(){}
    virtual ~BGPLog();
    virtual void open(BGPPeer *peer,const MSG_Open *)=0;
    virtual void close(BGPPeer *peer)=0;
    virtual void txt(BGPPeer *peer,const char *)=0;
    virtual void exception(BGPPeer *peer,const BGPException *)=0;
    virtual void notification(BGPPeer *peer,const MSG_Notify *)=0;
    virtual void keepalive(BGPPeer *peer)=0;
    virtual void nlri(BGPPeer *peer,const MSG_UpdateAttr &,const IPNets & )=0;
    virtual void withdraw(BGPPeer *peer,const MSG_UpdateAttr &,const IPNets & )=0;
    virtual void nlri_ipv4m(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPNets & )=0;
    virtual void withdraw_ipv4m(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPNets & )=0;
    virtual void nlri_ipv4lu(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & )=0;
    virtual void withdraw_ipv4lu(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & )=0;
    virtual void nlri_vpnv4u(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & )=0;
    virtual void withdraw_vpnv4u(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & )=0;
    virtual void nlri_ipv6u(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPNets & )=0;
    virtual void withdraw_ipv6u(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPNets & )=0;
    virtual void nlri_ipv6lu(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & )=0;
    virtual void withdraw_ipv6lu(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & )=0;
    virtual void nlri_vpnv6u(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & )=0;
    virtual void withdraw_vpnv6u(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & )=0;
    virtual void nlri_l2vpn(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & )=0;
    virtual void withdraw_l2vpn(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & )=0;
};

};

#endif
