#include "bgplogmysql.h"
#include "bgpcommon.h"
#include <unistd.h>
#include <iostream>
#include <sstream>
#include <mysql/mysql.h>

using namespace BGPHist;
using namespace std;

class MySqlException : public std::exception {
public:
    unsigned int uErrNo;
    std::string sMsg;
    MySqlException() throw();
    MySqlException(MYSQL *) throw();
    MySqlException(const std::string &) throw();
    MySqlException(const char *,MYSQL *) throw();
    virtual ~MySqlException() throw();
    virtual const char* what() const throw();
};
MySqlException::MySqlException() throw(){
}
MySqlException::MySqlException(const std::string &s) throw(){
    try {
        sMsg=s;
    } catch(...){}
}
MySqlException::MySqlException(MYSQL *m) throw(){
  try {
   uErrNo=mysql_errno(m);
   sMsg=mysql_error(m);
  } catch (...){
  }
}
MySqlException::MySqlException(const char *msg,MYSQL *m) throw(){
  try {
   uErrNo=mysql_errno(m);
   sMsg=msg;
   sMsg+=" : ";
   sMsg+=mysql_error(m);
  } catch (...){
  }
}

MySqlException::~MySqlException() throw(){
}
const char* MySqlException::what() const throw(){
    return sMsg.c_str();
}
std::string mysql_esc(MYSQL *m,const std::string &s){
	std::string r;
	r.resize(s.length()*2+1);
	r.resize(mysql_real_escape_string(m,const_cast<char *>(r.c_str()),s.c_str(),s.length()));
	return r;
}
#define CHECKMYSQL if (mysql_errno(mc)) throw MySqlException(mc);
#ifdef MYSQLSTMT
class BGPLogMy::MyStmt{
protected:
    size_t cBufSize;
    char * cBuf, *cSvBuf;
    void freeBuf();
    void allocBuf();
public:
     std::string sSql;
     BGPLogMy   *db;
     MYSQL_STMT *stmt;
     MYSQL_BIND *bind;
     unsigned params;
     size_t affected_rows;
     MyStmt(BGPLogMy *_db);
     ~MyStmt();
     void close();
     void init();
     void prepare(const char *);
     void bind_params();
     void execute();
     void save_buffer();
     void restore_buffer();
};
BGPLogMy::MyStmt::MyStmt(BGPLogMy *_db):db(_db),stmt(0),bind(0),cBuf(0),cSvBuf(0){
}
void BGPLogMy::MyStmt::init(){
    if (stmt) return ;
    MYSQL *mc=(MYSQL *)db->pmc;
    stmt=mysql_stmt_init(mc);
    if (!stmt) throw MySqlException("Unable to create prepared statement");
}
void BGPLogMy::MyStmt::close(){
    if (!stmt) return ;
    freeBuf();
    mysql_stmt_close(stmt);
    stmt=0;
}
BGPLogMy::MyStmt::~MyStmt(){
    close();
}
void BGPLogMy::MyStmt::freeBuf(){
    if (bind) free((void *)bind);bind=0;
    if (cBuf) free((void *)cBuf);cBuf=0;
    if (cSvBuf) free((void *)cSvBuf);cSvBuf=0;
}
void BGPLogMy::MyStmt::prepare(const char *_sql){
    if (!_sql) throw MySqlException("Invalid SQL string");
    sSql=_sql;
    init();
    if (mysql_stmt_prepare(stmt,_sql,strlen(_sql)))
       throw MySqlException(Sprintf("Unable to prepare '%s'\n",_sql));
    freeBuf();
    params=mysql_stmt_param_count(stmt);
    if (!params) return ;
    bind=(MYSQL_BIND *)malloc(sizeof(MYSQL_BIND)*params);
    memset(bind,0,sizeof(MYSQL_BIND)*params);
}
void BGPLogMy::MyStmt::bind_params(){
    unsigned i,cb;
    if (!bind || !params) return ;
    cb=0;
    for(i=0;i<params;i++){
        cb+=sizeof(my_bool)*2+sizeof(unsigned long)+bind[i].buffer_length;
    }
    if ((cb!=cBufSize) || !cBuf) {
        if (cBuf) free(cBuf);
        if (cSvBuf) free(cSvBuf);cSvBuf=0;
        cBuf=(char *)malloc(cb);
        cBufSize=cb;
        memset(cBuf,0,cb);
    }
    cb=0;
    for(i=0;i<params;i++){
        bind[i].is_null=(my_bool *)(cBuf+cb);
        bind[i].error=(my_bool *)(cBuf+cb+sizeof(my_bool));
        bind[i].length=(unsigned long *)(cBuf+cb+sizeof(my_bool)*2);
        bind[i].buffer=(void *)(cBuf+cb+sizeof(my_bool)*2+sizeof(unsigned long));
        cb+=sizeof(my_bool)*2+sizeof(unsigned long)+bind[i].buffer_length;
    }
    if (mysql_stmt_bind_param(stmt, bind)) throw MySqlException("Unable to bind params",(MYSQL *)db->pmc);
}
void BGPLogMy::MyStmt::save_buffer(){
    if (!stmt || !cBuf) return ;
    if (!cSvBuf) cSvBuf=(char *)malloc(cBufSize);
    memcpy(cSvBuf,cBuf,cBufSize);
}
void BGPLogMy::MyStmt::restore_buffer(){
    if (!stmt || !cBuf || !cSvBuf) return ;
    memcpy(cBuf,cSvBuf,cBufSize);
}

void BGPLogMy::MyStmt::execute(){
    if (!stmt) throw MySqlException("No statement");
    if (mysql_stmt_execute(stmt)) throw MySqlException(std::string(std::string("Error while executing '")+sSql+"'").c_str(),(MYSQL *)db->pmc); 
    affected_rows= mysql_stmt_affected_rows(stmt);
}
void BGPLogMy::prepare_vSQ(){
    MyStmt *st;
    unsigned i;
    vSQ[0]=st=new MyStmt(this);
    st->prepare("insert delayed into bgp_commonlog (srcid,msgcode,msgtxt) values (?,?,?)");
    st->bind[0].buffer_type= MYSQL_TYPE_LONG;
    st->bind[0].buffer_length=4;
    st->bind[1].buffer_type= MYSQL_TYPE_LONG;
    st->bind[1].buffer_length=4;
    st->bind[2].buffer_type= MYSQL_TYPE_STRING;
    st->bind[2].buffer_length=255;
    st->bind_params();
    *st->bind[0].length=4;
    *(unsigned long *)st->bind[0].buffer=uSrcId;
    *st->bind[0].is_null=false;
    *st->bind[1].length=4;
    *st->bind[1].is_null=false;
    *st->bind[2].is_null=false;

/*
            wr_ipv4(IN _srcid INT,IN state tinyint,IN _kind tinyint,IN _netaddr int unsigned,IN _netsize tinyint unsigned,IN _bgporigin tinyint unsigned,
 IN _bgpaspath varchar(255),IN _bgpcommunities varchar(255),IN _bgpnexthop int unsigned,IN _bgpmed int unsigned,IN _bgplocalpref int unsigned,
 IN _bgpatomicaggregate int unsigned,IN _bgpagregatoras int unsigned,IN _bgpagregatorIP int unsigned,IN _bgporiginatorIP int unsigned,IN _bgpextcom varchar(255),
 IN _bgplabels varchar(255)
*/
    vSQ[1]=st=new MyStmt(this);
    st->prepare("call wr_ipv4 (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)");
    st->bind[0].buffer_type= MYSQL_TYPE_LONG;
    st->bind[0].buffer_length=4;
    st->bind[1].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[1].buffer_length=2;
    st->bind[2].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[2].buffer_length=2;
    st->bind[3].buffer_type= MYSQL_TYPE_LONG;
    st->bind[3].buffer_length=4;
    st->bind[3].is_unsigned=true;
    st->bind[4].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[4].buffer_length=2;
    st->bind[5].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[5].buffer_length=2;
    st->bind[6].buffer_type= MYSQL_TYPE_STRING;
    st->bind[6].buffer_length=255;
    st->bind[7].buffer_type= MYSQL_TYPE_STRING;
    st->bind[7].buffer_length=255;
    st->bind[8].buffer_type= MYSQL_TYPE_LONG;
    st->bind[8].buffer_length=4;
    st->bind[8].is_unsigned=true;
    st->bind[9].buffer_type= MYSQL_TYPE_LONG;
    st->bind[9].buffer_length=4;
    st->bind[9].is_unsigned=true;
    st->bind[10].buffer_type= MYSQL_TYPE_LONG;
    st->bind[10].buffer_length=4;
    st->bind[10].is_unsigned=true;
    st->bind[11].buffer_type= MYSQL_TYPE_LONG;
    st->bind[11].buffer_length=4;
    st->bind[11].is_unsigned=true;
    st->bind[12].buffer_type= MYSQL_TYPE_LONG;
    st->bind[12].buffer_length=4;
    st->bind[12].is_unsigned=true;
    st->bind[13].buffer_type= MYSQL_TYPE_LONG;
    st->bind[13].buffer_length=4;
    st->bind[13].is_unsigned=true;
    st->bind[14].buffer_type= MYSQL_TYPE_LONG;
    st->bind[14].buffer_length=4;
    st->bind[14].is_unsigned=true;
    st->bind[15].buffer_type= MYSQL_TYPE_STRING;
    st->bind[15].buffer_length=255;
    st->bind[16].buffer_type= MYSQL_TYPE_STRING;
    st->bind[16].buffer_length=255;
    st->bind_params();
    for(i=0;i<st->params;i++) {*st->bind[i].length=st->bind[i].buffer_length;*st->bind[i].is_null=false;}
    *(unsigned long *)st->bind[0].buffer=uSrcId;
    
    /*
wr_vpnv4(IN _srcid INT,IN _state tinyint,IN _kind tinyint,IN _netaddrrdh int unsigned,IN _netaddrrdl int unsigned,IN _netaddr int unsigned,IN _netsize tinyint unsigned,
IN _bgporigin tinyint unsigned,IN _bgpaspath varchar(255),IN _bgpcommunities varchar(255),
IN _nhrrdh int unsigned,IN _nhrrdl int unsigned,IN _nh int unsigned,IN _bgpmed int unsigned,IN _bgplocalpref int unsigned,
 IN _bgpatomicaggregate int unsigned,IN _bgpaggregatoras int unsigned,IN _bgpagregatorIP int unsigned,IN _bgporiginatorIP int unsigned,IN _bgpextcom varchar(255),
 IN _bgplabels varchar(255)
     */
    vSQ[2]=st=new MyStmt(this);
    st->prepare("call wr_vpnv4 (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)");
    i=0;//srcid
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    i++;//state
    st->bind[i].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[i].buffer_length=2;
    i++;//kind
    st->bind[i].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[i].buffer_length=2;
    i++;//netaddrrdh
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//netaddrrdl
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//netaddr
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//netsize
    st->bind[i].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[i].buffer_length=2;
    i++;//bgporigin
    st->bind[i].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[i].buffer_length=2;
    i++;//bgpaspath
    st->bind[i].buffer_type= MYSQL_TYPE_STRING;
    st->bind[i].buffer_length=255;
    i++;//bgpcommunities
    st->bind[i].buffer_type= MYSQL_TYPE_STRING;
    st->bind[i].buffer_length=255;
    i++;//nexthoprdh
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//nexthoprdl
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//nexthopip
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgpmed
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgplocalpref
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgpatomicagregate
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgpagregatorAS
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgpagregatorIP
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgporiginatorIP
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgpextcommunity
    st->bind[i].buffer_type= MYSQL_TYPE_STRING;
    st->bind[i].buffer_length=255;
    i++;//bgplabels
    st->bind[i].buffer_type= MYSQL_TYPE_STRING;
    st->bind[i].buffer_length=255;
    st->bind_params();
    for(i=0;i<st->params;i++) {*st->bind[i].length=st->bind[i].buffer_length;*st->bind[i].is_null=false;}
    *(unsigned long *)st->bind[0].buffer=uSrcId;

    /*
CREATE PROCEDURE wr_ipv6(IN _srcid INT,IN _state tinyint,IN _kind tinyint,IN _netaddrh bigint unsigned,IN _netaddrl bigint unsigned,IN _netsize tinyint unsigned,IN _bgporigin tinyint unsigned,
 IN _bgpaspath varchar(255),IN _bgpcommunities varchar(255),IN _bgpnexthoph bigint unsigned,IN _bgpnexthopl bigint unsigned,IN _bgpmed int unsigned,IN _bgplocalpref int unsigned,
 IN _bgpatomicaggregateh bigint unsigned,IN _bgpatomicaggregatel bigint unsigned,IN _bgpaggregatoras int unsigned,
 IN _bgpagregatorIPh bigint unsigned,IN _bgpagregatorIPl bigint unsigned,
 IN _bgporiginatorIPh bigint unsigned,IN _bgporiginatorIPl bigint unsigned,IN _bgpextcom varchar(255),
 IN _bgplabels varchar(255)
)
    */    
    vSQ[3]=st=new MyStmt(this);
    st->prepare("call wr_ipv6 (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)");
    i=0;//srcid
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    i++;//state
    st->bind[i].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[i].buffer_length=2;
    i++;//kind
    st->bind[i].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[i].buffer_length=2;
    i++;//netaddrh
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//netaddrl
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//netsize
    st->bind[i].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[i].buffer_length=2;
    i++;//bgporigin
    st->bind[i].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[i].buffer_length=2;
    i++;//bgpaspath
    st->bind[i].buffer_type= MYSQL_TYPE_STRING;
    st->bind[i].buffer_length=255;
    i++;//bgpcommunities
    st->bind[i].buffer_type= MYSQL_TYPE_STRING;
    st->bind[i].buffer_length=255;
    i++;//nexthoph
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//nexthopl
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//bgpmed
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgplocalpref
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgpatomicagregateh
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//bgpatomicagregatel
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//bgpagregatorAS
    st->bind[i].buffer_type=MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgpagregatorIPh
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//bgpagregatorIPl
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//bgporiginatorIPh
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//bgporiginatorIPl
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//bgpextcommunity
    st->bind[i].buffer_type= MYSQL_TYPE_STRING;
    st->bind[i].buffer_length=255;
    i++;//bgplabels
    st->bind[i].buffer_type= MYSQL_TYPE_STRING;
    st->bind[i].buffer_length=255;
    st->bind_params();
    for(i=0;i<st->params;i++) {*st->bind[i].length=st->bind[i].buffer_length;*st->bind[i].is_null=false;}
    *(unsigned long *)st->bind[0].buffer=uSrcId;

    /*
CREATE PROCEDURE wr_vpnv6(IN _srcid INT,IN _state tinyint,IN _kind tinyint,IN _netaddrrdh int unsigned,IN _netaddrrdl int unsigned,IN _netaddrh bigint unsigned,IN _netaddrl bigint unsigned,IN _netsize tinyint unsigned,
IN _bgporigin tinyint unsigned,IN _bgpaspath varchar(255),IN _bgpcommunities varchar(255),
IN _nhrrdh int unsigned,IN _nhrrdl int unsigned,IN _nhh bigint unsigned,IN _nhl bigint unsigned,IN _bgpmed int unsigned,IN _bgplocalpref int unsigned,
 IN _bgpatomicaggregateh bigint unsigned,IN _bgpatomicaggregatel bigint unsigned,IN _bgpaggregatoras int unsigned,
 IN _bgpagregatorIPh bigint unsigned,IN _bgpagregatorIPl bigint unsigned,IN _bgporiginatorIPh bigint unsigned,IN _bgporiginatorIPl bigint unsigned,IN _bgpextcom varchar(255),
 IN _bgplabels varchar(255)
)
    */
    vSQ[4]=st=new MyStmt(this);
    st->prepare("call wr_vpnv6 (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)");
    i=0;//srcid
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    i++;//state
    st->bind[i].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[i].buffer_length=2;
    i++;//kind
    st->bind[i].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[i].buffer_length=2;
    i++;//netaddrrdh
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//netaddrrdl
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//netaddrh
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//netaddrl
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//netsize
    st->bind[i].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[i].buffer_length=2;
    i++;//bgporigin
    st->bind[i].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[i].buffer_length=2;
    i++;//bgpaspath
    st->bind[i].buffer_type= MYSQL_TYPE_STRING;
    st->bind[i].buffer_length=255;
    i++;//bgpcommunities
    st->bind[i].buffer_type= MYSQL_TYPE_STRING;
    st->bind[i].buffer_length=255;
    i++;//nexthoprdh
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//nexthoprdl
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//nexthoph
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//nexthopl
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//bgpmed
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgplocalpref
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgpatomicagregateh
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//bgpatomicagregatel
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//bgpagregatorAS
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgpagregatorIPh
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//bgpagregatorIPl
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//bgporiginatorIPh
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//bgporiginatorIPl
    st->bind[i].buffer_type=MYSQL_TYPE_LONGLONG;
    st->bind[i].buffer_length=8;
    st->bind[i].is_unsigned=true;
    i++;//bgpextcommunity
    st->bind[i].buffer_type= MYSQL_TYPE_STRING;
    st->bind[i].buffer_length=255;
    i++;//bgplabels
    st->bind[i].buffer_type= MYSQL_TYPE_STRING;
    st->bind[i].buffer_length=255;
    st->bind_params();
    for(i=0;i<st->params;i++) {*st->bind[i].length=st->bind[i].buffer_length;*st->bind[i].is_null=false;}
    *(unsigned long *)st->bind[0].buffer=uSrcId;

/*
            wr_l2vpn(IN _srcid INT,IN _state tinyint,IN _kind tinyint,IN _netaddrrdh int unsigned,IN _netaddrrdl int unsigned,IN _site smallint unsigned,IN _offset smallint unsigned,IN _srange smallint unsigned,
IN _bgporigin tinyint unsigned,IN _bgpaspath varchar(255),IN _bgpcommunities varchar(255),
IN _nh int unsigned,IN _bgpmed int unsigned,IN _bgplocalpref int unsigned,
 IN _bgpatomicaggregate int unsigned,IN _bgpaggregatoras int unsigned,IN _bgpagregatorIP int unsigned,IN _bgporiginatorIP int unsigned,IN _bgpextcom varchar(255),
 IN _bgplabels varchar(255)
)
*/
    vSQ[5]=st=new MyStmt(this);
    st->prepare("call wr_l2vpn (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)");
    i=0;//srcid
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    i++;//state
    st->bind[i].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[i].buffer_length=2;
    i++;//kind
    st->bind[i].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[i].buffer_length=2;
    i++;//netaddrrdh
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//netaddrrdl
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//site
    st->bind[i].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[i].buffer_length=2;
    st->bind[i].is_unsigned=true;
    i++;//offset
    st->bind[i].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[i].buffer_length=2;
    st->bind[i].is_unsigned=true;
    i++;//srange
    st->bind[i].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[i].buffer_length=2;
    st->bind[i].is_unsigned=true;
    i++;//bgporigin
    st->bind[i].buffer_type= MYSQL_TYPE_SHORT;
    st->bind[i].buffer_length=2;
    i++;//bgpaspath
    st->bind[i].buffer_type= MYSQL_TYPE_STRING;
    st->bind[i].buffer_length=255;
    i++;//bgpcommunities
    st->bind[i].buffer_type= MYSQL_TYPE_STRING;
    st->bind[i].buffer_length=255;
    i++;//nexthop
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgpmed
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgplocalpref
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgpatomicagregate
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgpagregatorAS
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgpagregatorIP
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgporiginatorIP
    st->bind[i].buffer_type= MYSQL_TYPE_LONG;
    st->bind[i].buffer_length=4;
    st->bind[i].is_unsigned=true;
    i++;//bgpextcommunity
    st->bind[i].buffer_type= MYSQL_TYPE_STRING;
    st->bind[i].buffer_length=255;
    i++;//bgplabels
    st->bind[i].buffer_type= MYSQL_TYPE_STRING;
    st->bind[i].buffer_length=255;
    st->bind_params();
    for(i=0;i<st->params;i++) {*st->bind[i].length=st->bind[i].buffer_length;*st->bind[i].is_null=false;}
    *(unsigned long *)st->bind[0].buffer=uSrcId;
}
#endif

BGPLogMy::BGPLogMy(const char *_myhost,const char *_mydb,const char *_myuser,const char *_mypwd,const char *_peerip,unsigned int _asn):pmc(0),uSrcId(0){
    MYSQL * mc=new MYSQL;
    int i;
    pmc=mc;
    MYSQL_RES *res;
    MYSQL_ROW row;
    mysql_init(mc);
    if (!mysql_real_connect(mc,_myhost,_myuser,_mypwd,0,0,0,0)){
       throw MySqlException(Sprintf("Unable to connect to %s:%s:%s\n",_myhost,_myuser,_mypwd));
    }
    if (mysql_select_db(mc,_mydb)){
       mysql_close(mc);
       throw MySqlException(Sprintf("Unable to select db %s\n",_mydb));
    }
    std::string sQ=Sprintf("select id from bgp_src where peerip='%s' and peeras='%u'",_peerip,_asn);
    mysql_query(mc,sQ.c_str());
    CHECKMYSQL;
    if (!(res = mysql_store_result(mc))) throw MySqlException("Unable to resolve source id");
    if((row = mysql_fetch_row(res))) {
        if (row[0]){
            sscanf(row[0],"%u",&uSrcId);
        }
    }
    mysql_free_result(res);
    if (!uSrcId) {
        sQ=Sprintf("insert into bgp_src (peerip,peeras) values ('%s','%u')",_peerip,_asn);
        mysql_query(mc,sQ.c_str());
        CHECKMYSQL;
        uSrcId=mysql_insert_id(mc);
    }
    const char *szTbls[]={"bgp_ipv4","bgp_vpnv4","bgp_ipv6","bgp_vpnv6",0};
    for(i=0;szTbls[i];i++){
        sQ=Sprintf("update %s set state=2 where srcid=%u",szTbls[i],uSrcId);
        mysql_query(mc,sQ.c_str());CHECKMYSQL;
        mysql_query(mc,"commit");CHECKMYSQL;
    }
#ifdef MYSQLSTMT
    prepare_vSQ();
#endif
    logtxt(1,"Started");
}
BGPLogMy::~BGPLogMy(){
    if (pmc){
        logtxt(2,"Finished");
#ifdef MYSQLSTMT
        TSQ::iterator i;
        for(i=vSQ.begin();i!=vSQ.end();i++) {if (i->second) delete i->second;i->second=0;}
        vSQ.clear();
#endif
        MYSQL * mc=(MYSQL *)pmc;
        mysql_close(mc);
        delete mc;
    }
}
void BGPLogMy::logtxt(unsigned uCode,const char *sTxt){
    if (!pmc) return ;
#ifndef MYSQLSTMT        
    MYSQL * mc=(MYSQL *)pmc;
    std::string sQ=Sprintf("insert delayed into bgp_commonlog (srcid,msgcode,msgtxt) values (%u,%u,'",uSrcId,uCode);
    sQ+=mysql_esc(mc,sTxt)+"')";
    mysql_query(mc,sQ.c_str());
    CHECKMYSQL;
#else
    MyStmt *st=vSQ[0];
    for(int i=0;i<st->params;i++) {*st->bind[i].length=st->bind[i].buffer_length;*st->bind[i].is_null=false;}
    *(unsigned long *)st->bind[1].buffer=uCode;
    *st->bind[2].length=strlen(sTxt);
    strncpy((char *)st->bind[2].buffer,sTxt,*st->bind[2].length);
    st->execute();
#endif    
}

void BGPLogMy::open(BGPPeer *peer,const MSG_Open *msg){
    //cout << (*msg) << endl;
    stringstream ss(stringstream::in | stringstream::out);
    ss << (*msg);
    std::string s;
    ss >> s;
    logtxt(3,s.c_str());
}
void BGPLogMy::close(BGPPeer *peer){
    logtxt(4,"close");
}
void BGPLogMy::txt(BGPPeer *peer,const char *txt){
	logtxt(0,txt);
}
void BGPLogMy::exception(BGPPeer *peer,const BGPException *e){
	if (e) cout << e->what() << endl;
}
void BGPLogMy::notification(BGPPeer *peer,const MSG_Notify *msg){
    /*
    stringstream ss(stringstream::in | stringstream::out);
    ss << (*msg);
    std::string s;
    ss >> s;
    logtxt(4,s.c_str());
    */
}
void BGPLogMy::keepalive(BGPPeer *peer){
    //logtxt(5,"keepalive");
}
void BGPLogMy::exec_ipv4(unsigned state,unsigned kind,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPNets & nets){
    if (!pmc) return ;
    MYSQL * mc=(MYSQL *)pmc;
    std::string sQ,sV,sS;
    unsigned i,u;
    IPNets::const_iterator ci;
#ifndef MYSQLSTMT 
    sQ=Sprintf("call wr_ipv4 (%u,%u,%u",uSrcId,state,kind);
    sV="";
    sV+=Sprintf(",%u",(unsigned)attr.origin);
    if (attr.vASPath.size()>0){
        sV+=",'";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        sV+="'";
    } else sV+=",null";
    if (attr.vCom.size()>0){
        sV+=",'";
        BGPCommunities::const_iterator cmi;
        i=0;
        for(cmi=attr.vCom.begin();cmi!=attr.vCom.end();cmi++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)(*cmi)) >> 16,((unsigned)(*cmi)) & 0xffff);
        }
        sV+="'";
    } else sV+=",null";
    sV+=Sprintf(",%u",(unsigned)nexthop.getIPv4Addr());
    if (attr.med==uEmpty) sV+=",null"; else sV+=Sprintf(",%u",(unsigned)attr.med);
    if (attr.localpref==uEmpty) sV+=",null"; else sV+=Sprintf(",%u",(unsigned)attr.localpref);
    sV+=Sprintf(",%u",(unsigned)attr.aAtomicAggregate.getIPv4Addr());
    if (attr.aAggregatorAS!=0){
        sV+=Sprintf(",%u",(unsigned)attr.aAggregatorAS);
    } else sV+=",null";
    if (attr.aAggregatorIP.getType()==atIPv4){
        sV+=Sprintf(",%u",(unsigned)attr.aAggregatorIP.getIPv4Addr());
    } else sV+=",null";
    if (attr.aOriginator.getType()==atIPv4){
        sV+=Sprintf(",%u",(unsigned)attr.aOriginator.getIPv4Addr());
    } else sV+=",null";
    if (attr.vExtCom.size()>0){
        sV+=",'";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        sV+="'";
    } else sV+=",null";
    sV+=",null";//labels
    sV+=")";
#else
    unsigned n;
    MyStmt *st=vSQ[1];
    for(n=0;n<st->params;n++) {*st->bind[n].length=st->bind[n].buffer_length;*st->bind[n].is_null=false;}
    *(unsigned *)st->bind[0].buffer=uSrcId;
    *(short *)st->bind[1].buffer=state;
    *(short *)st->bind[2].buffer=kind;
    *(short *)st->bind[5].buffer=attr.origin;
    if (attr.vASPath.size()>0){
        *st->bind[6].is_null=false;
        sV="";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        *st->bind[6].length=min<size_t>(sV.size(),st->bind[6].buffer_length);
        strncpy((char *)st->bind[6].buffer,sV.c_str(),st->bind[6].buffer_length);
    } else {
        *st->bind[6].is_null=true;
        *st->bind[6].length=0;
    }
    if (attr.vCom.size()>0){
        *st->bind[7].is_null=false;
        sV="";
        BGPCommunities::const_iterator cmi;
        i=0;
        for(cmi=attr.vCom.begin();cmi!=attr.vCom.end();cmi++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)(*cmi)) >> 16,((unsigned)(*cmi)) & 0xffff);
        }
        *st->bind[7].length=min<size_t>(sV.size(),st->bind[7].buffer_length);
        strncpy((char *)st->bind[7].buffer,sV.c_str(),st->bind[7].buffer_length);
    } else {
        *st->bind[7].is_null=true;
        *st->bind[7].length=0;
    }
    *(unsigned *)st->bind[8].buffer=nexthop.getIPv4Addr();
    if (attr.med==uEmpty) {*st->bind[9].is_null=true;*st->bind[9].length=0;} else {*st->bind[9].is_null=false;*st->bind[9].length=4;*(unsigned *)st->bind[9].buffer=attr.med;}
    if (attr.localpref==uEmpty) {*st->bind[10].is_null=true;*st->bind[10].length=0;} else {*st->bind[10].is_null=false;*st->bind[10].length=4;*(unsigned *)st->bind[10].buffer=attr.localpref;}
    *(unsigned *)st->bind[11].buffer=attr.aAtomicAggregate.getIPv4Addr();
    if (attr.aAggregatorAS!=0){
        *st->bind[12].is_null=false;*st->bind[12].length=4;*(unsigned *)st->bind[12].buffer=attr.aAggregatorAS;
    } else {*st->bind[12].is_null=true;*st->bind[12].length=0;}
    if (attr.aAggregatorIP.getType()==atIPv4){
        *st->bind[13].is_null=false;*st->bind[13].length=4;*(unsigned *)st->bind[13].buffer=attr.aAggregatorIP.getIPv4Addr();
    } else {*st->bind[13].is_null=true;*st->bind[13].length=0;}
    if (attr.aOriginator.getType()==atIPv4){
        *st->bind[14].is_null=false;*st->bind[14].length=4;*(unsigned *)st->bind[14].buffer=attr.aOriginator.getIPv4Addr();
    } else {*st->bind[14].is_null=true;*st->bind[14].length=0;}
    if (attr.vExtCom.size()>0){
        sV="";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        *st->bind[15].is_null=false;
        *st->bind[15].length=min<size_t>(sV.size(),st->bind[15].buffer_length);
        strncpy((char *)st->bind[15].buffer,sV.c_str(),st->bind[15].buffer_length);
    } else {
        *st->bind[15].is_null=true;
        *st->bind[15].length=0;
    }
    *st->bind[16].is_null=true;
    *st->bind[16].length=0;
#endif
    for(ci=nets.begin();ci!=nets.end();ci++){
#ifndef MYSQLSTMT        
        sS=sQ+Sprintf(",%u,%u",ci->getIPv4Addr(),ci->getPrefixLen())+sV+",'";
        for(i=0,mi=ci->labels.begin();mi!=ci->labels.end();mi++,i++){
            if (i) sS+=" ";
            sS+=Sprintf("%u",(unsigned)*mi);
        }
        sS+="')";
        mysql_query(mc,sS.c_str());
        CHECKMYSQL;
#else
        *(unsigned *)st->bind[3].buffer=ci->getIPv4Addr();
        *(short *)st->bind[4].buffer=ci->getPrefixLen();
        st->save_buffer();
        st->execute();
        st->restore_buffer();
#endif        
    }
}
void BGPLogMy::exec_ipv6(unsigned state,unsigned kind,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPNets & nets){
    if (!pmc) return ;
    MYSQL * mc=(MYSQL *)pmc;
    std::string sQ,sV,sS;
    unsigned i,u;
    IPNets::const_iterator ci;
#ifndef MYSQLSTMT
    sQ=Sprintf("call wr_ipv6 (%u,%u,%u",uSrcId,state,kind);
    sV="";
    sV+=Sprintf(",%u",(unsigned)attr.origin);
    if (attr.vASPath.size()>0){
        sV+=",'";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        sV+="'";
    } else sV+=",null";
    if (attr.vCom.size()>0){
        sV+=",'";
        BGPCommunities::const_iterator cmi;
        i=0;
        for(cmi=attr.vCom.begin();cmi!=attr.vCom.end();cmi++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)(*cmi)) >> 16,((unsigned)(*cmi)) & 0xffff);
        }
        sV+="'";
    } else sV+=",null";
    sV+=Sprintf(",%llu,%llu",nexthop.getIPv6AddrH(),nexthop.getIPv6AddrL());
    if (attr.med==uEmpty) sV+=",null"; else sV+=Sprintf(",%u",(unsigned)attr.med);
    if (attr.localpref==uEmpty) sV+=",null"; else sV+=Sprintf(",%u",(unsigned)attr.localpref);
    sV+=Sprintf(",%llu,%llu",attr.aAtomicAggregate.getIPv6AddrH(),attr.aAtomicAggregate.getIPv6AddrL());
    if (attr.aAggregatorAS!=0){
        sV+=Sprintf(",%u",(unsigned)attr.aAggregatorAS);
    } else sV+=",null";
    if (attr.aAggregatorIP.getType()==atIPv6){
        sV+=Sprintf(",%llu,%llu",attr.aAggregatorIP.getIPv6AddrH(),attr.aAggregatorIP.getIPv6AddrL());
    } else sV+=",null,null";
    if (attr.aOriginator.getType()==atIPv6){
        sV+=Sprintf(",%llu,%llu",attr.aOriginator.getIPv6AddrH(),attr.aOriginator.getIPv6AddrL());
    } else sV+=",null,null";
    if (attr.vExtCom.size()>0){
        sV+=",'";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        sV+="'";
    } else sV+=",null";
    sV+=",null";//labels
    sV+=")";
#else
    unsigned n;
    MyStmt *st=vSQ[3];
    for(n=0;n<st->params;n++) {*st->bind[n].length=st->bind[n].buffer_length;*st->bind[n].is_null=false;}
    *(unsigned *)st->bind[0].buffer=uSrcId;
    *(short *)st->bind[1].buffer=state;
    *(short *)st->bind[2].buffer=kind;
    *(short *)st->bind[6].buffer=attr.origin;
    n=7;
    if (attr.vASPath.size()>0){
        *st->bind[n].is_null=false;
        sV="";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        *st->bind[n].length=min<size_t>(sV.size(),st->bind[n].buffer_length);
        strncpy((char *)st->bind[n].buffer,sV.c_str(),st->bind[n].buffer_length);
    } else {
        *st->bind[n].is_null=true;
        *st->bind[n].length=0;
    }
    n=8;
    if (attr.vCom.size()>0){
        *st->bind[n].is_null=false;
        sV="";
        BGPCommunities::const_iterator cmi;
        i=0;
        for(cmi=attr.vCom.begin();cmi!=attr.vCom.end();cmi++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)(*cmi)) >> 16,((unsigned)(*cmi)) & 0xffff);
        }
        *st->bind[n].length=min<size_t>(sV.size(),st->bind[n].buffer_length);
        strncpy((char *)st->bind[n].buffer,sV.c_str(),st->bind[n].buffer_length);
    } else {
        *st->bind[n].is_null=true;
        *st->bind[n].length=0;
    }
    *(uint64_t *)st->bind[9].buffer=nexthop.getIPv6AddrH();
    *(uint64_t *)st->bind[10].buffer=nexthop.getIPv6AddrL();
    n=11;if (attr.med==uEmpty) {*st->bind[n].is_null=true;*st->bind[n].length=0;} else {*st->bind[n].is_null=false;*st->bind[n].length=4;*(unsigned *)st->bind[n].buffer=attr.med;}
    n=12;if (attr.localpref==uEmpty) {*st->bind[n].is_null=true;*st->bind[n].length=0;} else {*st->bind[n].is_null=false;*st->bind[n].length=4;*(unsigned *)st->bind[n].buffer=attr.localpref;}
    n=13;*(uint64_t *)st->bind[n].buffer=attr.aAtomicAggregate.getIPv6AddrH();*(uint64_t *)st->bind[n+1].buffer=attr.aAtomicAggregate.getIPv6AddrL();
    n=15;
    if (attr.aAggregatorAS!=0){
        *st->bind[n].is_null=false;*st->bind[n].length=4;*(unsigned *)st->bind[n].buffer=attr.aAggregatorAS;
    } else {*st->bind[n].is_null=true;*st->bind[n].length=0;}
    n=16;
    if (attr.aAggregatorIP.getType()==atIPv6){
        *st->bind[n].is_null=false;*st->bind[n].length=8;*(uint64_t *)st->bind[n].buffer=attr.aAggregatorIP.getIPv6AddrH();
        *st->bind[n+1].is_null=false;*st->bind[n+1].length=8;*(uint64_t *)st->bind[n+1].buffer=attr.aAggregatorIP.getIPv6AddrL();
    } else {*st->bind[n].is_null=true;*st->bind[n].length=0;*st->bind[n+1].is_null=true;*st->bind[n+1].length=0;}
    n=18;
    if (attr.aOriginator.getType()==atIPv6){
        *st->bind[n].is_null=false;*st->bind[n].length=8;*(uint64_t *)st->bind[n].buffer=attr.aOriginator.getIPv6AddrH();
        *st->bind[n+1].is_null=false;*st->bind[n+1].length=8;*(uint64_t *)st->bind[n+1].buffer=attr.aOriginator.getIPv6AddrL();
    } else {*st->bind[n].is_null=true;*st->bind[n].length=0;*st->bind[n+1].is_null=true;*st->bind[n+1].length=0;}
    n=20;
    if (attr.vExtCom.size()>0){
        sV="";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        *st->bind[n].is_null=false;
        *st->bind[n].length=min<size_t>(sV.size(),st->bind[n].buffer_length);
        strncpy((char *)st->bind[n].buffer,sV.c_str(),st->bind[n].buffer_length);
    } else {
        *st->bind[n].is_null=true;
        *st->bind[n].length=0;
    }
    n=21;
    *st->bind[n].is_null=true;
    *st->bind[n].length=0;
#endif
    for(ci=nets.begin();ci!=nets.end();ci++){
#ifndef MYSQLSTMT
        sS=sQ+Sprintf(",%llu,%llu,%u",ci->getIPv6AddrH(),ci->getIPv6AddrL(),ci->getPrefixLen())+sV;
        //fprintf(stderr,"%s\n",sS.c_str());
        mysql_query(mc,sS.c_str());
        CHECKMYSQL;
#else
        *(uint64_t *)st->bind[3].buffer=ci->getIPv6AddrH();
        *(uint64_t *)st->bind[4].buffer=ci->getIPv6AddrL();
        *(short *)st->bind[5].buffer=ci->getPrefixLen();
        st->save_buffer();
        st->execute();
        st->restore_buffer();
#endif
    }
}
void BGPLogMy::exec_ipv4l(unsigned state,unsigned kind,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    if (!pmc) return ;
    MYSQL * mc=(MYSQL *)pmc;
    std::string sQ,sV,sS;
    IPTagNets::const_iterator ci;
    MPLSLabels::const_iterator mi;
    unsigned i;
#ifndef MYSQLSTMT        
    sQ=Sprintf("call wr_ipv4 (%u,%u,%u",uSrcId,state,kind);
    sV="";
    sV+=Sprintf(",%u",(unsigned)attr.origin);
    if (attr.vASPath.size()>0){
        sV+=",'";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        sV+="'";
    } else sV+=",null";
    if (attr.vCom.size()>0){
        sV+=",'";
        BGPCommunities::const_iterator ci;
        i=0;
        for(ci=attr.vCom.begin();ci!=attr.vCom.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)*ci) >> 16,((unsigned)*ci) & 0xffff);
        }
        sV+="'";
    } else sV+=",null";
    sV+=Sprintf(",%u",(unsigned)nexthop.getIPv4Addr());
    if (attr.med==uEmpty) sV+=",null"; else sV+=Sprintf(",%u",(unsigned)attr.med);
    if (attr.localpref==uEmpty) sV+=",null"; else sV+=Sprintf(",%u",(unsigned)attr.localpref);
    sV+=Sprintf(",%u",(unsigned)attr.aAtomicAggregate.getIPv4Addr());
    if (attr.aAggregatorAS!=0){
        sV+=Sprintf(",%u",(unsigned)attr.aAggregatorAS);
    } else sV+=",null";
    if (attr.aAggregatorIP.getType()==atIPv4){
        sV+=Sprintf(",%u",(unsigned)attr.aAggregatorIP.getIPv4Addr());
    } else sV+=",null";
    if (attr.aOriginator.getType()==atIPv4){
        sV+=Sprintf(",%u",(unsigned)attr.aOriginator.getIPv4Addr());
    } else sV+=",null";
    if (attr.vExtCom.size()>0){
        sV+=",'";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        sV+="'";
    } else sV+=",null";
#else
    unsigned n;
    MyStmt *st=vSQ[1];
    for(n=0;n<st->params;n++) {*st->bind[n].length=st->bind[n].buffer_length;*st->bind[n].is_null=false;}
    *(unsigned *)st->bind[0].buffer=uSrcId;
    *(short *)st->bind[1].buffer=state;
    *(short *)st->bind[2].buffer=kind;
    *(short *)st->bind[5].buffer=attr.origin;
    if (attr.vASPath.size()>0){
        *st->bind[6].is_null=false;
        sV="";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        *st->bind[6].length=min<size_t>(sV.size(),st->bind[6].buffer_length);
        strncpy((char *)st->bind[6].buffer,sV.c_str(),st->bind[6].buffer_length);
    } else {
        *st->bind[6].is_null=true;
        *st->bind[6].length=0;
    }
    if (attr.vCom.size()>0){
        *st->bind[7].is_null=false;
        sV="";
        BGPCommunities::const_iterator cmi;
        i=0;
        for(cmi=attr.vCom.begin();cmi!=attr.vCom.end();cmi++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)(*cmi)) >> 16,((unsigned)(*cmi)) & 0xffff);
        }
        *st->bind[7].length=min<size_t>(sV.size(),st->bind[7].buffer_length);
        strncpy((char *)st->bind[7].buffer,sV.c_str(),st->bind[7].buffer_length);
    } else {
        *st->bind[7].is_null=true;
        *st->bind[7].length=0;
    }
    *(unsigned *)st->bind[8].buffer=nexthop.getIPv4Addr();
    if (attr.med==uEmpty) {*st->bind[9].is_null=true;*st->bind[9].length=0;} else {*st->bind[9].is_null=false;*st->bind[9].length=4;*(unsigned *)st->bind[9].buffer=attr.med;}
    if (attr.localpref==uEmpty) {*st->bind[10].is_null=true;*st->bind[10].length=0;} else {*st->bind[10].is_null=false;*st->bind[10].length=4;*(unsigned *)st->bind[10].buffer=attr.localpref;}
    *(unsigned *)st->bind[11].buffer=attr.aAtomicAggregate.getIPv4Addr();
    if (attr.aAggregatorAS!=0){
        *st->bind[12].is_null=false;*st->bind[12].length=4;*(unsigned *)st->bind[12].buffer=attr.aAggregatorAS;
    } else {*st->bind[12].is_null=true;*st->bind[12].length=0;}
    if (attr.aAggregatorIP.getType()==atIPv4){
        *st->bind[13].is_null=false;*st->bind[13].length=4;*(unsigned *)st->bind[13].buffer=attr.aAggregatorIP.getIPv4Addr();
    } else {*st->bind[13].is_null=true;*st->bind[13].length=0;}
    if (attr.aOriginator.getType()==atIPv4){
        *st->bind[14].is_null=false;*st->bind[14].length=4;*(unsigned *)st->bind[14].buffer=attr.aOriginator.getIPv4Addr();
    } else {*st->bind[14].is_null=true;*st->bind[14].length=0;}
    if (attr.vExtCom.size()>0){
        sV="";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        *st->bind[15].is_null=false;
        *st->bind[15].length=min<size_t>(sV.size(),st->bind[15].buffer_length);
        strncpy((char *)st->bind[15].buffer,sV.c_str(),st->bind[15].buffer_length);
    } else {
        *st->bind[15].is_null=true;
        *st->bind[15].length=0;
    }
#endif
    for(ci=nets.begin();ci!=nets.end();ci++){
#ifndef MYSQLSTMT
        sS=sQ+Sprintf(",%u,%u",ci->getIPv4Addr(),ci->getPrefixLen())+sV+",'";
        for(i=0,mi=ci->labels.begin();mi!=ci->labels.end();mi++,i++){
            if (i) sS+=" ";
            sS+=Sprintf("%u",(unsigned)*mi);
        }
        sS+="')";
        mysql_query(mc,sS.c_str());
        CHECKMYSQL;
#else
        *(unsigned *)st->bind[3].buffer=ci->getIPv4Addr();
        *(short *)st->bind[4].buffer=ci->getPrefixLen();
        sV="";
        for(i=0,mi=ci->labels.begin();mi!=ci->labels.end();mi++,i++){
            if (i) sV+=" ";
            sV+=Sprintf("%u",(unsigned)*mi);
        }
        if (sV.size()>0){
         *st->bind[16].is_null=false;
         *st->bind[16].length=min<size_t>(sV.size(),st->bind[16].buffer_length);
         strncpy((char *)st->bind[16].buffer,sV.c_str(),st->bind[16].buffer_length);
        } else {
        *st->bind[16].is_null=true;
        *st->bind[16].length=0;
        }
        st->save_buffer();
        st->execute();
        st->restore_buffer();
#endif        
    }
}
void BGPLogMy::exec_ipv6l(unsigned state,unsigned kind,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    if (!pmc) return ;
    MYSQL * mc=(MYSQL *)pmc;
    std::string sQ,sV,sS;
    unsigned i,u;
    IPTagNets::const_iterator ci;
    MPLSLabels::const_iterator mi;
#ifndef MYSQLSTMT
    sQ=Sprintf("call wr_ipv6 (%u,%u,%u",uSrcId,state,kind);
    sV="";
    sV+=Sprintf(",%u",(unsigned)attr.origin);
    if (attr.vASPath.size()>0){
        sV+=",'";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        sV+="'";
    } else sV+=",null";
    if (attr.vCom.size()>0){
        sV+=",'";
        BGPCommunities::const_iterator cmi;
        i=0;
        for(cmi=attr.vCom.begin();cmi!=attr.vCom.end();cmi++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)(*cmi)) >> 16,((unsigned)(*cmi)) & 0xffff);
        }
        sV+="'";
    } else sV+=",null";
    sV+=Sprintf(",%llu,%llu",nexthop.getIPv6AddrH(),nexthop.getIPv6AddrL());
    if (attr.med==uEmpty) sV+=",null"; else sV+=Sprintf(",%u",(unsigned)attr.med);
    if (attr.localpref==uEmpty) sV+=",null"; else sV+=Sprintf(",%u",(unsigned)attr.localpref);
    sV+=Sprintf(",%llu,%llu",attr.aAtomicAggregate.getIPv6AddrH(),attr.aAtomicAggregate.getIPv6AddrL());
    if (attr.aAggregatorAS!=0){
        sV+=Sprintf(",%u",(unsigned)attr.aAggregatorAS);
    } else sV+=",null";
    if (attr.aAggregatorIP.getType()==atIPv6){
        sV+=Sprintf(",%llu,%llu",attr.aAggregatorIP.getIPv6AddrH(),attr.aAggregatorIP.getIPv6AddrL());
    } else sV+=",null,null";
    if (attr.aOriginator.getType()==atIPv6){
        sV+=Sprintf(",%llu,%llu",attr.aOriginator.getIPv6AddrH(),attr.aOriginator.getIPv6AddrL());
    } else sV+=",null,null";
    if (attr.vExtCom.size()>0){
        sV+=",'";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        sV+="'";
    } else sV+=",null";
#else
    unsigned n;
    MyStmt *st=vSQ[3];
    for(n=0;n<st->params;n++) {*st->bind[n].length=st->bind[n].buffer_length;*st->bind[n].is_null=false;}
    *(unsigned *)st->bind[0].buffer=uSrcId;
    *(short *)st->bind[1].buffer=state;
    *(short *)st->bind[2].buffer=kind;
    *(short *)st->bind[6].buffer=attr.origin;
    n=7;
    if (attr.vASPath.size()>0){
        *st->bind[n].is_null=false;
        sV="";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        *st->bind[n].length=min<size_t>(sV.size(),st->bind[n].buffer_length);
        strncpy((char *)st->bind[n].buffer,sV.c_str(),st->bind[n].buffer_length);
    } else {
        *st->bind[n].is_null=true;
        *st->bind[n].length=0;
    }
    n=8;
    if (attr.vCom.size()>0){
        *st->bind[n].is_null=false;
        sV="";
        BGPCommunities::const_iterator cmi;
        i=0;
        for(cmi=attr.vCom.begin();cmi!=attr.vCom.end();cmi++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)(*cmi)) >> 16,((unsigned)(*cmi)) & 0xffff);
        }
        *st->bind[n].length=min<size_t>(sV.size(),st->bind[n].buffer_length);
        strncpy((char *)st->bind[n].buffer,sV.c_str(),st->bind[n].buffer_length);
    } else {
        *st->bind[n].is_null=true;
        *st->bind[n].length=0;
    }
    *(uint64_t *)st->bind[9].buffer=nexthop.getIPv6AddrH();
    *(uint64_t *)st->bind[10].buffer=nexthop.getIPv6AddrL();
    n=11;if (attr.med==uEmpty) {*st->bind[n].is_null=true;*st->bind[n].length=0;} else {*st->bind[n].is_null=false;*st->bind[n].length=4;*(unsigned *)st->bind[n].buffer=attr.med;}
    n=12;if (attr.localpref==uEmpty) {*st->bind[n].is_null=true;*st->bind[n].length=0;} else {*st->bind[n].is_null=false;*st->bind[n].length=4;*(unsigned *)st->bind[n].buffer=attr.localpref;}
    n=13;*(uint64_t *)st->bind[n].buffer=attr.aAtomicAggregate.getIPv6AddrH();*(uint64_t *)st->bind[n+1].buffer=attr.aAtomicAggregate.getIPv6AddrL();
    n=15;
    if (attr.aAggregatorAS!=0){
        *st->bind[n].is_null=false;*st->bind[n].length=4;*(unsigned *)st->bind[n].buffer=attr.aAggregatorAS;
    } else {*st->bind[n].is_null=true;*st->bind[n].length=0;}
    n=16;
    if (attr.aAggregatorIP.getType()==atIPv6){
        *st->bind[n].is_null=false;*st->bind[n].length=8;*(uint64_t *)st->bind[n].buffer=attr.aAggregatorIP.getIPv6AddrH();
        *st->bind[n+1].is_null=false;*st->bind[n+1].length=8;*(uint64_t *)st->bind[n+1].buffer=attr.aAggregatorIP.getIPv6AddrL();
    } else {*st->bind[n].is_null=true;*st->bind[n].length=0;*st->bind[n+1].is_null=true;*st->bind[n+1].length=0;}
    n=18;
    if (attr.aOriginator.getType()==atIPv6){
        *st->bind[n].is_null=false;*st->bind[n].length=8;*(uint64_t *)st->bind[n].buffer=attr.aOriginator.getIPv6AddrH();
        *st->bind[n+1].is_null=false;*st->bind[n+1].length=8;*(uint64_t *)st->bind[n+1].buffer=attr.aOriginator.getIPv6AddrL();
    } else {*st->bind[n].is_null=true;*st->bind[n].length=0;*st->bind[n+1].is_null=true;*st->bind[n+1].length=0;}
    n=20;
    if (attr.vExtCom.size()>0){
        sV="";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        *st->bind[n].is_null=false;
        *st->bind[n].length=min<size_t>(sV.size(),st->bind[n].buffer_length);
        strncpy((char *)st->bind[n].buffer,sV.c_str(),st->bind[n].buffer_length);
    } else {
        *st->bind[n].is_null=true;
        *st->bind[n].length=0;
    }
#endif
    for(ci=nets.begin();ci!=nets.end();ci++){
#ifndef MYSQLSTMT
        sS=sQ+Sprintf(",%llu,%llu,%u",ci->getIPv6AddrH(),ci->getIPv6AddrL(),ci->getPrefixLen())+sV+",'";
        //printf("%s\n",sS.c_str());
        for(i=0,mi=ci->labels.begin();mi!=ci->labels.end();mi++,i++){
            if (i) sS+=" ";
            sS+=Sprintf("%u",(unsigned)*mi);
        }
        sS+="')";
        mysql_query(mc,sS.c_str());
        CHECKMYSQL;
#else
        *(uint64_t *)st->bind[3].buffer=ci->getIPv6AddrH();
        *(uint64_t *)st->bind[4].buffer=ci->getIPv6AddrL();
        *(short *)st->bind[5].buffer=ci->getPrefixLen();
        sV="";
        for(i=0,mi=ci->labels.begin();mi!=ci->labels.end();mi++,i++){
            if (i) sV+=" ";
            sV+=Sprintf("%u",(unsigned)*mi);
        }
        n=21;
        if (sV.size()>0){
         *st->bind[n].is_null=false;
         *st->bind[n].length=min<size_t>(sV.size(),st->bind[n].buffer_length);
         strncpy((char *)st->bind[n].buffer,sV.c_str(),st->bind[n].buffer_length);
        } else {
        *st->bind[n].is_null=true;
        *st->bind[n].length=0;
        }
        st->save_buffer();
        st->execute();
        st->restore_buffer();
#endif
    }
}
void BGPLogMy::exec_vpnv4(unsigned state,unsigned kind,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    if (!pmc) return ;
    MYSQL * mc=(MYSQL *)pmc;
    std::string sQ,sV,sS;
    IPTagNets::const_iterator ci;
    MPLSLabels::const_iterator mi;
    unsigned i;
#ifndef MYSQLSTMT
    sQ=Sprintf("call wr_vpnv4 (%u,%u,%u",uSrcId,state,kind);
    sV="";
    sV+=Sprintf(",%u",(unsigned)attr.origin);
    if (attr.vASPath.size()>0){
        sV+=",'";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        sV+="'";
    } else sV+=",null";
    if (attr.vCom.size()>0){
        sV+=",'";
        BGPCommunities::const_iterator ci;
        i=0;
        for(ci=attr.vCom.begin();ci!=attr.vCom.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)*ci) >> 16,((unsigned)*ci) & 0xffff);
        }
        sV+="'";
    } else sV+=",null";
    sV+=Sprintf(",%u",(unsigned)nexthop.getRDH());
    sV+=Sprintf(",%u",(unsigned)nexthop.getRDL());
    sV+=Sprintf(",%u",(unsigned)nexthop.getIPv4Addr());
    if (attr.med==uEmpty) sV+=",null"; else sV+=Sprintf(",%u",(unsigned)attr.med);
    if (attr.localpref==uEmpty) sV+=",null"; else sV+=Sprintf(",%u",(unsigned)attr.localpref);
    sV+=Sprintf(",%u",(unsigned)attr.aAtomicAggregate.getIPv4Addr());
    if (attr.aAggregatorAS!=0){
        sV+=Sprintf(",%u",(unsigned)attr.aAggregatorAS);
    } else sV+=",null";
    if (attr.aAggregatorIP.getType()==atIPv4){
        sV+=Sprintf(",%u",(unsigned)attr.aAggregatorIP.getIPv4Addr());
    } else sV+=",null";
    if (attr.aOriginator.getType()==atIPv4){
        sV+=Sprintf(",%u",(unsigned)attr.aOriginator.getIPv4Addr());
    } else sV+=",null";
    if (attr.vExtCom.size()>0){
        sV+=",'";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        sV+="'";
    } else sV+=",null";
#else
    unsigned n;
    MyStmt *st=vSQ[2];
    for(n=0;n<st->params;n++) {*st->bind[n].length=st->bind[n].buffer_length;*st->bind[n].is_null=false;}
    *(unsigned *)st->bind[0].buffer=uSrcId;
    *(short *)st->bind[1].buffer=state;
    *(short *)st->bind[2].buffer=kind;
    *(short *)st->bind[7].buffer=attr.origin;
    if (attr.vASPath.size()>0){
        *st->bind[8].is_null=false;
        sV="";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        *st->bind[8].length=min<size_t>(sV.size(),st->bind[8].buffer_length);
        strncpy((char *)st->bind[8].buffer,sV.c_str(),st->bind[8].buffer_length);
    } else {
        *st->bind[8].is_null=true;
        *st->bind[8].length=0;
    }
    if (attr.vCom.size()>0){
        *st->bind[9].is_null=false;
        sV="";
        BGPCommunities::const_iterator cmi;
        i=0;
        for(cmi=attr.vCom.begin();cmi!=attr.vCom.end();cmi++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)(*cmi)) >> 16,((unsigned)(*cmi)) & 0xffff);
        }
        *st->bind[9].length=min<size_t>(sV.size(),st->bind[9].buffer_length);
        strncpy((char *)st->bind[9].buffer,sV.c_str(),st->bind[9].buffer_length);
    } else {
        *st->bind[9].is_null=true;
        *st->bind[9].length=0;
    }
    *(unsigned *)st->bind[10].buffer=nexthop.getRDH();
    *(unsigned *)st->bind[11].buffer=nexthop.getRDL();
    *(unsigned *)st->bind[12].buffer=nexthop.getIPv4Addr();
    if (attr.med==uEmpty) {*st->bind[13].is_null=true;*st->bind[13].length=0;} else {*st->bind[13].is_null=false;*st->bind[13].length=4;*(unsigned *)st->bind[13].buffer=attr.med;}
    if (attr.localpref==uEmpty) {*st->bind[14].is_null=true;*st->bind[14].length=0;} else {*st->bind[14].is_null=false;*st->bind[14].length=4;*(unsigned *)st->bind[14].buffer=attr.localpref;}
    *(unsigned *)st->bind[15].buffer=attr.aAtomicAggregate.getIPv4Addr();
    if (attr.aAggregatorAS!=0){
        *st->bind[16].is_null=false;*st->bind[16].length=4;*(unsigned *)st->bind[16].buffer=attr.aAggregatorAS;
    } else {*st->bind[16].is_null=true;*st->bind[16].length=0;}
    if (attr.aAggregatorIP.getType()==atIPv4){
        *st->bind[17].is_null=false;*st->bind[17].length=4;*(unsigned *)st->bind[17].buffer=attr.aAggregatorIP.getIPv4Addr();
    } else {*st->bind[17].is_null=true;*st->bind[1].length=0;}
    if (attr.aOriginator.getType()==atIPv4){
        *st->bind[18].is_null=false;*st->bind[18].length=4;*(unsigned *)st->bind[18].buffer=attr.aOriginator.getIPv4Addr();
    } else {*st->bind[18].is_null=true;*st->bind[18].length=0;}
    if (attr.vExtCom.size()>0){
        sV="";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        *st->bind[19].is_null=false;
        *st->bind[19].length=min<size_t>(sV.size(),st->bind[19].buffer_length);
        strncpy((char *)st->bind[19].buffer,sV.c_str(),st->bind[19].buffer_length);
    } else {
        *st->bind[19].is_null=true;
        *st->bind[19].length=0;
    }
#endif
    for(ci=nets.begin();ci!=nets.end();ci++){
#ifndef MYSQLSTMT
        sS=sQ+Sprintf(",%u,%u,%u,%u",ci->getRDH(),ci->getRDL(),ci->getIPv4Addr(),ci->getPrefixLen())+sV+",'";
        for(i=0,mi=ci->labels.begin();mi!=ci->labels.end();mi++,i++){
            if (i) sS+=" ";
            sS+=Sprintf("%u",(unsigned)*mi);
        }
        sS+="')";
        //printf("%s\n",sS.c_str());
        mysql_query(mc,sS.c_str());
        CHECKMYSQL;
#else
        *(unsigned *)st->bind[3].buffer=ci->getRDH();
        *(unsigned *)st->bind[4].buffer=ci->getRDL();
        *(unsigned *)st->bind[5].buffer=ci->getIPv4Addr();
        *(short *)st->bind[6].buffer=ci->getPrefixLen();
        sV="";
        for(i=0,mi=ci->labels.begin();mi!=ci->labels.end();mi++,i++){
            if (i) sV+=" ";
            sV+=Sprintf("%u",(unsigned)*mi);
        }
        if (sV.size()>0){
         *st->bind[20].is_null=false;
         *st->bind[20].length=min<size_t>(sV.size(),st->bind[20].buffer_length);
         strncpy((char *)st->bind[20].buffer,sV.c_str(),st->bind[20].buffer_length);
        } else {
        *st->bind[20].is_null=true;
        *st->bind[20].length=0;
        }
        st->save_buffer();
        st->execute();
        st->restore_buffer();
#endif
    }    
}
void BGPLogMy::exec_vpnv6(unsigned state,unsigned kind,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    if (!pmc) return ;
    MYSQL * mc=(MYSQL *)pmc;
    std::string sQ,sV,sS;
    IPTagNets::const_iterator ci;
    MPLSLabels::const_iterator mi;
    unsigned i;
#ifndef MYSQLSTMT
    sQ=Sprintf("call wr_vpnv6 (%u,%u,%u",uSrcId,state,kind);
    sV="";
    sV+=Sprintf(",%u",(unsigned)attr.origin);
    if (attr.vASPath.size()>0){
        sV+=",'";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        sV+="'";
    } else sV+=",null";
    if (attr.vCom.size()>0){
        sV+=",'";
        BGPCommunities::const_iterator ci;
        i=0;
        for(ci=attr.vCom.begin();ci!=attr.vCom.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)*ci) >> 16,((unsigned)*ci) & 0xffff);
        }
        sV+="'";
    } else sV+=",null";
    sV+=Sprintf(",%u",(unsigned)nexthop.getRDH());
    sV+=Sprintf(",%u",(unsigned)nexthop.getRDL());
    sV+=Sprintf(",%llu",nexthop.getIPv6AddrH());
    sV+=Sprintf(",%llu",nexthop.getIPv6AddrL());
    if (attr.med==uEmpty) sV+=",null"; else sV+=Sprintf(",%u",(unsigned)attr.med);
    if (attr.localpref==uEmpty) sV+=",null"; else sV+=Sprintf(",%u",(unsigned)attr.localpref);
    sV+=Sprintf(",%llu,%llu",attr.aAtomicAggregate.getIPv6AddrH(),attr.aAtomicAggregate.getIPv6AddrL());
    if (attr.aAggregatorAS!=0){
        sV+=Sprintf(",%u",(unsigned)attr.aAggregatorAS);
    } else sV+=",null";
    if (attr.aAggregatorIP.getType()==atIPv6){
        sV+=Sprintf(",%llu,%llu",attr.aAggregatorIP.getIPv6AddrH(),attr.aAggregatorIP.getIPv6AddrL());
    } else sV+=",null,null";
    if (attr.aOriginator.getType()==atIPv6){
        sV+=Sprintf(",%llu,%llu",attr.aOriginator.getIPv6AddrH(),attr.aOriginator.getIPv6AddrL());
    } else sV+=",null,null";
    if (attr.vExtCom.size()>0){
        sV+=",'";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        sV+="'";
    } else sV+=",null";
#else
    unsigned n;
    MyStmt *st=vSQ[4];
    for(n=0;n<st->params;n++) {*st->bind[n].length=st->bind[n].buffer_length;*st->bind[n].is_null=false;}
    *(unsigned *)st->bind[0].buffer=uSrcId;
    *(short *)st->bind[1].buffer=state;
    *(short *)st->bind[2].buffer=kind;
    *(short *)st->bind[8].buffer=attr.origin;
    n=9;
    if (attr.vASPath.size()>0){
        *st->bind[n].is_null=false;
        sV="";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        *st->bind[n].length=min<size_t>(sV.size(),st->bind[n].buffer_length);
        strncpy((char *)st->bind[n].buffer,sV.c_str(),st->bind[n].buffer_length);
    } else {
        *st->bind[n].is_null=true;
        *st->bind[n].length=0;
    }
    n=10;
    if (attr.vCom.size()>0){
        *st->bind[n].is_null=false;
        sV="";
        BGPCommunities::const_iterator cmi;
        i=0;
        for(cmi=attr.vCom.begin();cmi!=attr.vCom.end();cmi++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)(*cmi)) >> 16,((unsigned)(*cmi)) & 0xffff);
        }
        *st->bind[n].length=min<size_t>(sV.size(),st->bind[n].buffer_length);
        strncpy((char *)st->bind[n].buffer,sV.c_str(),st->bind[n].buffer_length);
    } else {
        *st->bind[n].is_null=true;
        *st->bind[n].length=0;
    }
    *(unsigned *)st->bind[11].buffer=nexthop.getRDH();
    *(unsigned *)st->bind[12].buffer=nexthop.getRDL();
    *(uint64_t *)st->bind[13].buffer=nexthop.getIPv6AddrH();
    *(uint64_t *)st->bind[14].buffer=nexthop.getIPv6AddrL();
    n=15;if (attr.med==uEmpty) {*st->bind[n].is_null=true;*st->bind[n].length=0;} else {*st->bind[n].is_null=false;*st->bind[n].length=4;*(unsigned *)st->bind[n].buffer=attr.med;}
    n=16;if (attr.localpref==uEmpty) {*st->bind[n].is_null=true;*st->bind[n].length=0;} else {*st->bind[n].is_null=false;*st->bind[n].length=4;*(unsigned *)st->bind[n].buffer=attr.localpref;}
    n=17;*(uint64_t *)st->bind[n].buffer=attr.aAtomicAggregate.getIPv6AddrH();*(uint64_t *)st->bind[n+1].buffer=attr.aAtomicAggregate.getIPv6AddrL();
    n=19;
    if (attr.aAggregatorAS!=0){
        *st->bind[n].is_null=false;*st->bind[n].length=4;*(unsigned *)st->bind[n].buffer=attr.aAggregatorAS;
    } else {*st->bind[n].is_null=true;*st->bind[n].length=0;}
    n=20;
    if (attr.aAggregatorIP.getType()==atIPv6){
        *st->bind[n].is_null=false;*st->bind[n].length=8;*(uint64_t *)st->bind[n].buffer=attr.aAggregatorIP.getIPv6AddrH();
        *st->bind[n+1].is_null=false;*st->bind[n+1].length=8;*(uint64_t *)st->bind[n+1].buffer=attr.aAggregatorIP.getIPv6AddrL();
    } else {*st->bind[n].is_null=true;*st->bind[n].length=0;*st->bind[n+1].is_null=true;*st->bind[n+1].length=0;}
    n=22;
    if (attr.aOriginator.getType()==atIPv6){
        *st->bind[n].is_null=false;*st->bind[n].length=8;*(uint64_t *)st->bind[n].buffer=attr.aOriginator.getIPv6AddrH();
        *st->bind[n+1].is_null=false;*st->bind[n+1].length=8;*(uint64_t *)st->bind[n+1].buffer=attr.aOriginator.getIPv6AddrL();
    } else {*st->bind[n].is_null=true;*st->bind[n].length=0;*st->bind[n+1].is_null=true;*st->bind[n+1].length=0;}
    n=24;
    if (attr.vExtCom.size()>0){
        sV="";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        *st->bind[n].is_null=false;
        *st->bind[n].length=min<size_t>(sV.size(),st->bind[n].buffer_length);
        strncpy((char *)st->bind[n].buffer,sV.c_str(),st->bind[n].buffer_length);
    } else {
        *st->bind[n].is_null=true;
        *st->bind[n].length=0;
    }
#endif        
    for(ci=nets.begin();ci!=nets.end();ci++){
#ifndef MYSQLSTMT
        sS=sQ+Sprintf(",%u,%u,%llu,%llu,%u",ci->getRDH(),ci->getRDL(),ci->getIPv6AddrH(),ci->getIPv6AddrL(),ci->getPrefixLen())+sV+",'";
        for(i=0,mi=ci->labels.begin();mi!=ci->labels.end();mi++,i++){
            if (i) sS+=" ";
            sS+=Sprintf("%u",(unsigned)*mi);
        }
        sS+="')";
        //fprintf(stderr,"%s\n",sS.c_str());
        mysql_query(mc,sS.c_str());
        CHECKMYSQL;
#else
        *(unsigned *)st->bind[3].buffer=ci->getRDH();
        *(unsigned *)st->bind[4].buffer=ci->getRDL();
        *(uint64_t *)st->bind[5].buffer=ci->getIPv6AddrH();
        *(uint64_t *)st->bind[6].buffer=ci->getIPv6AddrL();
        *(short *)st->bind[7].buffer=ci->getPrefixLen();
        sV="";
        for(i=0,mi=ci->labels.begin();mi!=ci->labels.end();mi++,i++){
            if (i) sV+=" ";
            sV+=Sprintf("%u",(unsigned)*mi);
        }
        n=25;
        if (sV.size()>0){
         *st->bind[n].is_null=false;
         *st->bind[n].length=min<size_t>(sV.size(),st->bind[n].buffer_length);
         strncpy((char *)st->bind[n].buffer,sV.c_str(),st->bind[n].buffer_length);
        } else {
        *st->bind[n].is_null=true;
        *st->bind[n].length=0;
        }
        st->save_buffer();
        st->execute();
        st->restore_buffer();
#endif
    }    
}
void BGPLogMy::exec_l2vpn(unsigned state,unsigned kind,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    if (!pmc) return ;
    MYSQL * mc=(MYSQL *)pmc;
    std::string sQ,sV,sS;
    IPTagNets::const_iterator ci;
    MPLSLabels::const_iterator mi;
    unsigned i;
#ifndef MYSQLSTMT
    sQ=Sprintf("call wr_l2vpn (%u,%u,%u",uSrcId,state,kind);
    sV="";
    sV+=Sprintf(",%u",(unsigned)attr.origin);
    if (attr.vASPath.size()>0){
        sV+=",'";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        sV+="'";
    } else sV+=",null";
    if (attr.vCom.size()>0){
        sV+=",'";
        BGPCommunities::const_iterator ci;
        i=0;
        for(ci=attr.vCom.begin();ci!=attr.vCom.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)*ci) >> 16,((unsigned)*ci) & 0xffff);
        }
        sV+="'";
    } else sV+=",null";
    sV+=Sprintf(",%u",(unsigned)nexthop.getIPv4Addr());
    if (attr.med==uEmpty) sV+=",null"; else sV+=Sprintf(",%u",(unsigned)attr.med);
    if (attr.localpref==uEmpty) sV+=",null"; else sV+=Sprintf(",%u",(unsigned)attr.localpref);
    sV+=Sprintf(",%u",(unsigned)attr.aAtomicAggregate.getIPv4Addr());
    if (attr.aAggregatorAS!=0){
        sV+=Sprintf(",%u",(unsigned)attr.aAggregatorAS);
    } else sV+=",null";
    if (attr.aAggregatorIP.getType()==atIPv4){
        sV+=Sprintf(",%u",(unsigned)attr.aAggregatorIP.getIPv4Addr());
    } else sV+=",null";
    if (attr.aOriginator.getType()==atIPv4){
        sV+=Sprintf(",%u",(unsigned)attr.aOriginator.getIPv4Addr());
    } else sV+=",null";
    if (attr.vExtCom.size()>0){
        sV+=",'";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        sV+="'";
    } else sV+=",null";
#else
    unsigned n;
    MyStmt *st=vSQ[5];
    for(n=0;n<st->params;n++) {*st->bind[n].length=st->bind[n].buffer_length;*st->bind[n].is_null=false;}
    *(unsigned *)st->bind[0].buffer=uSrcId;
    *(short *)st->bind[1].buffer=state;
    *(short *)st->bind[2].buffer=kind;
    *(short *)st->bind[8].buffer=attr.origin;
    n=9;
    if (attr.vASPath.size()>0){
        *st->bind[n].is_null=false;
        sV="";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        *st->bind[n].length=min<size_t>(sV.size(),st->bind[n].buffer_length);
        strncpy((char *)st->bind[n].buffer,sV.c_str(),st->bind[n].buffer_length);
    } else {
        *st->bind[n].is_null=true;
        *st->bind[n].length=0;
    }
    n=10;
    if (attr.vCom.size()>0){
        *st->bind[n].is_null=false;
        sV="";
        BGPCommunities::const_iterator cmi;
        i=0;
        for(cmi=attr.vCom.begin();cmi!=attr.vCom.end();cmi++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)(*cmi)) >> 16,((unsigned)(*cmi)) & 0xffff);
        }
        *st->bind[n].length=min<size_t>(sV.size(),st->bind[n].buffer_length);
        strncpy((char *)st->bind[n].buffer,sV.c_str(),st->bind[n].buffer_length);
    } else {
        *st->bind[n].is_null=true;
        *st->bind[n].length=0;
    }
    *(unsigned *)st->bind[11].buffer=nexthop.getIPv4Addr();
    n=12;if (attr.med==uEmpty) {*st->bind[n].is_null=true;*st->bind[n].length=0;} else {*st->bind[n].is_null=false;*st->bind[n].length=4;*(unsigned *)st->bind[n].buffer=attr.med;}
    n=13;if (attr.localpref==uEmpty) {*st->bind[n].is_null=true;*st->bind[n].length=0;} else {*st->bind[n].is_null=false;*st->bind[n].length=4;*(unsigned *)st->bind[n].buffer=attr.localpref;}
    n=14;*(unsigned *)st->bind[n].buffer=attr.aAtomicAggregate.getIPv4Addr();
    n=15;
    if (attr.aAggregatorAS!=0){
        *st->bind[n].is_null=false;*st->bind[n].length=4;*(unsigned *)st->bind[n].buffer=attr.aAggregatorAS;
    } else {*st->bind[n].is_null=true;*st->bind[n].length=0;}
    n=16;
    if (attr.aAggregatorIP.getType()==atIPv4){
        *st->bind[n].is_null=false;*st->bind[n].length=4;*(unsigned *)st->bind[n].buffer=attr.aAggregatorIP.getIPv4Addr();
    } else {*st->bind[n].is_null=true;*st->bind[n].length=0;}
    n=17;
    if (attr.aOriginator.getType()==atIPv4){
        *st->bind[n].is_null=false;*st->bind[n].length=4;*(unsigned *)st->bind[n].buffer=attr.aOriginator.getIPv4Addr();
    } else {*st->bind[n].is_null=true;*st->bind[n].length=0;}
    n=18;
    if (attr.vExtCom.size()>0){
        sV="";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        *st->bind[n].is_null=false;
        *st->bind[n].length=min<size_t>(sV.size(),st->bind[n].buffer_length);
        strncpy((char *)st->bind[n].buffer,sV.c_str(),st->bind[n].buffer_length);
    } else {
        *st->bind[n].is_null=true;
        *st->bind[n].length=0;
    }
#endif
    for(ci=nets.begin();ci!=nets.end();ci++){
#ifndef MYSQLSTMT
        sS=sQ+Sprintf(",%u,%u,%u,%u,%u",ci->getRDH(),ci->getRDL(),ci->getL2Site(),ci->getL2Offset(),ci->getL2Range())+sV+",'";
        for(i=0,mi=ci->labels.begin();mi!=ci->labels.end();mi++,i++){
            if (i) sS+=" ";
            sS+=Sprintf("%u",(unsigned)*mi);
        }
        sS+="')";
        mysql_query(mc,sS.c_str());
        CHECKMYSQL;
#else
        *(unsigned *)st->bind[3].buffer=ci->getRDH();
        *(unsigned *)st->bind[4].buffer=ci->getRDL();
        *(unsigned short *)st->bind[5].buffer=ci->getL2Site();
        *(unsigned short *)st->bind[6].buffer=ci->getL2Offset();
        *(unsigned short *)st->bind[7].buffer=ci->getL2Range();
        sV="";
        for(i=0,mi=ci->labels.begin();mi!=ci->labels.end();mi++,i++){
            if (i) sV+=" ";
            sV+=Sprintf("%u",(unsigned)*mi);
        }
        n=19;
        if (sV.size()>0){
         *st->bind[n].is_null=false;
         *st->bind[n].length=min<size_t>(sV.size(),st->bind[n].buffer_length);
         strncpy((char *)st->bind[n].buffer,sV.c_str(),st->bind[n].buffer_length);
        } else {
         *st->bind[n].is_null=true;
         *st->bind[n].length=0;
        }
        st->save_buffer();
        st->execute();
        st->restore_buffer();
#endif
    }

}
void BGPLogMy::nlri(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPNets & nets){
    exec_ipv4(1,1,attr,attr.aNextHop,nets);
}
void BGPLogMy::withdraw(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPNets & nets){
    exec_ipv4(2,1,attr,attr.aNextHop,nets);
}
void BGPLogMy::nlri_ipv4m(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPNets & nets){
	exec_ipv4(1,2,attr,nexthop,nets);
}
void BGPLogMy::withdraw_ipv4m(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPNets & nets){
	exec_ipv4(2,2,attr,nexthop,nets);
}
void BGPLogMy::nlri_ipv4lu(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
	exec_ipv4l(1,4,attr,nexthop,nets);
}
void BGPLogMy::withdraw_ipv4lu(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
	exec_ipv4l(2,4,attr,nexthop,nets);
}
void BGPLogMy::nlri_vpnv4u(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    exec_vpnv4(1,128,attr,nexthop,nets);
}
void BGPLogMy::withdraw_vpnv4u(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    exec_vpnv4(2,128,attr,nexthop,nets);
}
void BGPLogMy::nlri_ipv6u(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPNets & nets){
	exec_ipv6(1,4,attr,nexthop,nets);
}
void BGPLogMy::withdraw_ipv6u(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPNets & nets){
	exec_ipv6(2,4,attr,nexthop,nets);
}
void BGPLogMy::nlri_ipv6lu(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
	exec_ipv6l(1,4,attr,nexthop,nets);
}
void BGPLogMy::withdraw_ipv6lu(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
	exec_ipv6l(2,4,attr,nexthop,nets);
}
void BGPLogMy::nlri_vpnv6u(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    exec_vpnv6(1,128,attr,nexthop,nets);
}
void BGPLogMy::withdraw_vpnv6u(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    exec_vpnv6(2,128,attr,nexthop,nets);
}
void BGPLogMy::nlri_l2vpn(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    exec_l2vpn(1,1,attr,nexthop,nets);
}
void BGPLogMy::withdraw_l2vpn(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    exec_l2vpn(2,1,attr,nexthop,nets);
}
