#include "bgplogoci.h"
#include "bgpcommon.h"
#include <unistd.h>
#include <string.h>
#include <iostream>
#include <sstream>
extern "C"{
#include <oci.h>
}
bool bOCIWasInited=false;
#define MAX_SQLMSG 2048
using namespace std;
namespace BGPHist{

class OCIBuffer;
class OCIDesc;
class OCIQuery {
friend class OCIBuffer;
friend class OCIDesc;
friend class BGPLogOCI;
  OCIQuery(const OCIQuery &){}
 protected:
  int recN,recC;
  bool bWantsRO,bEof,bBof,bAutoCommit,bOpened,bCancelled,bhasDataGot;
  OCIStmt * stmt;
  OCIError *err;
  std::string sSql;
  OCIBuffer *cBuffer,*cBind;
  bool bIsScrollable;
 public:
  BGPLogOCI *sess;
  OCIQuery(BGPLogOCI *);
  virtual ~OCIQuery();
  bool isOpened() const;
  void cursorOpen();
  void cursorParse();
  void cursorExecute(unsigned int iter);
  void cursorClose();
  void checkStatus(sword status=0);
  void cursorBind();
  void cursorDescribe();
  void cursorDefine();
  
  bool isEof();
  void next();
  void open();
  int execute();
  virtual void setSQL(const char *);
  bool prepareExecute();
  long executePrepared();
  bool unprepareExecute();
};

struct _OCIDate
{
        ub1 oraCentury;
    ub1 oraYear;
    ub1 oraMonth;
    ub1 oraDay;
    ub1 oraHour;
    ub1 oraMinute;
    ub1 oraSecond;
};

void OCI_atexit(void){
        OCITerminate(OCI_DEFAULT);
}
dvoid * ocimalocfp(dvoid *ctxp,size_t size){
    return malloc(size);
}
dvoid * ociralocfp(dvoid *ctxp,dvoid *memptr, size_t newsize){
    return realloc(memptr,newsize);
}
void ocimfreefp(dvoid *ctxp,dvoid *memptr){
    free(memptr);
}

class OCIException : public std::exception {
public:
    unsigned int uErrNo;
    std::string sMsg;
    OCIException() throw();
    OCIException(unsigned int ,const std::string &) throw();
    virtual ~OCIException() throw();
    virtual const char* what() const throw();
};
OCIException::OCIException() throw(){
}
OCIException::OCIException(unsigned int c,const std::string &s) throw(){
    try {
        uErrNo=c;
        sMsg=s;
    } catch(...){}
}
OCIException::~OCIException() throw(){
}
const char* OCIException::what() const throw(){
    return sMsg.c_str();
}

//===============================================================================================================================================
struct OCIFldBuf {
                sb2     ind;
                ub2     rlen;
                ub2 rc;
#ifdef OCIOBJ
                ub4 objlen,objindsz;
                dvoid * objptr,*objind;
#endif
                OCILobLocator *loc;

                char data[1];
};

class OCIDesc {
 public:
  OCIDesc();
  OCIDesc(const char *_nm,sb2 _Type,ub2 _Size);
  char szName[256];
  ub1 NameLen;
#ifdef OCIOBJ
  char szTypeName[256],szTypeSchema[256],szTypeSub[256];
  ub1 TypeNameLen,TypeSchemaLen,TypeSubLen;
  OCITypeDesc *ociType;
#endif
  ub2 ColSize;
  sb2 ColType;
  sb4 DispWidth;
  ub1 ColPrec;
  sb1 ColScale,NullOk;
  sword descr(OCIQuery * ,int );
  sword define(OCIQuery *,int ,OCIFldBuf * );
  sword bind(OCIQuery *,OCIFldBuf * );
};
OCIDesc::OCIDesc(){
        NameLen=szName[0]=0;
#ifdef OCIOBJ
        TypeNameLen=TypeSchemaLen=TypeSubLen=szTypeSub[0]=szTypeSchema[0]=szTypeName[0]=0;
#endif
}
OCIDesc::OCIDesc(const char *_nm,sb2 _Type,ub2 _Size){
        NameLen=strlen(_nm);if(NameLen>=sizeof(szName)) NameLen=sizeof(szName)-1;
        strncpy(szName,_nm,NameLen);
        ColType=_Type;
        ColSize=_Size;
#ifdef OCIOBJ
        TypeNameLen=TypeSchemaLen=TypeSubLen=szTypeSub[0]=szTypeSchema[0]=szTypeName[0]=0;
#endif
}

sword _getstrattr(OCIError * err,OCIParam *prm,ub4 attrid,char * szName,ub1 & len){
        char *nm;
        ub4 lng;
        lng=255;
        sword status=OCIAttrGet((dvoid*) prm, (ub4) OCI_DTYPE_PARAM, (dvoid*) &nm,(ub4 *) &lng, (ub4) attrid, err );
        if (status) {szName[0]=0;len=0;return status;}
        strncpy(szName,nm,lng);
        szName[lng]=0;
        len=lng;
        return status;
}
sword OCIDesc::descr(OCIQuery *q,int i){
        sword status;
        NameLen=sizeof(szName)-1;
        OCIParam *prm=0;
        status=OCIParamGet(q->stmt, OCI_HTYPE_STMT, q->err, (dvoid **)&prm,i);
        if (status) return status;
        status=OCIAttrGet((dvoid*) prm, (ub4) OCI_DTYPE_PARAM,(dvoid*) &ColType,(ub4 *) 0, (ub4) OCI_ATTR_DATA_TYPE,q->err  );
        if (status) return status;

        status=_getstrattr(q->err,prm,OCI_ATTR_NAME,szName,NameLen);
        if (status) return status;
#ifdef OCIOBJ
        status=_getstrattr(q->err,prm,OCI_ATTR_TYPE_NAME,szTypeName,TypeNameLen);
        status=_getstrattr(q->err,prm,OCI_ATTR_SCHEMA_NAME,szTypeSchema,TypeSchemaLen);
        status=_getstrattr(q->err,prm,OCI_ATTR_SUB_NAME,szTypeSub,TypeSubLen);
#endif
        status=OCIAttrGet((dvoid*) prm, (ub4) OCI_DTYPE_PARAM, (dvoid*) &ColSize,(ub4 *) 0, (ub4) OCI_ATTR_DATA_SIZE, q->err );
        if (status) return status;
        status=OCIAttrGet((dvoid*) prm, (ub4) OCI_DTYPE_PARAM, (dvoid*) &ColPrec,(ub4 *) 0, (ub4) OCI_ATTR_PRECISION, q->err );
        status=OCIAttrGet((dvoid*) prm, (ub4) OCI_DTYPE_PARAM, (dvoid*) &ColScale,(ub4 *) 0, (ub4) OCI_ATTR_SCALE, q->err );
        DispWidth=0;
        status=OCIAttrGet((dvoid*) prm, (ub4) OCI_DTYPE_PARAM, (dvoid*) &DispWidth,(ub4 *) 0, (ub4) OCI_ATTR_DISP_SIZE, q->err );
        NullOk=1;
        return status;
}
sword OCIDesc::define(OCIQuery *q,int i,OCIFldBuf *buf){
        sword status=0;
        OCIDefine *defn;
        if (buf) buf->loc=0;
        if ((ColType==SQLT_BLOB) || (ColType==SQLT_CLOB)){
                if (buf)status=OCIDescriptorAlloc(q->sess->env,(dvoid **) &buf->loc,(ub4) OCI_DTYPE_LOB, (size_t) 0, (dvoid **) 0);
                if (status) return status;
                status=OCIDefineByPos (q->stmt, &defn, q->err, i,(buf ? &buf->loc:0),(buf ? ColSize:0),ColType,(buf ? &buf->ind:0),0,0,OCI_DEFAULT);
        } else {
                status=OCIDefineByPos (q->stmt, &defn, q->err, i,(buf ? buf->data:0),(buf ? ColSize:0),ColType,(buf ? &buf->ind:0),(buf ? &buf->rlen:0),(buf ? &buf->rc:0),OCI_DEFAULT);
                if (status) return status;
#ifdef OCIOBJ
                if ((ColType==SQLT_NTY) || (ColType==SQLT_REF)){
                        if (buf){
                                buf->objptr=0;//&buf->data;
                                buf->objlen=0;//ColSize;
                                buf->objindsz=4;
                        }
                        ociType=&q->sess->getOCIType(OCIObjName(szTypeSchema,szTypeName,szTypeSub));
                        status=OCIDefineObject(defn, q->err,*ociType,(buf ? &buf->objptr:0),(buf ? &buf->objlen:0),(buf ? &buf->objind:0),(buf ? &buf->objindsz:0));
                }
#endif
        }
        return status;
}

sword OCIDesc::bind(OCIQuery *q,OCIFldBuf * buf){
        sword status;
        OCIBind *bnd=0;
        status=OCIBindByName ( q->stmt, &bnd,q->err,(text *)szName,NameLen,buf ? (dvoid*)buf->data:0,buf ? ColSize:0,ColType,buf ? &buf->ind:0,buf ? &buf->rlen:0,buf ? &buf->rc:0,0,0,OCI_DEFAULT);
        if (status) return status;
#ifdef OCIOBJ
                if ((ColType==SQLT_NTY) || (ColType==SQLT_REF)){
                        status=OCIBindObject(bnd, q->err,*ociType,(buf ? &buf->objptr:0),(buf ? &buf->objlen:0),(buf ? &buf->objind:0),(buf ? &buf->objindsz:0));
                }
#endif
        return status;
}
//===============================================================================================================================================
class OCIBuffer: public std::vector<OCIDesc *> {
  typedef std::vector<OCIDesc *> inherited;
 public:
  OCIQuery *own;
  char* cRowBuffer;size_t szRowBuffer;

  OCIBuffer(OCIQuery *);
  ~OCIBuffer();

  void describe();
  void define();
  void undefine();
  void bind();
  void unbind();
  void clearList();
  void allocBuffer();
  void freeBuffer();
  OCIFldBuf* getField(unsigned int idx);

};
OCIBuffer::OCIBuffer(OCIQuery *q):cRowBuffer(0),own(q){}

void OCIBuffer::clearList(){
        freeBuffer();
        for(iterator i=begin();i!=end();i++) delete *i;
        clear();
}
void OCIBuffer::allocBuffer(){
        freeBuffer();
        szRowBuffer=0;
        for(iterator i=begin();i!=end();i++){
                szRowBuffer+=sizeof(OCIFldBuf)+(*i)->ColSize-1;
        }
        if (!szRowBuffer) return ;
        cRowBuffer=(char *)malloc(szRowBuffer);
        memset(cRowBuffer,0,szRowBuffer);
}
void OCIBuffer::freeBuffer(){
        iterator i;
        if (cRowBuffer) {
                char* lpB=cRowBuffer;
                for(i=begin();i!=end();i++) {
                        OCIFldBuf *fld=(OCIFldBuf *)lpB;
                        if (fld->loc)
                                OCIDescriptorFree(((OCIFldBuf *)lpB)->loc,OCI_DTYPE_LOB);
#ifdef OCIOBJ
                        if (fld->objptr)
                                OCIObjectFree(own->sess->env,own->err,fld->objptr,0);
                        fld->objptr=0;
                        fld->objind=0;
#endif
                        lpB+=sizeof(OCIFldBuf)+(*i)->ColSize-1;
                }
                free(cRowBuffer);
        }
        cRowBuffer=0;
}
OCIBuffer::~OCIBuffer(){
        clearList();
}

void OCIBuffer::describe(){
        sword i,status;
        clearList();
        for(i=1;;i++){
                OCIDesc* od=new OCIDesc();
                status=od->descr(own,i);
                if (status) {delete od;break;}
                push_back(od);
        }
}

void OCIBuffer::undefine(){
        int j;
        iterator i;
        for(i=begin(),j=1;i!=end();i++,j++){
                OCIDesc & od=*(*i);
                own->checkStatus(od.define(own,j,0));
        }
}

void OCIBuffer::define(){
        szRowBuffer=0;
        char *ptr;
        allocBuffer();
        if (!szRowBuffer) return ;
        ptr=cRowBuffer;
        int j;
        iterator i;
        for(i=begin(),j=1;i!=end();i++,j++){
                OCIDesc & od=*(*i);
                own->checkStatus(od.define(own,j,(OCIFldBuf *)ptr));
                ptr+=sizeof(OCIFldBuf)+od.ColSize-1;
        }
}

void OCIBuffer::bind(){
        if (!szRowBuffer) return ;
        char * ptr=cRowBuffer;
        int j;
        iterator i;
        for(i=begin(),j=1;i!=end();i++,j++){
                OCIDesc & od=*(*i);
                own->checkStatus(od.bind(own,(OCIFldBuf *)ptr));
                ptr+=sizeof(OCIFldBuf)+od.ColSize-1;
        }
}

void OCIBuffer::unbind(){
        int j;
        iterator i;
        for(i=begin(),j=1;i!=end();i++,j++){
                OCIDesc & od=*(*i);
                own->checkStatus(od.bind(own,0));
        }
}
OCIFldBuf* OCIBuffer::getField(unsigned int idx){
        if (!cRowBuffer) return 0;
        if (!idx) return (OCIFldBuf *)cRowBuffer;
        char * ptr=cRowBuffer;
        size_t i;
        for(i=0;i<size();i++){
                OCIDesc & od=*(operator[](i));
                ptr+=sizeof(OCIFldBuf)+od.ColSize-1;
                if ((i+1)==idx) return (OCIFldBuf *)ptr;
        }
        return 0;
}

//===============================================================================================================================================
OCIQuery::OCIQuery(BGPLogOCI *_s):sess(_s),recN(0),recC(0),bWantsRO(true),bEof(false),bBof(false),bAutoCommit(false),bOpened(false),bCancelled(false),bhasDataGot(false),
  stmt(0),err(0){
  cBuffer=new OCIBuffer(this);
  cBind=new OCIBuffer(this);
}
OCIQuery::~OCIQuery(){
    delete cBind;
    delete cBuffer;
    cursorClose();
}
void OCIQuery::setSQL(const char *s){sSql=s;}
void OCIQuery::open(){
 cursorOpen();
 recN=0;recC=0;
 try {
  cursorParse();
  cursorBind();
  //cursorBind();
  cursorExecute(0);
  cursorDescribe();
  cursorDefine();
 } catch(...){
  try {cursorClose();} catch(...){}
  throw;
 }
 bOpened=true;
 bBof=true;
 bEof=false;
}

void OCIQuery::cursorOpen(){
 if (bOpened) return ;
 if (!sess) return ;
 bhasDataGot=false;
 sword state;
 state=OCIHandleAlloc( sess->env, (dvoid **) &stmt, (ub4) OCI_HTYPE_STMT, 0, 0);
 checkStatus( state);
 state=OCIHandleAlloc( sess->env, (dvoid **) &err, (ub4) OCI_HTYPE_ERROR, 0, 0);
 checkStatus(state);
}
void OCIQuery::cursorParse(){
 sword state;
 state=OCIStmtPrepare(stmt,err,(text *)sSql.c_str(), sSql.length(),OCI_NTV_SYNTAX,OCI_DEFAULT);
 checkStatus( state);
}
void OCIQuery::cursorExecute(unsigned int iter){
 ub4 mode;
 sword state;
 mode=0;
 mode|=(bAutoCommit ? OCI_COMMIT_ON_SUCCESS:0);
 state=OCIStmtExecute(sess->srv,stmt,err,iter,0,0,0,mode);
 checkStatus(state);
 /*
 long l=10;//prefetch rows
 if (l)
  checkStatus(state=OCIAttrSet(stmt,OCI_HTYPE_STMT,&l,0,OCI_ATTR_PREFETCH_ROWS,err));
 */
 checkStatus(state=OCIAttrGet(stmt, OCI_HTYPE_STMT,&recC, 0, (ub4) OCI_ATTR_ROW_COUNT,err));
}
void OCIQuery::cursorClose(){
 sword state;
 if (stmt){
  state=OCIHandleFree(stmt,OCI_HTYPE_STMT);stmt=0;
 }
 if (err){
  OCIHandleFree(err,OCI_HTYPE_ERROR);err=0;
 }
}
void OCIQuery::cursorBind(){
 if (cBind->empty()) return ;
 cBind->bind();
}

void OCIQuery::cursorDescribe(){
 cBuffer->describe();
 //cBuffer->setTo(fields);
}

void OCIQuery::cursorDefine(){
 cBuffer->define();
}
bool OCIQuery::isOpened() const{return bOpened;}
bool OCIQuery::isEof(){return bEof;}
void OCIQuery::next(){
 if (bEof) return ;
 sword stat;
 stat=OCIStmtFetch(stmt,err,1,OCI_FETCH_NEXT,OCI_DEFAULT);
 if (!stat || (stat==1)) recN++;
 if (!stat || (stat==1)) bhasDataGot=false;
 if (stat==OCI_NO_DATA) bEof=true; else checkStatus(stat);
 bBof=false;
}
void OCIQuery::checkStatus(sword status){
 if(!sess) return;
 sess->checkStatus(status,err);
}
int OCIQuery::execute(){
 int ret;
 cursorOpen();
 try {
  cursorParse();
  cursorBind();
  cursorExecute(1);
  //cursorDescribe();
  //cursorDefine();
  ret=0;//cda.rpc;
  sword status=OCIAttrGet(stmt, OCI_HTYPE_STMT,&ret, 0, (ub4) OCI_ATTR_ROW_COUNT,err  );
 } catch(...) {
 try {cursorClose();} catch(...){}
 throw;
        }
        cursorClose();
        return ret;
}

//===============================================================================================================================================
void BGPLogOCI::checkStatus(int status,OCIError *_err){
        if (!status || (status==1)) return ;
        if(!_err) _err=err;
        sb4 errcode;
        char errMsg[MAX_SQLMSG];
        switch (status){
                case OCI_SUCCESS: return ;
                case OCI_SUCCESS_WITH_INFO:return ;
                case OCI_NEED_DATA:strcpy(errMsg,"Error - OCI_NEED_DATA");break;
                case OCI_NO_DATA:strcpy(errMsg,"Error - OCI_NO_DATA");break;
                case OCI_INVALID_HANDLE:strcpy(errMsg,"Error - OCI_INVALID_HANDLE");break;
                case OCI_STILL_EXECUTING:strcpy(errMsg,"Error - OCI_STILL_EXECUTE");break;
                case OCI_CONTINUE:strcpy(errMsg,"Error - OCI_CONTINUE");break;
                //case OCI_ERROR:
                default:{
                        OCIErrorGet(_err,1,0,&errcode,(text *)errMsg,MAX_SQLMSG,OCI_HTYPE_ERROR);
                        break;
                }
        }
        throw OCIException(errcode,errMsg);
}
BGPLogOCI::BGPLogOCI(const char *_ocicon,const char *_dbuser,const char *_dbpwd,const char *_peerip,unsigned int _asn):pmc(0),uSrcId(0){
    sword status;
    OCIThreadProcessInit();
    OCIEnv *g_env;
    int i;
    if (!bOCIWasInited) {
     bOCIWasInited=true;
     status=OCIInitialize((ub4) OCI_THREADED, (dvoid *)0,  ocimalocfp,  ociralocfp,  ocimfreefp );
     status=OCIEnvInit( &g_env, (ub4) OCI_DEFAULT, 0, 0 );
     atexit(OCI_atexit);
    }
    env=g_env;
    status=OCIHandleAlloc( env, (dvoid **) &err, (ub4) OCI_HTYPE_ERROR, 0, 0);
    checkStatus(status);
    status=OCILogon(env,err,&srv,(text *)_dbuser, strlen(_dbuser),
                                        (text *)_dbpwd, strlen(_dbpwd),
                                        (text *)_ocicon, strlen(_ocicon));
    checkStatus(status);
    ub4 sz;sz=4;
    status=OCIAttrGet(srv,OCI_HTYPE_SVCCTX,(dvoid *)&srvr,&sz,OCI_ATTR_SERVER,err);
    checkStatus(status);
    uSrcId=0;
    try{
     std::string sQ;
     OCIQuery q(this);
     q.setSQL("call bgplog_getsrcif(:id,:peerip,:peeras)");
     q.cursorOpen();
     q.cursorParse();
     q.cBind->push_back(new OCIDesc(":id",SQLT_INT,4));
     q.cBind->push_back(new OCIDesc(":peerip",1,40));
     q.cBind->push_back(new OCIDesc(":peeras",SQLT_INT,4));
     q.cBind->allocBuffer();
     OCIFldBuf *fb=q.cBind->getField(1);
     fb->rlen=min<size_t>(40,strlen(_peerip));
     strncpy(fb->data,_peerip,fb->rlen);
     fb=q.cBind->getField(2);
     fb->rlen=min<size_t>(4,sizeof(_asn));
     memcpy(fb->data,&_asn,fb->rlen);
     q.cursorBind();
     q.cursorExecute(1);
     fb=q.cBind->getField(0);
     memcpy(&uSrcId,fb->data,min<size_t>(fb->rlen,sizeof(uSrcId)));
     q.cursorClose();
     commit();
     q.cBind->clearList();
     const char *szTbls[]={"bgp_ipv4","bgp_vpnv4","bgp_ipv6","bgp_vpnv6",0};
     for(i=0;szTbls[i];i++){
        sQ=Sprintf("update %s set state=2 where srcid=%u",szTbls[i],uSrcId);
        q.setSQL(sQ.c_str());
        try {
            q.execute();
        } catch(const std::exception & e){
         fprintf(stderr,"Exception caught while initializing BGPLogOCI: '%s'\n",e.what());
        } catch(...){
         fprintf(stderr,"Exception caught while initializing BGPLogOCI\n");
        }
        commit();
     }
     OCIQuery *pq;
     vSQ[1]=pq=new OCIQuery(this);
     pq->setSQL("insert into bgp_commonlog (msgcode,msgtxt,srcid) values (:msgcode,:msgtxt,:srcid)");
     pq->cursorOpen();
     pq->cursorParse();
     pq->cBind->push_back(new OCIDesc(":msgcode",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":msgtxt",1,256));
     pq->cBind->push_back(new OCIDesc(":srcid",SQLT_INT,4));
     pq->cBind->allocBuffer();
     fb=pq->cBind->getField(2);
     fb->rlen=4;
     memcpy(fb->data,&uSrcId,fb->rlen);
     pq->cursorBind();
     
     vSQ[2]=pq=new OCIQuery(this);
     pq->setSQL("call wr_ipv4(:srcid,:state,:kind,:netaddr,:netsize,:bgporigin,:bgpaspath,:bgpcommunities,:bgpnexthop,:bgpmed,:bgplocalpref,:bgpatomicaggregate,:bgpaggregatoras,:bgpagregatorIP,:bgporiginatorIP,:bgpextcom,:bgplabels)");
     pq->cursorOpen();
     pq->cursorParse();
     pq->cBind->push_back(new OCIDesc(":srcid",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":state",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":kind",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":netaddr",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":netsize",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgporigin",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpaspath",1,256));
     pq->cBind->push_back(new OCIDesc(":bgpcommunities",1,256));
     pq->cBind->push_back(new OCIDesc(":bgpnexthop",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpmed",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgplocalpref",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpatomicaggregate",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpaggregatoras",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpagregatorIP",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgporiginatorIP",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpextcom",1,256));
     pq->cBind->push_back(new OCIDesc(":bgplabels",1,256));
     pq->cBind->allocBuffer();
     fb=pq->cBind->getField(0);
     fb->rlen=4;
     memcpy(fb->data,&uSrcId,fb->rlen);
     pq->cursorBind();
     
     vSQ[3]=pq=new OCIQuery(this);
     pq->setSQL("call wr_vpnv4(:srcid,:state,:kind,:netaddrrdh,:netaddrrdl,:netaddr,:netsize,:bgporigin,:bgpaspath,:bgpcommunities,:nexthoprdh,:nexthoprdl,:nexthop,:bgpmed,:bgplocalpref,:bgpatomicaggregate,:bgpaggregatoras,:bgpagregatorIP,:bgporiginatorIP,:bgpextcom,:bgplabels)");
     pq->cursorOpen();
     pq->cursorParse();
     pq->cBind->push_back(new OCIDesc(":srcid",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":state",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":kind",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":netaddrrdh",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":netaddrrdl",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":netaddr",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":netsize",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgporigin",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpaspath",1,256));
     pq->cBind->push_back(new OCIDesc(":bgpcommunities",1,256));
     pq->cBind->push_back(new OCIDesc(":nexthoprdh",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":nexthoprdl",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":nexthop",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpmed",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgplocalpref",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpatomicaggregate",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpaggregatoras",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpagregatorIP",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgporiginatorIP",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpextcom",1,256));
     pq->cBind->push_back(new OCIDesc(":bgplabels",1,256));
     pq->cBind->allocBuffer();
     fb=pq->cBind->getField(0);
     fb->rlen=4;
     memcpy(fb->data,&uSrcId,fb->rlen);
     pq->cursorBind();

     vSQ[4]=pq=new OCIQuery(this);
     pq->setSQL("call wr_ipv6(:srcid,:state,:kind,:netaddrh,:netaddrl,:netsize,:bgporigin ,:bgpaspath,:bgpcommunities,:bgpnexthoph,:bgpnexthopl,:bgpmed,:bgplocalpref,:bgpatomicaggregateh,:bgpatomicaggregatel,:bgpaggregatoras,:bgpagregatorIPh,:bgpagregatorIPl,:bgporiginatorIPh,:bgporiginatorIPl,:bgpextcom,:bgplabels)");
     pq->cursorOpen();
     pq->cursorParse();
     pq->cBind->push_back(new OCIDesc(":srcid",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":state",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":kind",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":netaddrh",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":netaddrl",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":netsize",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgporigin",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpaspath",1,256));
     pq->cBind->push_back(new OCIDesc(":bgpcommunities",1,256));
     pq->cBind->push_back(new OCIDesc(":bgpnexthoph",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":bgpnexthopl",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":bgpmed",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgplocalpref",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpatomicaggregateh",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":bgpatomicaggregatel",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":bgpaggregatoras",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpagregatorIPh",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":bgpagregatorIPl",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":bgporiginatorIPh",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":bgporiginatorIPl",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":bgpextcom",1,256));
     pq->cBind->push_back(new OCIDesc(":bgplabels",1,256));
     pq->cBind->allocBuffer();
     fb=pq->cBind->getField(0);
     fb->rlen=4;
     memcpy(fb->data,&uSrcId,fb->rlen);
     pq->cursorBind();
     
     vSQ[5]=pq=new OCIQuery(this);
     pq->setSQL("call wr_vpnv6(:srcid,:state,:kind,:netaddrrdh,:netaddrrdl,:netaddrh,:netaddrl,:netsize,:bgporigin ,:bgpaspath,:bgpcommunities,:nhrdh,:nhrdl,:nhh,:nhl,:bgpmed,:bgplocalpref,:bgpatomicaggregateh,:bgpatomicaggregatel,:bgpaggregatoras,:bgpagregatorIPh,:bgpagregatorIPl,:bgporiginatorIPh,:bgporiginatorIPl,:bgpextcom,:bgplabels)");
     pq->cursorOpen();
     pq->cursorParse();
     pq->cBind->push_back(new OCIDesc(":srcid",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":state",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":kind",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":netaddrrdh",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":netaddrrdl",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":netaddrh",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":netaddrl",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":netsize",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgporigin",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpaspath",1,256));
     pq->cBind->push_back(new OCIDesc(":bgpcommunities",1,256));
     pq->cBind->push_back(new OCIDesc(":nhrdh",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":nhrdl",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":nhh",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":nhl",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":bgpmed",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgplocalpref",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpatomicaggregateh",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":bgpatomicaggregatel",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":bgpaggregatoras",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpagregatorIPh",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":bgpagregatorIPl",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":bgporiginatorIPh",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":bgporiginatorIPl",SQLT_INT,8));
     pq->cBind->push_back(new OCIDesc(":bgpextcom",1,256));
     pq->cBind->push_back(new OCIDesc(":bgplabels",1,256));
     pq->cBind->allocBuffer();
     fb=pq->cBind->getField(0);
     fb->rlen=4;
     memcpy(fb->data,&uSrcId,fb->rlen);
     pq->cursorBind();
     
     vSQ[6]=pq=new OCIQuery(this);
     pq->setSQL("call wr_l2vpn(:srcid,:state,:kind,:site,:offset,:range,:bgporigin,:bgpaspath,:bgpcommunities,:bgpnexthop,:bgpmed,:bgplocalpref,:bgpatomicaggregate,:bgpaggregatoras,:bgpagregatorIP,:bgporiginatorIP,:bgpextcom,:bgplabels)");
     pq->cursorOpen();
     pq->cursorParse();
     pq->cBind->push_back(new OCIDesc(":srcid",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":state",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":kind",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":site",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":offset",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":range",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgporigin",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpaspath",1,256));
     pq->cBind->push_back(new OCIDesc(":bgpcommunities",1,256));
     pq->cBind->push_back(new OCIDesc(":bgpnexthop",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpmed",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgplocalpref",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpatomicaggregate",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpaggregatoras",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpagregatorIP",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgporiginatorIP",SQLT_INT,4));
     pq->cBind->push_back(new OCIDesc(":bgpextcom",1,256));
     pq->cBind->push_back(new OCIDesc(":bgplabels",1,256));
     pq->cBind->allocBuffer();
     fb=pq->cBind->getField(0);
     fb->rlen=4;
     memcpy(fb->data,&uSrcId,fb->rlen);
     pq->cursorBind();
     
    } catch(const std::exception & e){
        fprintf(stderr,"Exception caught while initializing BGPLogOCI: '%s'\n",e.what());
    } catch(...){
        fprintf(stderr,"Exception caught while initializing BGPLogOCI\n");
    }
    logtxt(1,"Started");
}
BGPLogOCI::~BGPLogOCI(){
    sword status;
    if (!srv) return;
    logtxt(2,"Finished");
    std::map<int,OCIQuery *>::iterator qi;
    for(qi=vSQ.begin();qi!=vSQ.end();qi++){
        delete qi->second;
        qi->second=0;
    }
    status=OCILogoff(srv,err);
    OCIHandleFree(srv,OCI_HTYPE_SVCCTX);srv=0;
    OCIHandleFree(err,OCI_HTYPE_ERROR);err=0;
    
}
void BGPLogOCI::commit(){
        if (!srv) return ;
        sword status;
        status=OCITransCommit(srv,err,0);
        checkStatus(status);
}

void BGPLogOCI::rollback(){
        if (!srv) return ;
        sword status;
        status=OCITransRollback(srv,err,0);
        checkStatus(status);
}

void BGPLogOCI::logtxt(unsigned uCode,const char *sTxt){
    if (!srv) return ;
    OCIQuery *pq=vSQ[1];
    if (!pq) return ;
    OCIFldBuf *fb=pq->cBind->getField(0);
    fb->rlen=4;
    memcpy(fb->data,&uCode,fb->rlen);
    fb=pq->cBind->getField(1);
    fb->rlen=strlen(sTxt);
    memcpy(fb->data,sTxt,fb->rlen);
    pq->cursorExecute(1);
    commit();
}

void BGPLogOCI::open(BGPPeer *peer,const MSG_Open *msg){
    //cout << (*msg) << endl;
    stringstream ss(stringstream::in | stringstream::out);
    ss << (*msg);
    std::string s;
    ss >> s;
    logtxt(3,s.c_str());
}
void BGPLogOCI::close(BGPPeer *peer){
    logtxt(4,"close");
}
void BGPLogOCI::txt(BGPPeer *peer,const char *txt){
	//cout << txt << endl;
	logtxt(0,txt);
}
void BGPLogOCI::exception(BGPPeer *peer,const BGPException *e){
	if (e) cout << e->what() << endl;
}
void BGPLogOCI::notification(BGPPeer *peer,const MSG_Notify *msg){
    /*
    stringstream ss(stringstream::in | stringstream::out);
    ss << (*msg);
    std::string s;
    ss >> s;
    logtxt(4,s.c_str());
    */
}
void BGPLogOCI::keepalive(BGPPeer *peer){
    //logtxt(5,"keepalive");
}
void BGPLogOCI::exec_ipv4(unsigned state,unsigned kind,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPNets & nets){
    if (!srv) return ;
    std::string sQ,sV,sS;
    unsigned i,u;
    IPNets::const_iterator ci;
    OCIQuery *q=vSQ[2];
    if (!q) return;
    OCIFldBuf *fb=q->cBind->getField(1);fb->rlen=4;memcpy(fb->data,&state,4);
    fb=q->cBind->getField(2);fb->rlen=4;memcpy(fb->data,&kind,4);
    fb=q->cBind->getField(5);fb->rlen=4;*(unsigned *)(fb->data)=attr.origin;
    fb=q->cBind->getField(6);
    if (attr.vASPath.size()>0){
        sV="";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    fb=q->cBind->getField(7);
    if (attr.vCom.size()>0){
        sV="";
        BGPCommunities::const_iterator cmi;
        i=0;
        for(cmi=attr.vCom.begin();cmi!=attr.vCom.end();cmi++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)(*cmi)) >> 16,((unsigned)(*cmi)) & 0xffff);
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    fb=q->cBind->getField(8);fb->rlen=4;*(unsigned *)(fb->data)=(unsigned)nexthop.getIPv4Addr();
    fb=q->cBind->getField(9);if (attr.med==uEmpty) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.med;}
    fb=q->cBind->getField(10);if (attr.localpref==uEmpty) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.localpref;}
    fb=q->cBind->getField(11);if (attr.aAtomicAggregate.getIPv4Addr()==0) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aAtomicAggregate.getIPv4Addr();}
    fb=q->cBind->getField(12);if (attr.aAggregatorAS==0) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aAggregatorAS;}
    fb=q->cBind->getField(13);if (attr.aAggregatorIP.getType()!=atIPv4) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aAggregatorIP.getIPv4Addr();}
    fb=q->cBind->getField(14);if (attr.aOriginator.getType()!=atIPv4) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aOriginator.getIPv4Addr();}
    fb=q->cBind->getField(15);
    if (attr.vExtCom.size()>0){
        sV="";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    fb=q->cBind->getField(16);fb->ind=-1;
    for(ci=nets.begin();ci!=nets.end();ci++){
        fb=q->cBind->getField(3);
        fb->rlen=4;fb->ind=0;
        *(unsigned *)(fb->data)=(unsigned)ci->getIPv4Addr();
        fb=q->cBind->getField(4);
        fb->rlen=4;fb->ind=0;
        *(unsigned *)(fb->data)=(unsigned)ci->getPrefixLen();
        q->cursorExecute(1);
    }
}
void BGPLogOCI::exec_ipv6(unsigned state,unsigned kind,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPNets & nets){
    if (!srv) return ;
    std::string sQ,sV,sS;
    unsigned i,u;
    IPNets::const_iterator ci;
    OCIQuery *q=vSQ[4];
    if (!q) return;
    OCIFldBuf *fb=q->cBind->getField(1);fb->rlen=4;memcpy(fb->data,&state,4);
    fb=q->cBind->getField(2);fb->rlen=4;memcpy(fb->data,&kind,4);
    fb=q->cBind->getField(6);fb->rlen=4;*(unsigned *)(fb->data)=attr.origin;
    fb=q->cBind->getField(7);
    if (attr.vASPath.size()>0){
        sV="";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    fb=q->cBind->getField(8);
    if (attr.vCom.size()>0){
        sV="";
        BGPCommunities::const_iterator cmi;
        i=0;
        for(cmi=attr.vCom.begin();cmi!=attr.vCom.end();cmi++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)(*cmi)) >> 16,((unsigned)(*cmi)) & 0xffff);
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    fb=q->cBind->getField(9);fb->rlen=8;*(uint64_t *)(fb->data)=nexthop.getIPv6AddrH();
    fb=q->cBind->getField(10);fb->rlen=8;*(uint64_t *)(fb->data)=nexthop.getIPv6AddrL();
    fb=q->cBind->getField(11);if (attr.med==uEmpty) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.med;}
    fb=q->cBind->getField(12);if (attr.localpref==uEmpty) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.localpref;}
    fb=q->cBind->getField(13);if (attr.aAtomicAggregate.getIPv6AddrH()==0) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=(uint64_t)attr.aAtomicAggregate.getIPv6AddrH();}
    fb=q->cBind->getField(14);if (attr.aAtomicAggregate.getIPv6AddrL()==0) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=(uint64_t)attr.aAtomicAggregate.getIPv6AddrL();}
    fb=q->cBind->getField(15);if (attr.aAggregatorAS==0) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aAggregatorAS;}
    fb=q->cBind->getField(16);if (attr.aAggregatorIP.getType()!=atIPv6) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=(uint64_t)attr.aAggregatorIP.getIPv6AddrH();}
    fb=q->cBind->getField(17);if (attr.aAggregatorIP.getType()!=atIPv6) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=(uint64_t)attr.aAggregatorIP.getIPv6AddrL();}
    fb=q->cBind->getField(18);if (attr.aOriginator.getType()!=atIPv6) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=(uint64_t)attr.aOriginator.getIPv6AddrH();}
    fb=q->cBind->getField(19);if (attr.aOriginator.getType()!=atIPv6) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=(uint64_t)attr.aOriginator.getIPv6AddrL();}
    fb=q->cBind->getField(20);
    if (attr.vExtCom.size()>0){
        sV="";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    fb=q->cBind->getField(21);fb->ind=-1;fb->rlen=0;
    for(ci=nets.begin();ci!=nets.end();ci++){
        fb=q->cBind->getField(3);
        fb->rlen=8;fb->ind=0;
        *(uint64_t *)(fb->data)=ci->getIPv6AddrH();
        fb=q->cBind->getField(4);
        fb->rlen=8;fb->ind=0;
        *(uint64_t *)(fb->data)=ci->getIPv6AddrL();
        fb=q->cBind->getField(5);
        fb->rlen=4;fb->ind=0;
        *(unsigned *)(fb->data)=(unsigned)ci->getPrefixLen();
        q->cursorExecute(1);
    }
}
void BGPLogOCI::exec_ipv4l(unsigned state,unsigned kind,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    if (!srv) return ;
    std::string sQ,sV,sS;
    unsigned i,u;
    IPTagNets::const_iterator ci;
    MPLSLabels::const_iterator mi;
    OCIQuery *q=vSQ[2];
    if (!q) return;
    OCIFldBuf *fb=q->cBind->getField(1);fb->rlen=4;memcpy(fb->data,&state,4);
    fb=q->cBind->getField(2);fb->rlen=4;memcpy(fb->data,&kind,4);
    fb=q->cBind->getField(5);fb->rlen=4;*(unsigned *)(fb->data)=attr.origin;
    fb=q->cBind->getField(6);
    if (attr.vASPath.size()>0){
        sV="";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    fb=q->cBind->getField(7);
    if (attr.vCom.size()>0){
        sV="";
        BGPCommunities::const_iterator cmi;
        i=0;
        for(cmi=attr.vCom.begin();cmi!=attr.vCom.end();cmi++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)(*cmi)) >> 16,((unsigned)(*cmi)) & 0xffff);
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    fb=q->cBind->getField(8);fb->rlen=4;*(unsigned *)(fb->data)=(unsigned)nexthop.getIPv4Addr();
    fb=q->cBind->getField(9);if (attr.med==uEmpty) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.med;}
    fb=q->cBind->getField(10);if (attr.localpref==uEmpty) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.localpref;}
    fb=q->cBind->getField(11);if (attr.aAtomicAggregate.getIPv4Addr()==0) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aAtomicAggregate.getIPv4Addr();}
    fb=q->cBind->getField(12);if (attr.aAggregatorAS==0) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aAggregatorAS;}
    fb=q->cBind->getField(13);if (attr.aAggregatorIP.getType()!=atIPv4) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aAggregatorIP.getIPv4Addr();}
    fb=q->cBind->getField(14);if (attr.aOriginator.getType()!=atIPv4) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aOriginator.getIPv4Addr();}
    fb=q->cBind->getField(15);
    if (attr.vExtCom.size()>0){
        sV="";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    fb=q->cBind->getField(16);fb->ind=-1;
    for(ci=nets.begin();ci!=nets.end();ci++){
        fb=q->cBind->getField(3);
        fb->rlen=4;fb->ind=0;
        *(unsigned *)(fb->data)=(unsigned)ci->getIPv4Addr();
        fb=q->cBind->getField(4);
        fb->rlen=4;fb->ind=0;
        *(unsigned *)(fb->data)=(unsigned)ci->getPrefixLen();
        sV="";
        for(i=0,mi=ci->labels.begin();mi!=ci->labels.end();mi++,i++){
            if (i) sV+=" ";
            sV+=Sprintf("%u",(unsigned)*mi);
        }
        fb=q->cBind->getField(16);fb->ind=0;
        fb->rlen=min<size_t>(sV.size(),256);
        strncpy(fb->data,sV.c_str(),fb->rlen);
        q->cursorExecute(1);
    }
}
void BGPLogOCI::exec_ipv6l(unsigned state,unsigned kind,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    if (!srv) return ;
    std::string sQ,sV,sS;
    unsigned i,u;
    IPTagNets::const_iterator ci;
    MPLSLabels::const_iterator mi;
    OCIQuery *q=vSQ[4];
    if (!q) return;
    OCIFldBuf *fb=q->cBind->getField(1);fb->rlen=4;memcpy(fb->data,&state,4);
    fb=q->cBind->getField(2);fb->rlen=4;memcpy(fb->data,&kind,4);
    fb=q->cBind->getField(6);fb->rlen=4;*(unsigned *)(fb->data)=attr.origin;
    fb=q->cBind->getField(7);
    if (attr.vASPath.size()>0){
        sV="";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    fb=q->cBind->getField(8);
    if (attr.vCom.size()>0){
        sV="";
        BGPCommunities::const_iterator cmi;
        i=0;
        for(cmi=attr.vCom.begin();cmi!=attr.vCom.end();cmi++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)(*cmi)) >> 16,((unsigned)(*cmi)) & 0xffff);
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    fb=q->cBind->getField(9);fb->rlen=8;*(uint64_t *)(fb->data)=nexthop.getIPv6AddrH();
    fb=q->cBind->getField(10);fb->rlen=8;*(uint64_t *)(fb->data)=nexthop.getIPv6AddrL();
    fb=q->cBind->getField(11);if (attr.med==uEmpty) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.med;}
    fb=q->cBind->getField(12);if (attr.localpref==uEmpty) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.localpref;}
    fb=q->cBind->getField(13);if (attr.aAtomicAggregate.getIPv6AddrH()==0) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=(uint64_t)attr.aAtomicAggregate.getIPv6AddrH();}
    fb=q->cBind->getField(14);if (attr.aAtomicAggregate.getIPv6AddrL()==0) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=(uint64_t)attr.aAtomicAggregate.getIPv6AddrL();}
    fb=q->cBind->getField(15);if (attr.aAggregatorAS==0) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aAggregatorAS;}
    fb=q->cBind->getField(16);if (attr.aAggregatorIP.getType()!=atIPv6) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=(uint64_t)attr.aAggregatorIP.getIPv6AddrH();}
    fb=q->cBind->getField(17);if (attr.aAggregatorIP.getType()!=atIPv6) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=(uint64_t)attr.aAggregatorIP.getIPv6AddrL();}
    fb=q->cBind->getField(18);if (attr.aOriginator.getType()!=atIPv6) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=(uint64_t)attr.aOriginator.getIPv6AddrH();}
    fb=q->cBind->getField(19);if (attr.aOriginator.getType()!=atIPv6) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=(uint64_t)attr.aOriginator.getIPv6AddrL();}
    fb=q->cBind->getField(20);
    if (attr.vExtCom.size()>0){
        sV="";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    for(ci=nets.begin();ci!=nets.end();ci++){
        fb=q->cBind->getField(3);
        fb->rlen=8;fb->ind=0;
        *(uint64_t *)(fb->data)=ci->getIPv6AddrH();
        fb=q->cBind->getField(4);
        fb->rlen=8;fb->ind=0;
        *(uint64_t *)(fb->data)=ci->getIPv6AddrL();
        fb=q->cBind->getField(5);
        fb->rlen=4;fb->ind=0;
        *(unsigned *)(fb->data)=(unsigned)ci->getPrefixLen();
        sS="";
        for(i=0,mi=ci->labels.begin();mi!=ci->labels.end();mi++,i++){
            if (i) sS+=" ";
            sS+=Sprintf("%u",(unsigned)*mi);
        }        
        fb=q->cBind->getField(21);fb->ind=0;fb->rlen=sS.length();
        if (fb->rlen>256) fb->rlen=256;
        strncpy(fb->data,sS.c_str(),fb->rlen);
        q->cursorExecute(1);
    }
}
void BGPLogOCI::exec_vpnv4(unsigned state,unsigned kind,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    if (!srv) return ;
    std::string sQ,sV,sS;
    unsigned i,u;
    IPTagNets::const_iterator ci;
    MPLSLabels::const_iterator mi;
    OCIQuery *q=vSQ[3];
    if (!q) return;
    OCIFldBuf *fb=q->cBind->getField(1);fb->rlen=4;memcpy(fb->data,&state,4);
    fb=q->cBind->getField(2);fb->rlen=4;memcpy(fb->data,&kind,4);
    fb=q->cBind->getField(7);fb->rlen=4;*(unsigned *)(fb->data)=attr.origin;
    fb=q->cBind->getField(8);
    if (attr.vASPath.size()>0){
        sV="";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    fb=q->cBind->getField(9);
    if (attr.vCom.size()>0){
        sV="";
        BGPCommunities::const_iterator cmi;
        i=0;
        for(cmi=attr.vCom.begin();cmi!=attr.vCom.end();cmi++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)(*cmi)) >> 16,((unsigned)(*cmi)) & 0xffff);
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    fb=q->cBind->getField(10);fb->rlen=4;*(unsigned *)(fb->data)=(unsigned)nexthop.getRDH();
    fb=q->cBind->getField(11);fb->rlen=4;*(unsigned *)(fb->data)=(unsigned)nexthop.getRDL();
    fb=q->cBind->getField(12);fb->rlen=4;*(unsigned *)(fb->data)=(unsigned)nexthop.getIPv4Addr();

    fb=q->cBind->getField(13);if (attr.med==uEmpty) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.med;}
    fb=q->cBind->getField(14);if (attr.localpref==uEmpty) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.localpref;}
    fb=q->cBind->getField(15);if (attr.aAtomicAggregate.getIPv4Addr()==0) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aAtomicAggregate.getIPv4Addr();}
    fb=q->cBind->getField(16);if (attr.aAggregatorAS==0) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aAggregatorAS;}
    fb=q->cBind->getField(17);if (attr.aAggregatorIP.getType()!=atIPv4) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aAggregatorIP.getIPv4Addr();}
    fb=q->cBind->getField(18);if (attr.aOriginator.getType()!=atIPv4) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aOriginator.getIPv4Addr();}
    fb=q->cBind->getField(19);
    if (attr.vExtCom.size()>0){
        sV="";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    fb=q->cBind->getField(16);fb->ind=-1;
    for(ci=nets.begin();ci!=nets.end();ci++){
        fb=q->cBind->getField(3);
        fb->rlen=4;fb->ind=0;
        *(unsigned *)(fb->data)=(unsigned)ci->getRDH();
        fb=q->cBind->getField(4);
        fb->rlen=4;fb->ind=0;
        *(unsigned *)(fb->data)=(unsigned)ci->getRDL();
        fb=q->cBind->getField(5);
        fb->rlen=4;fb->ind=0;
        *(unsigned *)(fb->data)=(unsigned)ci->getIPv4Addr();
        fb=q->cBind->getField(6);
        fb->rlen=4;fb->ind=0;
        *(unsigned *)(fb->data)=(unsigned)ci->getPrefixLen();
        sV="";
        for(i=0,mi=ci->labels.begin();mi!=ci->labels.end();mi++,i++){
            if (i) sV+=" ";
            sV+=Sprintf("%u",(unsigned)*mi);
        }
        fb=q->cBind->getField(20);fb->ind=0;
        fb->rlen=min<size_t>(sV.size(),256);
        strncpy(fb->data,sV.c_str(),fb->rlen);
        q->cursorExecute(1);
    }    
}
void BGPLogOCI::exec_vpnv6(unsigned state,unsigned kind,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    if (!srv) return ;
    std::string sQ,sV,sS;
    unsigned i,u;
    IPTagNets::const_iterator ci;
    MPLSLabels::const_iterator mi;
    OCIQuery *q=vSQ[5];
    if (!q) return;
    OCIFldBuf *fb=q->cBind->getField(1);fb->rlen=4;memcpy(fb->data,&state,4);
    fb=q->cBind->getField(2);fb->rlen=4;memcpy(fb->data,&kind,4);
    fb=q->cBind->getField(8);fb->rlen=4;*(unsigned *)(fb->data)=attr.origin;
    fb=q->cBind->getField(9);
    if (attr.vASPath.size()>0){
        sV="";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    fb=q->cBind->getField(10);
    if (attr.vCom.size()>0){
        sV="";
        BGPCommunities::const_iterator cmi;
        i=0;
        for(cmi=attr.vCom.begin();cmi!=attr.vCom.end();cmi++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)(*cmi)) >> 16,((unsigned)(*cmi)) & 0xffff);
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    fb=q->cBind->getField(11);fb->rlen=4;*(unsigned *)(fb->data)=nexthop.getRDH();
    fb=q->cBind->getField(12);fb->rlen=4;*(unsigned *)(fb->data)=nexthop.getRDL();
    fb=q->cBind->getField(13);fb->rlen=8;*(uint64_t *)(fb->data)=nexthop.getIPv6AddrH();
    fb=q->cBind->getField(14);fb->rlen=8;*(uint64_t *)(fb->data)=nexthop.getIPv6AddrL();
    fb=q->cBind->getField(15);if (attr.med==uEmpty) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.med;}
    fb=q->cBind->getField(16);if (attr.localpref==uEmpty) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.localpref;}
    fb=q->cBind->getField(17);if (attr.aAtomicAggregate.getIPv6AddrH()==0) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=(uint64_t)attr.aAtomicAggregate.getIPv6AddrH();}
    fb=q->cBind->getField(18);if (attr.aAtomicAggregate.getIPv6AddrL()==0) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=(uint64_t)attr.aAtomicAggregate.getIPv6AddrL();}
    fb=q->cBind->getField(19);if (attr.aAggregatorAS==0) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aAggregatorAS;}
    fb=q->cBind->getField(20);if (attr.aAggregatorIP.getType()!=atIPv6) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=(uint64_t)attr.aAggregatorIP.getIPv6AddrH();}
    fb=q->cBind->getField(21);if (attr.aAggregatorIP.getType()!=atIPv6) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=(uint64_t)attr.aAggregatorIP.getIPv6AddrL();}
    fb=q->cBind->getField(22);if (attr.aOriginator.getType()!=atIPv6) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=(uint64_t)attr.aOriginator.getIPv6AddrH();}
    fb=q->cBind->getField(23);if (attr.aOriginator.getType()!=atIPv6) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=(uint64_t)attr.aOriginator.getIPv6AddrL();}
    fb=q->cBind->getField(24);
    if (attr.vExtCom.size()>0){
        sV="";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    for(ci=nets.begin();ci!=nets.end();ci++){
        fb=q->cBind->getField(3);fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)ci->getIPv6AddrH();
        fb=q->cBind->getField(4);fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)ci->getIPv6AddrL();
        fb=q->cBind->getField(5);fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=ci->getIPv6AddrH();
        fb=q->cBind->getField(6);fb->rlen=8;fb->ind=0;*(uint64_t *)(fb->data)=ci->getIPv6AddrL();
        fb=q->cBind->getField(7);fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)ci->getPrefixLen();
        sS="";
        for(i=0,mi=ci->labels.begin();mi!=ci->labels.end();mi++,i++){
            if (i) sS+=" ";
            sS+=Sprintf("%u",(unsigned)*mi);
        }        
        fb=q->cBind->getField(25);fb->ind=0;fb->rlen=sS.length();
        if (fb->rlen>256) fb->rlen=256;
        strncpy(fb->data,sS.c_str(),fb->rlen);
        q->cursorExecute(1);
    }
}
void BGPLogOCI::exec_l2vpn(unsigned state,unsigned kind,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    if (!srv) return ;
    std::string sQ,sV,sS;
    unsigned i,u;
    IPTagNets::const_iterator ci;
    MPLSLabels::const_iterator mi;
    OCIQuery *q=vSQ[6];
    if (!q) return;
    OCIFldBuf *fb=q->cBind->getField(1);fb->rlen=4;memcpy(fb->data,&state,4);
    fb=q->cBind->getField(2);fb->rlen=4;memcpy(fb->data,&kind,4);
    fb=q->cBind->getField(6);fb->rlen=4;*(unsigned *)(fb->data)=attr.origin;
    fb=q->cBind->getField(7);
    if (attr.vASPath.size()>0){
        sV="";
        BGPASNs::const_iterator ci;
        for(i=0,ci=attr.vASPath.begin();ci!=attr.vASPath.end();ci++,i++){
            sV+=Sprintf(i==0 ? "%u":" %u",(unsigned)*ci);
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    fb=q->cBind->getField(8);
    if (attr.vCom.size()>0){
        sV="";
        BGPCommunities::const_iterator cmi;
        i=0;
        for(cmi=attr.vCom.begin();cmi!=attr.vCom.end();cmi++,i++){
            sV+=Sprintf(i==0 ? "%u:%u":" %u:%u",((unsigned)(*cmi)) >> 16,((unsigned)(*cmi)) & 0xffff);
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    fb=q->cBind->getField(9);fb->rlen=4;*(unsigned *)(fb->data)=(unsigned)nexthop.getIPv4Addr();
    fb=q->cBind->getField(10);if (attr.med==uEmpty) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.med;}
    fb=q->cBind->getField(11);if (attr.localpref==uEmpty) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.localpref;}
    fb=q->cBind->getField(12);if (attr.aAtomicAggregate.getIPv4Addr()==0) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aAtomicAggregate.getIPv4Addr();}
    fb=q->cBind->getField(13);if (attr.aAggregatorAS==0) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aAggregatorAS;}
    fb=q->cBind->getField(14);if (attr.aAggregatorIP.getType()!=atIPv4) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aAggregatorIP.getIPv4Addr();}
    fb=q->cBind->getField(15);if (attr.aOriginator.getType()!=atIPv4) {fb->rlen=0;fb->ind=-1;} else {fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)attr.aOriginator.getIPv4Addr();}
    fb=q->cBind->getField(16);
    if (attr.vExtCom.size()>0){
        sV="";
        BGPExtCommunities::const_iterator ci;
        for(i=0,ci=attr.vExtCom.begin();ci!=attr.vExtCom.end();ci++,i++){
            sV+=(i==0 ? "":" ")+ci->str();
        }
        fb->rlen=min<size_t>(sV.size(),256);
        fb->ind=0;
        strncpy(fb->data,sV.c_str(),fb->rlen);
    } else {
        fb->rlen=0;
        fb->ind=-1;
    }
    for(ci=nets.begin();ci!=nets.end();ci++){
        fb=q->cBind->getField(3);fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)ci->getL2Site();
        fb=q->cBind->getField(4);fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)ci->getL2Offset();
        fb=q->cBind->getField(3);fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)ci->getL2Site();
        fb=q->cBind->getField(5);fb->rlen=4;fb->ind=0;*(unsigned *)(fb->data)=(unsigned)ci->getL2Range();
        sV="";
        for(i=0,mi=ci->labels.begin();mi!=ci->labels.end();mi++,i++){
            if (i) sV+=" ";
            sV+=Sprintf("%u",(unsigned)*mi);
        }
        fb=q->cBind->getField(17);fb->ind=0;
        fb->rlen=min<size_t>(sV.size(),256);
        strncpy(fb->data,sV.c_str(),fb->rlen);
        q->cursorExecute(1);
    }
}
void BGPLogOCI::nlri(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPNets & nets){
    //cout << "nlri: " << attr << nets << endl;
    exec_ipv4(1,1,attr,attr.aNextHop,nets);
}
void BGPLogOCI::withdraw(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPNets & nets){
    //cout << "withdraw: " << attr << nets << endl;
    exec_ipv4(2,1,attr,attr.aNextHop,nets);
}
void BGPLogOCI::nlri_ipv4m(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPNets & nets){
	//cout << "nlri_ipv4lu: attr(" << attr << ") nh " << nexthop << " nets: " << nets  << endl;
	exec_ipv4(1,2,attr,nexthop,nets);
}
void BGPLogOCI::withdraw_ipv4m(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPNets & nets){
	//cout << "withdraw_ipv4lu: attr(" << attr << ") nh " << nexthop << " nets: " << nets  << endl;
	exec_ipv4(2,2,attr,nexthop,nets);
}
void BGPLogOCI::nlri_ipv4lu(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
	//cout << "nlri_ipv4lu: attr(" << attr << ") nh " << nexthop << " nets: " << nets  << endl;
	exec_ipv4l(1,4,attr,nexthop,nets);
}
void BGPLogOCI::withdraw_ipv4lu(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
	//cout << "withdraw_ipv4lu: attr(" << attr << ") nh " << nexthop << " nets: " << nets  << endl;
	exec_ipv4l(2,4,attr,nexthop,nets);
}
void BGPLogOCI::nlri_vpnv4u(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    //cout << "nlri_vpnv4u: attr(" << attr << ") nh " << nexthop  << " nets: " << nets  << endl;
    exec_vpnv4(1,128,attr,nexthop,nets);
}
void BGPLogOCI::withdraw_vpnv4u(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    //cout << "withdraw_vpnv4u: attr(" << attr << ") nh " << nexthop  << " nets: " << nets  << endl;
    exec_vpnv4(2,128,attr,nexthop,nets);
}
void BGPLogOCI::nlri_ipv6u(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPNets & nets){
	exec_ipv6(1,4,attr,nexthop,nets);
}
void BGPLogOCI::withdraw_ipv6u(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPNets & nets){
	exec_ipv6(2,4,attr,nexthop,nets);
}
void BGPLogOCI::nlri_ipv6lu(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
	exec_ipv6l(1,4,attr,nexthop,nets);
}
void BGPLogOCI::withdraw_ipv6lu(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
	exec_ipv6l(2,4,attr,nexthop,nets);
}
void BGPLogOCI::nlri_vpnv6u(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    exec_vpnv6(1,128,attr,nexthop,nets);
}
void BGPLogOCI::withdraw_vpnv6u(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    exec_vpnv6(2,128,attr,nexthop,nets);
}
void BGPLogOCI::nlri_l2vpn(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    exec_l2vpn(1,1,attr,nexthop,nets);
}
void BGPLogOCI::withdraw_l2vpn(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
    exec_l2vpn(2,1,attr,nexthop,nets);
}

}
