#ifndef __bgplogoci_h_
#define __bgplogoci_h_
#include "bgplog.h"

struct OCIEnv;
struct OCISvcCtx;
struct OCIError;
struct OCIServer;
namespace BGPHist {
class OCIQuery;
class BGPLogOCI: public BGPLog {
friend class OCIQuery;
friend class OCIDesc;
protected:
    OCIEnv *env;
    OCISvcCtx *srv;
    OCIError *err;
    OCIServer *srvr;
    std::map<int,OCIQuery *> vSQ;
    void logtxt(unsigned uCode,const char *);
    void exec_ipv4(unsigned state,unsigned kind,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPNets & );
    void exec_ipv4l(unsigned state,unsigned kind,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    void exec_vpnv4(unsigned state,unsigned kind,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    void exec_ipv6(unsigned state,unsigned kind,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPNets & );
    void exec_ipv6l(unsigned state,unsigned kind,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    void exec_vpnv6(unsigned state,unsigned kind,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    void exec_l2vpn(unsigned state,unsigned kind,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    void checkStatus(int,OCIError * =0);
    void commit();
    void rollback();
public:
    void * pmc;
    const char *ocicon,*dbuser,*dbpwd;
    unsigned uSrcId;
    BGPLogOCI(const char *_ocicon,const char *_dbuser,const char *_dbpwd,const char *_peerip,unsigned int _asn);
    virtual ~BGPLogOCI();
    virtual void open(BGPPeer *peer,const MSG_Open *);
    virtual void close(BGPPeer *peer);
    virtual void txt(BGPPeer *peer,const char *);
    virtual void exception(BGPPeer *peer,const BGPException *);
    virtual void notification(BGPPeer *peer,const MSG_Notify *);
    virtual void keepalive(BGPPeer *peer);
    virtual void nlri(BGPPeer *peer,const MSG_UpdateAttr &,const IPNets & );
    virtual void withdraw(BGPPeer *peer,const MSG_UpdateAttr &,const IPNets & );
    virtual void nlri_ipv4m(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPNets & );
    virtual void withdraw_ipv4m(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPNets & );
    virtual void nlri_ipv4lu(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    virtual void withdraw_ipv4lu(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    virtual void nlri_vpnv4u(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    virtual void withdraw_vpnv4u(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    virtual void nlri_ipv6u(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPNets & );
    virtual void withdraw_ipv6u(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPNets & );
    virtual void nlri_ipv6lu(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    virtual void withdraw_ipv6lu(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    virtual void nlri_vpnv6u(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    virtual void withdraw_vpnv6u(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    virtual void nlri_l2vpn(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    virtual void withdraw_l2vpn(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
};

};

#endif
