#include "bgplogproxy.h"
#include <unistd.h>
#include <iostream>
#include <pthread.h>


using namespace BGPHist;
using namespace std;

void * _async_thread(void *_lp){
    BGPLogProxy *work=(BGPLogProxy *)_lp;
    work->async_proc();
}

BGPItemAttr::BGPItemAttr():bTouch(false),bWithdrawn(false){};
BGPItemAttr::BGPItemAttr(const BGPItemAttr &x):MSG_UpdateAttr(x),bTouch(x.bTouch),bWithdrawn(x.bWithdrawn){}
BGPItemAttr::BGPItemAttr(const MSG_UpdateAttr &x):MSG_UpdateAttr(x),bTouch(true),bWithdrawn(false){}
BGPItemAttr::BGPItemAttr(const MSG_UpdateAttr &x,bool t):MSG_UpdateAttr(x),bTouch(t),bWithdrawn(false){}

BGPLogProxy::BGPLogProxy(BGPLog *_l,unsigned _asyncwait):log(_l),uWait(_asyncwait),bAsync(_asyncwait>0),bExit(false),uThread(0){
    if (bAsync){
        pthread_mutexattr_init(&uAttr);
        pthread_mutexattr_settype(&uAttr,PTHREAD_MUTEX_RECURSIVE);
        pthread_mutex_init(&uLock,&uAttr);
        pthread_create(&uThread,0,_async_thread,this);
    }
}
BGPLogProxy::~BGPLogProxy(){
    if (bAsync){
        bExit=true;
        void *p;
        pthread_join(uThread,&p);
        pthread_mutex_destroy(&uLock);
        pthread_mutexattr_destroy(&uAttr);
    }
}
bool BGPLogProxy::lock_acquire(){
    if (!bAsync) return true;
    int rt=pthread_mutex_lock(&uLock);
    if (rt!=0) {
        cerr << "pthread_mutex_lock returned " << rt << "\t" << strerror(rt) << std::endl;
        return false;
    }
    return true;
}
bool BGPLogProxy::lock_release(){
    if (!bAsync) return true;
    int rt=pthread_mutex_unlock(&uLock);
    if (rt!=0) {
        cerr << "pthread_mutex_unlock returned " << rt << "\t" << strerror(rt) << std::endl;
        return false;
    }
    return rt==0;
}
std::string BGPLogProxy::get_stats(){
    std::string ret;
    unsigned s,t;
    BGPIPTable::iterator bi;
    BGPIPTagTable::iterator bti;
    if (!lock_acquire()) return "Can't acquire lock";
    s=t=0;for(bi=ipv4u.begin();bi!=ipv4u.end();bi++) {s++;if (bi->second.bTouch) t++;}
    ret+=Sprintf("ipv4u(%u,%u) ",s,t);
    s=t=0;for(bi=ipv4m.begin();bi!=ipv4m.end();bi++) {s++;if (bi->second.bTouch) t++;}
    ret+=Sprintf("ipv4m(%u,%u) ",s,t);
    s=t=0;for(bti=ipv4l.begin();bti!=ipv4l.end();bti++) {s++;if (bti->second.bTouch) t++;}
    ret+=Sprintf("ipv4lu(%u,%u) ",s,t);
    s=t=0;for(bti=vpnv4u.begin();bti!=vpnv4u.end();bti++) {s++;if (bti->second.bTouch) t++;}
    ret+=Sprintf("vpnv4u(%u,%u) ",s,t);
    s=t=0;for(bi=ipv6u.begin();bi!=ipv6u.end();bi++) {s++;if (bi->second.bTouch) t++;}
    ret+=Sprintf("ipv6u(%u,%u) ",s,t);
    s=t=0;for(bti=ipv6l.begin();bti!=ipv6l.end();bti++) {s++;if (bti->second.bTouch) t++;}
    ret+=Sprintf("ipv6lu(%u,%u) ",s,t);
    s=t=0;for(bti=vpnv6u.begin();bti!=vpnv6u.end();bti++) {s++;if (bti->second.bTouch) t++;}
    ret+=Sprintf("vpnv6u(%u,%u) ",s,t);
    s=t=0;for(bti=l2vpn.begin();bti!=l2vpn.end();bti++) {s++;if (bti->second.bTouch) t++;}
    ret+=Sprintf("l2vpn(%u,%u) ",s,t);
    lock_release();
    return ret;
}
void BGPLogProxy::async_proc(){
    BGPIPTable::iterator bi;
    BGPIPTagTable::iterator bti;
    IPNet net;
    IPTagNet tnet;
    BGPItemAttr iattr;
    for(;!bExit;){
        if (!lock_acquire()) {
            usleep(uWait*1000);
            continue;
        }
        try {
        for(bti=l2vpn.begin();bti!=l2vpn.end();bti++) if (bti->second.bTouch) break;
        if (bti!=l2vpn.end()){
            bti->second.bTouch=false;
            tnet=bti->first;
            iattr=bti->second;
            lock_release();
            IPTagNets nts;nts.push_back(tnet);
            if (iattr.bWithdrawn)
                log->withdraw_l2vpn(_peer,iattr,iattr.aNextHop,nts);
            else
                log->nlri_l2vpn(_peer,iattr,iattr.aNextHop,nts);
            continue;
        }
        for(bi=ipv4m.begin();bi!=ipv4m.end();bi++) if (bi->second.bTouch) break;
        if (bi!=ipv4m.end()){
            bi->second.bTouch=false;
            net=bi->first;
            iattr=bi->second;
            lock_release();
            IPNets nts;nts.push_back(net);
            if (iattr.bWithdrawn)
                log->withdraw_ipv4m(_peer,iattr,iattr.aNextHop,nts);
            else
                log->nlri_ipv4m(_peer,iattr,iattr.aNextHop,nts);
            continue;
        }
        for(bti=ipv4l.begin();bti!=ipv4l.end();bti++) if (bti->second.bTouch) break;
        if (bti!=ipv4l.end()){
            bti->second.bTouch=false;
            tnet=bti->first;
            iattr=bti->second;
            lock_release();
            IPTagNets nts;nts.push_back(tnet);
            if (iattr.bWithdrawn)
                log->withdraw_ipv4lu(_peer,iattr,iattr.aNextHop,nts);
            else
                log->nlri_ipv4lu(_peer,iattr,iattr.aNextHop,nts);
            continue;
        }
        for(bti=vpnv4u.begin();bti!=vpnv4u.end();bti++) if (bti->second.bTouch) break;
        if (bti!=vpnv4u.end()){
            bti->second.bTouch=false;
            tnet=bti->first;
            iattr=bti->second;
            lock_release();
            IPTagNets nts;nts.push_back(tnet);
            if (iattr.bWithdrawn)
                log->withdraw_vpnv4u(_peer,iattr,iattr.aNextHop,nts);
            else
                log->nlri_vpnv4u(_peer,iattr,iattr.aNextHop,nts);
            continue;
        }
        for(bi=ipv6u.begin();bi!=ipv6u.end();bi++) if (bi->second.bTouch) break;
        if (bi!=ipv6u.end()){
            bi->second.bTouch=false;
            net=bi->first;
            iattr=bi->second;
            lock_release();
            IPNets nts;nts.push_back(net);
            if (iattr.bWithdrawn)
                log->withdraw_ipv6u(_peer,iattr,iattr.aNextHop,nts);
            else
                log->nlri_ipv6u(_peer,iattr,iattr.aNextHop,nts);
            continue;
        }
        for(bti=ipv6l.begin();bti!=ipv6l.end();bti++) if (bti->second.bTouch) break;
        if (bti!=ipv6l.end()){
            bti->second.bTouch=false;
            tnet=bti->first;
            iattr=bti->second;
            lock_release();
            IPTagNets nts;nts.push_back(tnet);
            if (iattr.bWithdrawn)
                log->withdraw_ipv6lu(_peer,iattr,iattr.aNextHop,nts);
            else
                log->nlri_ipv6lu(_peer,iattr,iattr.aNextHop,nts);
            continue;
        }
        for(bti=vpnv6u.begin();bti!=vpnv6u.end();bti++) if (bti->second.bTouch) break;
        if (bti!=vpnv6u.end()){
            bti->second.bTouch=false;
            tnet=bti->first;
            iattr=bti->second;
            lock_release();
            IPTagNets nts;nts.push_back(tnet);
            if (iattr.bWithdrawn)
                log->withdraw_vpnv6u(_peer,iattr,iattr.aNextHop,nts);
            else
                log->nlri_vpnv6u(_peer,iattr,iattr.aNextHop,nts);
            continue;
        }
        for(bi=ipv4u.begin();bi!=ipv4u.end();bi++) if (bi->second.bTouch) break;
        if (bi!=ipv4u.end()){
            bi->second.bTouch=false;
            net=bi->first;
            iattr=bi->second;
            lock_release();
            IPNets nts;nts.push_back(net);
            if (iattr.bWithdrawn)
                log->withdraw(_peer,iattr,nts);
            else
                log->nlri(_peer,iattr,nts);
            continue;
        }
        } catch(const std::exception & e){
            std::cerr << e.what() << std::endl;
        }
        lock_release();
        usleep(uWait*1000);
    }
}
void BGPLogProxy::open(BGPPeer *peer,const MSG_Open *msg){
    _peer=peer;
	if (log) log->open(peer,msg);
}
void BGPLogProxy::close(BGPPeer *peer){
	if (log) log->close(peer);
}
void BGPLogProxy::txt(BGPPeer *peer,const char *txt){
	if (log) log->txt(peer,txt);
}
void BGPLogProxy::exception(BGPPeer *peer,const BGPException *e){
	if (log) log->exception(peer,e);
}
void BGPLogProxy::notification(BGPPeer *peer,const MSG_Notify *n){
	if (log) log->notification(peer,n);
}
void BGPLogProxy::keepalive(BGPPeer *peer){
	if (log) log->keepalive(peer);
}
void BGPLogProxy::nlri(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPNets & nets){
 if (!log) return ;
 if (!lock_acquire()) return ;
 IPNets nlri;
 IPNets::const_iterator i;
 BGPIPTable::iterator j;
 for(i=nets.begin();i!=nets.end();i++){
  j=ipv4u.find(*i);
  if (j!=ipv4u.end()){
   if (j->second==attr) continue;
  }
  nlri.push_back(*i);
 }
 if (nlri.empty()) {lock_release();return ;}
 for(i=nlri.begin();i!=nlri.end();i++){
  ipv4u[*i]=attr;
 }
 if (!bAsync) log->nlri(peer,attr,nlri);
 lock_release();
}
void BGPLogProxy::withdraw(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPNets & nets){
 if (!log) return ;
 if (!lock_acquire()) return ;
 IPNets nlri,cnet;
 IPNets::const_iterator i;
 BGPIPTable::iterator j;
 for(i=nets.begin();i!=nets.end();i++){
  j=ipv4u.find(*i);
  if (j==ipv4u.end()) continue;
  nlri.push_back(*i);
 }
 if (nlri.empty()) {lock_release();return ;}
 cnet.push_back(nlri[0]);
 for(i=nlri.begin();i!=nlri.end();i++){
  j=ipv4u.find(*i);
  //ipv4u.erase(j);
  if (!j->second.bWithdrawn){
   cnet[0]=*i;
   if (!bAsync) log->withdraw(peer,j->second,cnet);
   j->second.bWithdrawn=true;//invalidate
   j->second.bTouch=true;
  }
 }
 lock_release();
 //log->withdraw(peer,attr,nlri);
}
void BGPLogProxy::nlri_ipv4m(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPNets & nets){
 if (!log) return ;
 if (!lock_acquire()) return ;
 IPNets nlri;
 IPNets::const_iterator i;
 BGPIPTable::iterator j;
 MSG_UpdateAttr attrm=attr;
 attrm.aNextHop=nexthop;
 for(i=nets.begin();i!=nets.end();i++){
  j=ipv4m.find(*i);
  if (j!=ipv4m.end()){
   if (j->second==attrm) continue;
  }
  nlri.push_back(*i);
 }
 if (nlri.empty()) {lock_release();return ;}
 for(i=nlri.begin();i!=nlri.end();i++){
  ipv4m[*i]=attrm;
 }
 if (!bAsync) log->nlri_ipv4m(peer,attr,nexthop,nlri);
 lock_release();
}
void BGPLogProxy::withdraw_ipv4m(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPNets & nets){
 if (!log) return ;
 if (!lock_acquire()) return ;
 IPNets nlri,cnet;
 IPNets::const_iterator i;
 BGPIPTable::iterator j;
 for(i=nets.begin();i!=nets.end();i++){
  j=ipv4m.find(*i);
  if (j==ipv4m.end()) continue;
  nlri.push_back(*i);
 }
 if (nlri.empty()) {lock_release();return ;}
 cnet.push_back(nlri[0]);
 for(i=nlri.begin();i!=nlri.end();i++){
  j=ipv4m.find(*i);
  //ipv4m.erase(j);
  if (!j->second.bWithdrawn){
   cnet[0]=*i;
   j->second.bWithdrawn=true;//invalidate
   j->second.bTouch=true;
   if (!bAsync) log->withdraw_ipv4m(peer,j->second,j->second.aNextHop,cnet);
  }
 }
 lock_release();
}
void BGPLogProxy::nlri_ipv4lu(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
 if (!log) return ;
 if (!lock_acquire()) return ;
 IPTagNets nlri;
 IPTagNets::const_iterator i;
 BGPIPTagTable::iterator j;
 MSG_UpdateAttr attrm=attr;
 attrm.aNextHop=nexthop;
 for(i=nets.begin();i!=nets.end();i++){
  j=ipv4l.find(*i);
  if (j!=ipv4l.end()){
   if (j->second==attrm) continue;
  }
  nlri.push_back(*i);
 }
 if (nlri.empty()) {lock_release();return ;}
 for(i=nlri.begin();i!=nlri.end();i++){
  ipv4l[*i]=attrm;
 }
 if (!bAsync) log->nlri_ipv4lu(peer,attr,nexthop,nlri);
 lock_release();
}
void BGPLogProxy::withdraw_ipv4lu(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
 if (!log) return ;
 if (!lock_acquire()) return ;
 IPTagNets nlri,cnet;
 IPTagNets::const_iterator i;
 BGPIPTagTable::iterator j;
 for(i=nets.begin();i!=nets.end();i++){
  j=ipv4l.find(*i);
  if (j==ipv4l.end()) continue;
  nlri.push_back(*i);
 }
 if (nlri.empty()) {lock_release();return ;}
 cnet.push_back(nlri[0]);
 for(i=nlri.begin();i!=nlri.end();i++){
  j=ipv4l.find(*i);
  //ipv4l.erase(j);
  if (!j->second.bWithdrawn){
   cnet[0]=*i;
   j->second.bWithdrawn=true;//invalidate
   j->second.bTouch=true;
   if (!bAsync) log->withdraw_ipv4lu(peer,j->second,j->second.aNextHop,cnet);
  }
 }
 lock_release();
 //log->withdraw_ipv4lu(peer,attr,nexthop,nlri);
}
void BGPLogProxy::nlri_vpnv4u(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
 if (!log) return ;
 if (!lock_acquire()) return ;
 IPTagNets nlri;
 IPTagNets::const_iterator i;
 BGPIPTagTable::iterator j;
 MSG_UpdateAttr attrm=attr;
 attrm.aNextHop=nexthop;
 for(i=nets.begin();i!=nets.end();i++){
  j=vpnv4u.find(*i);
  if (j!=vpnv4u.end()){
   if (j->second==attrm) continue;
  }
  nlri.push_back(*i);
 }
 if (nlri.empty()) {lock_release();return ;}
 for(i=nlri.begin();i!=nlri.end();i++){
  vpnv4u[*i]=attrm;
 }
 if (!bAsync) log->nlri_vpnv4u(peer,attr,nexthop,nlri);
 lock_release();
}
void BGPLogProxy::withdraw_vpnv4u(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
 if (!log) return ;
 if (!lock_acquire()) return ;
 IPTagNets nlri,cnet;
 IPTagNets::const_iterator i;
 BGPIPTagTable::iterator j;
 for(i=nets.begin();i!=nets.end();i++){
  j=vpnv4u.find(*i);
  if (j==vpnv4u.end()) continue;
  nlri.push_back(*i);
 }
 if (nlri.empty()) {lock_release();return ;}
 cnet.push_back(nlri[0]);
 for(i=nlri.begin();i!=nlri.end();i++){
  j=vpnv4u.find(*i);
  //vpnv4u.erase(j);
  if (!j->second.bWithdrawn){
   cnet[0]=*i;
   j->second.bWithdrawn=true;//invalidate
   j->second.bTouch=true;
   if (!bAsync) log->withdraw_vpnv4u(peer,j->second,j->second.aNextHop,cnet);
  }
 }
 lock_release();
 //log->withdraw_vpnv4u(peer,attr,nexthop,nlri);
}

void BGPLogProxy::nlri_ipv6u(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPNets & nets){
 if (!log) return ;
 if (!lock_acquire()) return ;
 IPNets nlri;
 IPNets::const_iterator i;
 BGPIPTable::iterator j;
 MSG_UpdateAttr attrm=attr;
 attrm.aNextHop=nexthop;
 for(i=nets.begin();i!=nets.end();i++){
  j=ipv6u.find(*i);
  if (j!=ipv6u.end()){
   if (j->second==attrm) continue;
  }
  nlri.push_back(*i);
 }
 if (nlri.empty()) {lock_release();return ;}
 for(i=nlri.begin();i!=nlri.end();i++){
  ipv6u[*i]=attrm;
 }
 if (!bAsync) log->nlri_ipv6u(peer,attr,nexthop,nlri);
 lock_release();
}

void BGPLogProxy::withdraw_ipv6u(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPNets & nets){
 if (!log) return ;
 if (!lock_acquire()) return ;
 IPNets nlri,cnet;
 IPNets::const_iterator i;
 BGPIPTable::iterator j;
 for(i=nets.begin();i!=nets.end();i++){
  j=ipv6u.find(*i);
  if (j==ipv6u.end()) continue;
  nlri.push_back(*i);
 }
 if (nlri.empty()) {lock_release();return ;}
 cnet.push_back(nlri[0]);
 for(i=nlri.begin();i!=nlri.end();i++){
  j=ipv6u.find(*i);
  //ipv4u.erase(j);
  if (!j->second.bWithdrawn){
   cnet[0]=*i;
   j->second.bWithdrawn=true;//invalidate
   j->second.bTouch=true;
   if (!bAsync) log->withdraw_ipv6u(peer,j->second,nexthop,cnet);
  }
 }
 lock_release();
}
void BGPLogProxy::nlri_ipv6lu(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
 if (!log) return ;
 if (!lock_acquire()) return ;
 IPTagNets nlri;
 IPTagNets::const_iterator i;
 BGPIPTagTable::iterator j;
 MSG_UpdateAttr attrm=attr;
 attrm.aNextHop=nexthop;
 for(i=nets.begin();i!=nets.end();i++){
  j=ipv6l.find(*i);
  if (j!=ipv6l.end()){
   if (j->second==attrm) continue;
  }
  nlri.push_back(*i);
 }
 if (nlri.empty()) {lock_release();return ;}
 for(i=nlri.begin();i!=nlri.end();i++){
  ipv6l[*i]=attrm;
 }
 if (!bAsync) log->nlri_ipv6lu(peer,attr,nexthop,nlri);
 lock_release();
}
void BGPLogProxy::withdraw_ipv6lu(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
 if (!log) return ;
 if (!lock_acquire()) return ;
 IPTagNets nlri,cnet;
 IPTagNets::const_iterator i;
 BGPIPTagTable::iterator j;
 for(i=nets.begin();i!=nets.end();i++){
  j=ipv6l.find(*i);
  if (j==ipv6l.end()) continue;
  nlri.push_back(*i);
 }
 if (nlri.empty()) {lock_release();return ;}
 cnet.push_back(nlri[0]);
 for(i=nlri.begin();i!=nlri.end();i++){
  j=ipv6l.find(*i);
  //ipv4l.erase(j);
  if (!j->second.bWithdrawn){
   cnet[0]=*i;
   j->second.bWithdrawn=true;//invalidate
   j->second.bTouch=true;
   if (!bAsync) log->withdraw_ipv6lu(peer,j->second,j->second.aNextHop,cnet);
  }
 }
 lock_release();
}

void BGPLogProxy::nlri_vpnv6u(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
 if (!log) return ;
 if (!lock_acquire()) return ;
 IPTagNets nlri;
 IPTagNets::const_iterator i;
 BGPIPTagTable::iterator j;
 MSG_UpdateAttr attrm=attr;
 attrm.aNextHop=nexthop;
 for(i=nets.begin();i!=nets.end();i++){
  j=vpnv6u.find(*i);
  if (j!=vpnv6u.end()){
   if (j->second==attrm) continue;
  }
  nlri.push_back(*i);
 }
 if (nlri.empty()) {lock_release();return ;}
 for(i=nlri.begin();i!=nlri.end();i++){
  vpnv6u[*i]=attrm;
 }
 if (!bAsync) log->nlri_vpnv6u(peer,attr,nexthop,nlri);
 lock_release();
}
void BGPLogProxy::withdraw_vpnv6u(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
 if (!log) return ;
 if (!lock_acquire()) return ;
 IPTagNets nlri,cnet;
 IPTagNets::const_iterator i;
 BGPIPTagTable::iterator j;
 for(i=nets.begin();i!=nets.end();i++){
  j=vpnv6u.find(*i);
  if (j==vpnv6u.end()) continue;
  nlri.push_back(*i);
 }
 if (nlri.empty()) {lock_release();return ;}
 cnet.push_back(nlri[0]);
 for(i=nlri.begin();i!=nlri.end();i++){
  j=vpnv6u.find(*i);
  //vpnv4u.erase(j);
  if (!j->second.bWithdrawn){
   cnet[0]=*i;
   j->second.bWithdrawn=true;//invalidate
   j->second.bTouch=true;
   if (!bAsync) log->withdraw_vpnv6u(peer,j->second,j->second.aNextHop,cnet);
  }
 }
 lock_release();
}

void BGPLogProxy::nlri_l2vpn(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
 if (!log) return ;
 if (!lock_acquire()) return ;
 IPTagNets nlri;
 IPTagNets::const_iterator i;
 BGPIPTagTable::iterator j;
 MSG_UpdateAttr attrm=attr;
 attrm.aNextHop=nexthop;
 for(i=nets.begin();i!=nets.end();i++){
  j=l2vpn.find(*i);
  if (j!=l2vpn.end()){
   if (j->second==attrm) continue;
  }
  nlri.push_back(*i);
 }
 if (nlri.empty()) {lock_release();return ;}
 for(i=nlri.begin();i!=nlri.end();i++){
  l2vpn[*i]=attrm;
 }
 if (!bAsync) log->nlri_l2vpn(peer,attr,nexthop,nlri);
 lock_release();
}
void BGPLogProxy::withdraw_l2vpn(BGPPeer *peer,const MSG_UpdateAttr &attr,const IPAddress & nexthop,const IPTagNets & nets){
 if (!log) return ;
 if (!lock_acquire()) return ;
 IPTagNets nlri,cnet;
 IPTagNets::const_iterator i;
 BGPIPTagTable::iterator j;
 for(i=nets.begin();i!=nets.end();i++){
  j=l2vpn.find(*i);
  if (j==l2vpn.end()) continue;
  nlri.push_back(*i);
 }
 if (nlri.empty()) {lock_release();return ;}
 cnet.push_back(nlri[0]);
 for(i=nlri.begin();i!=nlri.end();i++){
  j=l2vpn.find(*i);
  //vpnv4u.erase(j);
  if (!j->second.bWithdrawn){
   cnet[0]=*i;
   j->second.bWithdrawn=true;//invalidate
   j->second.bTouch=true;
   if (!bAsync) log->withdraw_l2vpn(peer,j->second,j->second.aNextHop,cnet);
  }
 }
 lock_release();
}
