#ifndef __bgplogproxy_h_
#define __bgplogproxy_h_
#include "bgplog.h"
#include <map>
#include <pthread.h>

namespace BGPHist {

class BGPItemAttr: public MSG_UpdateAttr{
public:
    bool bTouch,bWithdrawn;
    BGPItemAttr();
    BGPItemAttr(const BGPItemAttr &);
    BGPItemAttr(const MSG_UpdateAttr &);
    BGPItemAttr(const MSG_UpdateAttr &,bool );
};
typedef std::map<IPNet,BGPItemAttr> BGPIPTable;
typedef std::map<IPTagNet,BGPItemAttr> BGPIPTagTable;
    
class BGPLogProxy: public BGPLog {
protected:
    bool bAsync,bExit;
    unsigned uWait;
    BGPPeer *_peer;
    pthread_mutex_t uLock;
    pthread_mutexattr_t uAttr;
    pthread_t uThread;
    BGPLog * log;
    BGPIPTable ipv4u;
    BGPIPTable ipv4m;
    BGPIPTagTable ipv4l;
    BGPIPTagTable vpnv4u;
    BGPIPTable ipv6u;
    BGPIPTagTable ipv6l;
    BGPIPTagTable vpnv6u;
    BGPIPTagTable l2vpn;
    bool lock_acquire();
    bool lock_release();
public:
    BGPLogProxy(BGPLog *_log,unsigned _asyncwait=0);
    void async_proc();
    std::string get_stats();
    virtual ~BGPLogProxy();
    virtual void open(BGPPeer *peer,const MSG_Open *);
    virtual void close(BGPPeer *peer);
    virtual void txt(BGPPeer *peer,const char *);
    virtual void exception(BGPPeer *peer,const BGPException *);
    virtual void notification(BGPPeer *peer,const MSG_Notify *);
    virtual void keepalive(BGPPeer *peer);
    virtual void nlri(BGPPeer *peer,const MSG_UpdateAttr &,const IPNets & );
    virtual void withdraw(BGPPeer *peer,const MSG_UpdateAttr &,const IPNets & );
    virtual void nlri_ipv4m(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPNets & );
    virtual void withdraw_ipv4m(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPNets & );
    virtual void nlri_ipv4lu(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    virtual void withdraw_ipv4lu(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    virtual void nlri_vpnv4u(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    virtual void withdraw_vpnv4u(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    virtual void nlri_ipv6u(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPNets & );
    virtual void withdraw_ipv6u(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPNets & );
    virtual void nlri_ipv6lu(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    virtual void withdraw_ipv6lu(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    virtual void nlri_vpnv6u(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    virtual void withdraw_vpnv6u(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    virtual void nlri_l2vpn(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
    virtual void withdraw_l2vpn(BGPPeer *peer,const MSG_UpdateAttr &,const IPAddress & nexthop,const IPTagNets & );
};

};

#endif
