extern "C"{
#include <sys/stat.h>
#include <netinet/in.h>
#include <unistd.h>
}
#include "bgppeer.h"
#include "bgplog.h"

using namespace std;
using namespace BGPHist;

BGPPeer::BGPPeer(BGPLog *_log):log(_log),state(0),uTimeout(3),uHoldTime(180),uRouterId(0x01010101),asnMy(65534),asnPeer(65534),bIs32bitASN(true),bIsIPV6(false),
iFD(-1),bFDMode(false),bFEof(false)
{
     sock.sockopts.uOpt_keepalive=1;
     sock.sockopts.uOpt_sndbuf=8192;
     sock.sockopts.uOpt_rcvbuf=8192;
}
BGPPeer::~BGPPeer(){
    close();
}
void BGPPeer::playfile(int _fd){
    close();
    bFDMode=true;
    iFD=_fd;
    bFEof=false;
}

bool BGPPeer::connect(){
    char szAddr[256];
    close();
#ifdef INET6
    if (sPeer.find(":")!=std::string::npos){
        snprintf(szAddr,sizeof(szAddr),"tcp6:%s.179",sPeer.c_str());
        bIsIPV6=true;
    } else
#endif        
        snprintf(szAddr,sizeof(szAddr),"tcp4:%s:179",sPeer.c_str());
    return sock.openSocket(szAddr,somConnect);
}
void BGPPeer::connectTo(BGPLog *log,const char * addr,BGPASN asn){
    BGPPeer * peer=new BGPPeer(log);
    try {
        peer->asnPeer=asn;
        peer->asnMy=asn;
        peer->sPeer=addr;
        if (peer->connect())
            peer->lifecycle();
    }
    catch (const BGPException & e){
        log->exception(0,&e);
        cout << "caught BGPException " << e.what() << endl;
  }
  catch (const SocketException & e) {
     log->txt(0,e.what());
     cout << "Connecting to " << addr << "caught sockerr " << e.what() << endl;
  }
  catch(const std::exception & s){
     cout << "caught " << s.what() << endl;
  }
  delete peer;
}

void BGPPeer::listen(BGPLog *log,unsigned myaddr,BGPASN myasn){
    try {
     cout << "bind" << endl;
     Socket sock;
     sock.sockopts.uOpt_keepalive=1;
     sock.sockopts.uOpt_sndbuf=8192;
     sock.sockopts.uOpt_rcvbuf=8192;
     sock.openSocket("tcp4:0.0.0.0:179",somBind);
     for (;;) {
      Socket sAcc;
      if (!sock.accept(sAcc)) return ;
      pid_t p=fork();
      if (p==0){
       BGPPeer * peer=new BGPPeer(log);
       peer->asnMy=myasn;
       peer->sock=sAcc;
       peer->lifecycle();
       delete peer;
       return ;
      }
     }
    }
    catch (const BGPException & e){
        log->exception(0,&e);
        cout << "caught BGPException " << e.what() << endl;
    }
    catch (const SocketException & s) {
     log->txt(0,s.what());
     cout << "caught sockerr " << s.what() << endl;
    }
    catch(const std::exception & s){
     cout << "caught " << s.what() << endl;
    }
    return ;
}
bool BGPPeer::eof(){
    if (!bFDMode) return sock.eof();
    if (iFD<0) return true;
    return bFEof;
}

void BGPPeer::close(){
    try {
        if (!bFDMode) sock.close();
        iFD=-1;
        bFDMode=false;
        bFEof=true;
        log->close(this);
  }
  catch (...){
                // if the client break the connection, then
                // the shutdown method throws sockerr
  }
}
    
void BGPPeer::lifecycle(){
 time_t t0,t1;
 unsigned uKeepalive;
 try {
  unsigned char c;
  char cOpen[]={255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,//marker
                0,69+32,//length big endian
                1,//type
                4,//version
                0x0,0x1,//my ASN = 1, big endian
                0,180,//hold time, big endian
                1,1,1,1,//bgp router id
                72,//40,//rest length
                2,6,1,4,0,1,0,1,//multiproto capability ipv4 u
                2,6,1,4,0,1,0,4,//multiproto capability ipv4 m
                2,6,1,4,0,1,0,128,//multiproto capability vpnv4 u
                2,6,1,4,0,1,0,2,//multiproto capability ipv4 lu
                2,6,1,4,0,2,0,1,//multiproto capability ipv6 u
                2,6,1,4,0,2,0,4,//multiproto capability ipv6 lu
                2,6,1,4,0,2,0,128,//multiproto capability vpnv6 u
                2,6,1,4,0,25,0,65,//vpls
                2,6,65,4,0,0,0xac,0xcd//32-bit asn
                //2,2,128,0,
                //2,2,2,0
  };
  cOpen[20]=(asnMy >> 8) & 0xff;
  cOpen[21]=(asnMy) & 0xff;
  cOpen[22]=(uHoldTime >> 8) & 0xff;
  cOpen[23]=uHoldTime & 0xff;
  memcpy(&cOpen[24],&uRouterId,4);
  c=28+1+cOpen[28]-4;
  cOpen[c]=(asnMy >> 24) & 0xff;
  cOpen[c+1]=(asnMy >> 16) & 0xff;
  cOpen[c+2]=(asnMy >> 8) & 0xff;
  cOpen[c+3]=asnMy & 0xff;
  char cKeepalive[]={255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,//marker
         0,0x13,//length big endian
  4};
  if (!bFDMode) sock.Write(cOpen,sizeof(cOpen));
  read_bgpmsg();
  if (!bFDMode) sock.Write(cKeepalive,sizeof(cKeepalive));
  t0=time(0);
  uKeepalive=uHoldTime/3;
  while(!eof()){
   try {
    read_bgpmsg();
   }
   catch(const BGPTimeout &s){  }
   catch(const BGPException &s){
    cout << "caught BGPException " << s.what() << endl;
   }
   if (eof()) break;
   if (!bFDMode){
    t1=time(0);
    if (t1>(t0+uKeepalive)){
     if (sock.is_writeready(1)){
      sock.Write(cKeepalive,sizeof(cKeepalive));
     }
     t0=t1;
    }
   }
  }
 }
 catch (const BGPException &s){
                cout << "caught BGPException " << s.what() << endl;
 }
 catch (const SocketException & s) {
                cout << "caught sockerr " << s.what() << endl;
 }
 catch (int t) {
                cout << "caught int " << t << endl;
 }
}

void BGPPeer::readdata(void *d,size_t sz){
    void *p;
    char *c,*c1;
    size_t i,j;
    if (bFDMode){
     c=(char *)d;
     j=read(iFD,d,sz);
     if (!j) {bFEof=true;throw BGPException(2,"eof");}
    } else {
     if (!sock.valid()) throw BGPException(2,"invalid socket");
     if (sock.eof()) throw BGPException(2,"eof");
     c=(char *)d;
     for(i=0;i<sz;){
      if (!sock.is_readready(uTimeout)) throw BGPTimeout();
      j=sock.Read(c,sz-i);
      if (!j) throw BGPException(2,"null read");
      c+=j;
      i+=j;
     }
    }
}


bool BGPPeer::read_bgpmsg(){
    unsigned char c,m_type;
    unsigned i,j,k,l,pt,info_len,n,tc,m,pl;
    unsigned short p_len;
    unsigned afi,safi,ps,c_snpa,nl,lbl;
    int ii;
    void *p;
    i=0;
    for(;!eof();){
      c=0;
      readdata(&c,1);
      if (c!=0xff) {
#ifdef DEBUG
        cout << hex << (unsigned)c << dec << endl;
#endif
        return false;
      }//throw BGPException(3);
      i++;
      if (i>15) break;
    }
    if (eof()) return false;//throw BGPException(2);
    readdata(&c,1);
    p_len=c;p_len=p_len << 8;
    readdata(&c,1);
    p_len|=c;
    readdata(&c,1);
    m_type=c;
#ifdef DEBUG    
    cout << endl << dec << "============" << endl;
    cout << "BGP message type = " << (unsigned) m_type << " length = " << p_len << " actual data len = " << (p_len-19) << hex << endl;
#endif    
    if (p_len<20) return true;
    info_len=p_len-19;
    unsigned char *buf=(unsigned char *)alloca(info_len);
    if (!buf) return false;
    readdata(buf,info_len);
#ifdef DEBUG    
    for(i=0;i<info_len;i++){
        cout << (unsigned) buf[i] << " ";
    }
    cout << endl << dec << "------------" << endl;
#endif    
    switch(m_type){
        case 1:{
            MSG_Open msg;
            msg.version=(unsigned)buf[0];
            msg.asn16=(buf[1] << 8 | buf[2]);
            msg.holdtimer=(buf[3] << 8 | buf[4]);
            msg.aRouterId.assignIPv4(&buf[5]);
            uHoldTime=msg.holdtimer;
#ifdef DEBUG    
            cout << "OPEN Message" << endl;
            cout << "Version " << (unsigned)buf[0] << endl;
            cout << "ASN " << (buf[1] << 8 | buf[2]) << endl;
            cout << "HoldTimer " << (buf[3] << 8 | buf[4]) << endl;
            cout << "RouterID " << (unsigned)buf[5] << "." << (unsigned)buf[6] << "." << (unsigned)buf[7] << "." << (unsigned)buf[8] << endl;
#endif    
            if(buf[9]>0) {
                for(i=0;i<buf[9];i+=buf[11+i]+2){
#ifdef DEBUG    
                    cout << "param type " << (unsigned)buf[10+i] << " length " << (unsigned)buf[11+i] << endl;
#endif                            
                    if (buf[11+i]>(buf[9]-i)) {
                        //cout << "invalid length" << endl;
                        break;
                    }
#ifdef DEBUG    
                    cout << hex << " hex dump: ";for(j=0;j<buf[11+i];j++) cout << (unsigned)buf[12+i+j] << " ";cout << dec << endl;
#endif                    
                    switch(buf[10+i]){
                        case 1:{
                            //cout << " auth param" << endl;
                            break;
                        }
                        case 2:{
#ifdef DEBUG    
                            cout << " capability: code " << (unsigned)buf[12+i] << " length " << (unsigned)buf[13+i] << endl;
#endif                            
                            BGPCapabiliy cap;cap.code=(unsigned)buf[12+i];cap.length=(unsigned)buf[13+i];
                            if (buf[13+i]>(buf[11+i]-2)) {
                                //cout << "invalid length" << endl;
                                break;
                            }
#ifdef DEBUG    
                            cout << hex << "   hex dump: ";for(j=0;j<buf[13+i];j++) cout << (unsigned)buf[14+i+j] << " ";cout << dec << endl;
#endif                            
                            memcpy(&cap.data,&buf[14+i],std::min<size_t>((size_t)sizeof(cap.data),(size_t)buf[13+i]));
                            msg.caps.push_back(cap);
                            break;
                        }
                    }
                }
            }
            //else cout << "no optional params" << endl;
            log->open(this,&msg);
            break;
        }
        case 2:{
            try {
#ifdef DEBUG    
            cout << "UPDATE Message" << endl;
#endif    
            MSG_UpdateAttr attrs;
            attrs.clear();
            pt=0;
            uint32_t u32;
            //withdraw routes
            IPNets withdraw;
            k=be16(&buf[pt]);pt+=2;
            for(i=0;i<k;){
                //handle withdraws
                //cout << "withdraw # " << (i+1) << " : ";
                pl=buf[pt];
                if (bIsIPV6){
                    withdraw.push_back(IPNet().assignNLRI6(pl,&buf[pt+1]));
                } else {
                    withdraw.push_back(IPNet().assignNLRI4(pl,&buf[pt+1]));
                }
                pt+=((pl+7)/8)+1;
                i+=((pl+7)/8)+1;
            }
            //path attributes
            k=be16(&buf[pt]);pt+=2;
            //k=k & 0xff;//???
#ifdef DEBUG    
            cout << " PA block length " << k << endl;
#endif
            k+=pt;
            if (k>p_len) throw BGPException(Sprintf("Incorrect PA block length %u>%u",(unsigned)k,(unsigned)p_len));
            for(n=1;pt<k;n++){
                //handle path attributes
                tc=buf[pt+1];
#ifdef DEBUG    
                cout << "PA # " << n << " : flags=" << hex << (unsigned)buf[pt] << "(";
                cout << ((buf[pt] & 128) ? "optional ":"well-known ") << ((buf[pt] & 64) ? "transitive ":"non-transitive ") << ((buf[pt] & 32) ? "partial ":"complete ") << ((buf[pt] & 16) ? "long ":"short ") << ")";
                cout << " tc=" << tc << dec;
#endif
                
                if (buf[pt] & 16) {
                    l=be16(&buf[pt+2]);
                    pt+=4;
                    //if (l>2) l-=2; else l=0;
                } else {
                    l=buf[pt+2];
                    pt+=3;
                    //if (l>1) l-=1; else l=0;
                }
                //l=l & 0xff;//???
#ifdef DEBUG    
                cout << " len=" << l;
#endif
                if (l>(p_len-pt)) throw BGPException(Sprintf("Incorrect PA item length %d>%d",(int)l,(int)(p_len-pt)));
#ifdef DEBUG
                for(j=0;j<l;j++){
                    cout << " " << (unsigned)buf[pt+j];
                }
                cout << endl << "-";
#endif
                switch(tc){
                    case 1:{
                        //cout << "ORIGIN " << (unsigned) buf[pt] << endl;
                        attrs.origin=buf[pt];
                        break;
                    }
                    case 2:{
                        attrs.vASPath.clear();
                        //cout << "ASPATH ";
                        for(j=0;j<l;){
                            //cout << "[" << (unsigned)buf[pt+j] << "]";
                            j++;
                            m=buf[pt+j];j++;//size
                            for(n=0;n<m;n++){
                                if (bIs32bitASN) {
                                    //cout << " " << be32(&buf[pt+j]);
                                    attrs.vASPath.push_back(be32(&buf[pt+j]));
                                    j+=4;
                                } else {
                                    //cout << " " << be16(&buf[pt+j]);
                                    attrs.vASPath.push_back(be16(&buf[pt+j]));
                                    j+=2;
                                }
                            }
                        }
                        //cout << endl;
                        break;
                    }
                    case 3:{
#ifdef DEBUG
                        cout << "NEXTHOP";for(j=0;j<l;j++) cout << (j ? ".":" ") << (unsigned)buf[pt+j];cout << endl;
#endif
                        attrs.aNextHop.clear();
                        if (l>3) attrs.aNextHop.assign(&buf[pt],l);
                        break;
                    }
                    case 4:{
                        //cout << "MED " << be32(&buf[pt]) << endl;
                        attrs.med=be32(&buf[pt]);
                        break;
                    }
                    case 5:{
                        //cout << "LOCALPREF " << be32(&buf[pt]) << endl;
                        attrs.localpref=be32(&buf[pt]);
                        break;
                    }
                    case 6:{
                        //cout << "ATOMIC-AGGREGATE ";for(j=0;j<l;j++) cout << (j ? ".":" ") << (unsigned)buf[pt+j+n];cout  << endl;
                        attrs.aAtomicAggregate.clear();
                        if (l>3) attrs.aAtomicAggregate.assign(&buf[pt+n],l);
                        break;
                    }
                    case 7:{
                        //cout << "AGGREGATOR AS ";for(j=0;j<l;j++) cout << (j ? ".":" ") << (unsigned)buf[pt+j+n];cout << endl;
                        if (bIs32bitASN){
                            if (l!=8) break;
                            attrs.aAggregatorIP.assign(&buf[pt+4],4);
                            attrs.aAggregatorAS=be32(&buf[pt]);
                        } else {
                            if (l!=6) break;
                            attrs.aAggregatorIP.assign(&buf[pt+2],4);
                            attrs.aAggregatorAS=be16(&buf[pt]);
                        }
                        break;
                    }
                    case 9:{
                        //cout << "ORIGINATOR-ID ";for(j=0;j<l;j++) cout << (j ? ".":" ") << (unsigned)buf[pt+j];cout  << endl;
                        attrs.aOriginator.clear();
                        if (l>3) attrs.aOriginator.assign(&buf[pt],l);
                        break;
                    }
                    case 8:{
#ifdef DEBUG
                        cout << "COMMUNITY ";
                        for(j=0;j<l;j++) cout << (j ? ".":" ") << (unsigned)buf[pt+j];
                        cout  << endl;
#endif
                        attrs.vCom.clear();
                        for(j=0;j<l;j+=4) attrs.vCom.push_back(be32(&buf[pt+j]));
                        break;
                    }
                    case 10:{
                        for(j=0;j<l;j+=4) attrs.vaClusterList.push_back(IPAddress().assignIPv4(&buf[pt+j]));
                        //cout << "CLUSTER-LIST ";for(j=0;j<l;j++) cout << (j ? ".":" ") << (unsigned)buf[pt+j];cout  << endl;
                        break;
                    }
                    case 14:{
                        afi=be16(&buf[pt]);
                        safi=buf[pt+2];
                        IPAddress nexthop;
#ifdef DEBUG
                        cout << "MP_REACH_NLRI afi(" << afi << ") safi(" << safi <<") ";
#endif
                        ps=4+buf[pt+3];
                        if ((afi==2) && (safi==128)){//vpnv6 unicast
                            nexthop.assignIPv6RD(&buf[pt+4]);
#ifdef DEBUG
                          cout << "vpnv4 unicast nexthop " ;
                          for(j=0;j<buf[pt+3];j++) cout << (j ? ".":" ") << (unsigned)buf[pt+j+4];
#endif
                        } else
                        if ((afi==1) && (safi==128)){//vpnv4 unicast
                            nexthop.assignIPv4RD(&buf[pt+4]);
#ifdef DEBUG
                          cout << "vpnv4 unicast nexthop rd " << be32(&buf[pt+4]) << ":" << be32(&buf[pt+8]) << " ip" ;
                          for(j=0;j<4;j++) cout << (j ? ".":" ") << (unsigned)buf[pt+j+12];
#endif
                        } else
                        if ((afi==1) && (safi==4) && (buf[pt+3]==4)){//ipv4 labeled-unicast
                            nexthop.assignIPv4(&buf[pt+4]);
#ifdef DEBUG
                          cout << "ipv4 labeled-unicast nexthop ip" ;
                          for(j=0;j<4;j++) cout << (j ? ".":" ") << (unsigned)buf[pt+j+4];
#endif
                        } else
                        if ((afi==1) && (safi==2)){//ipv4 multicast
                            nexthop.assignIPv4(&buf[pt+4]);
#ifdef DEBUG
                          cout << "ipv4 multicast nexthop ip" ;
                          for(j=0;j<4;j++) cout << (j ? ".":" ") << (unsigned)buf[pt+j+4];
#endif
                        } else
                        if ((afi==2) && (safi==1)){//ipv6 unicast
                            nexthop.assignIPv6(&buf[pt+4]);
#ifdef DEBUG
                          cout << "ipv6 unicast nexthop ip" ;
                          for(j=0;j<16;j++) cout << (j ? ".":" ") << (unsigned)buf[pt+j+4];
#endif
                        } else
                        if ((afi==2) && (safi==4)){//ipv6 labeled-unicast
                            nexthop.assignIPv6(&buf[pt+4]);
#ifdef DEBUG
                          cout << "ipv6 labeled-unicast nexthop ip" ;
                          for(j=0;j<16;j++) cout << (j ? ".":" ") << (unsigned)buf[pt+j+4];
#endif
                        }
                        else
                        if((afi==25) && (safi==65)){//l2vpn
                            if (buf[pt+3]==4)
                                nexthop.assignIPv4(&buf[pt+4]);
                        }
#ifdef DEBUG
                        else
                             cout << "nh_unknown (" << afi << "," << safi << ")";
                        cout << endl;
#endif
                        attrs.aNextHop=nexthop;
                        c_snpa=buf[pt+ps];
                        ps++;
#ifdef DEBUG
                        cout << "SNPA count " << c_snpa << endl;
#endif
                        for(k=0;k<c_snpa;k++){
#ifdef DEBUG
                            cout << "SNPA#" << (k+1) << " len=" << (unsigned)buf[pt+ps];
                            for(j=0;j<buf[pt+ps];j++) cout << " " << (unsigned)buf[pt+ps+1+j];
                            cout << endl;
#endif
                            ps+=1+buf[pt+ps];
                        }
                        //nlri
#ifdef DEBUG
                        cout << "NLRI dump [" << (l-ps) << "]:" << endl;
                        for(j=0;j<(l-ps);j++) cout << " " << (unsigned)buf[pt+ps+j];
                        cout << endl;
#endif
                        if((afi==2) && (safi==4)){//ipv6 labeled-unicast
                            IPTagNets nlri;
                            while(ps<l){
                                nl=buf[pt+ps];
                                nlri.push_back(IPTagNet().assignNLRI6(nl,&buf[pt+ps+1]));
                                ps+=((nl+7)/8)+1;
                            }
                            log->nlri_ipv6lu(this,attrs,nexthop,nlri);
                        }
                        if((afi==2) && (safi==1)){//ipv6 unicast
                            IPNets nlri;
                            n=1;
                            while(ps<l){
                                pl=buf[pt+ps];
                                if (pl>128) throw BGPException(Sprintf("Incorrect prefix length %d>128",(int)l));
                                nlri.push_back(IPNet().assignNLRI6(pl,&buf[pt+ps+1]));
                                ps+=((pl+7)/8)+1;
                                n++;
                            }
                            log->nlri_ipv6u(this,attrs,nexthop,nlri);
                        }
                        if((afi==1) && (safi==2)){//ipv4 multicast
                            IPNets nlri;
                            n=1;
                            while(ps<l){
                                pl=buf[pt+ps];
                                if (pl>32) throw BGPException(Sprintf("Incorrect prefix length %d>32",(int)l));
                                nlri.push_back(IPNet().assignNLRI4(pl,&buf[pt+ps+1]));
                                ps+=((pl+7)/8)+1;
                                n++;
                            }
                            log->nlri_ipv4m(this,attrs,nexthop,nlri);
                        }
                        if((afi==1) && (safi==4)){//ipv4 labeled-unicast
                            IPTagNets nlri;
                            while(ps<l){
                                nl=buf[pt+ps];
                                nlri.push_back(IPTagNet().assignNLRI4(nl,&buf[pt+ps+1]));
                                ps+=((nl+7)/8)+1;
                            }
                            log->nlri_ipv4lu(this,attrs,nexthop,nlri);
                        }
                        if((afi==1) && (safi==128)){//vpnv4 unicast
                            IPTagNets nlri;
                            while(ps<l){
                                nl=buf[pt+ps];
                                nlri.push_back(IPTagNet().assignNLRIRD4(nl,&buf[pt+ps+1]));
                                ps+=((nl+7)/8)+1;
                            }
                            log->nlri_vpnv4u(this,attrs,nexthop,nlri);
                        }
                        if((afi==2) && (safi==128)){//vpnv6 unicast
                            IPTagNets nlri;
                            while(ps<l){
                                nl=buf[pt+ps];
                                nlri.push_back(IPTagNet().assignNLRIRD6(nl,&buf[pt+ps+1]));
                                ps+=((nl+7)/8)+1;
                            }
                            log->nlri_vpnv6u(this,attrs,nexthop,nlri);
                        }
                        if((afi==25) && (safi==65)){//l2vpn vpls-vpws
                            IPTagNets nlri;
                            while(ps<l){
                                nl=be16(&buf[pt+ps]);
                                nlri.push_back(IPTagNet().assignNLRIL2(nl,&buf[pt+ps+2]));
                                ps+=nl+2;
                            }
                            log->nlri_l2vpn(this,attrs,nexthop,nlri);
                        }
                        break;
                    }
                    case 15:{
                        afi=be16(&buf[pt]);
                        safi=buf[pt+2];
                        IPAddress nexthop;
                        ps=3;
#ifdef DEBUG
                        cout << "MP_UNREACH_NLRI afi(" << afi << ") safi(" << safi <<") " << endl;
                        cout << "UNLRI dump:" << endl;
                        for(j=0;j<(l-ps);j++) cout << " " << (unsigned)buf[pt+ps+j];
                        cout << endl;
#endif
                        if((afi==2) && (safi==1)){//ipv6 unicast
                            IPNets nlri;
                            n=1;
                            while(ps<l-3){
                                pl=buf[pt+ps];
                                if (pl>128) break;
                                nlri.push_back(IPNet().assignNLRI6(pl,&buf[pt+ps+1]));
                                ps+=((pl+7)/8)+1;
                                n++;
                            }
                            log->withdraw_ipv6u(this,attrs,nexthop,nlri);
                        } else
                        if((afi==2) && (safi==4)){//ipv6 labeled-unicast
                            n=1;
                            IPTagNets nlri;
                            while(ps<l-3){
                                nl=buf[pt+ps];
                                nlri.push_back(IPTagNet().assignNLRI6(nl,&buf[pt+ps+1]));
                                ps+=((nl+7)/8)+1;
                                n++;
                            }
                            log->withdraw_ipv6lu(this,attrs,nexthop,nlri);
                        } else
                        if((afi==1) && (safi==2)){//ipv4 multicast
                            IPNets nlri;
                            n=1;
                            while(ps<l-3){
                                pl=buf[pt+ps];
                                nlri.push_back(IPNet().assignNLRI4(pl,&buf[pt+ps+1]));
                                ps+=((pl+7)/8)+1;
                                n++;
                            }
                            log->withdraw_ipv4m(this,attrs,nexthop,nlri);
                        } else
                        if((afi==1) && (safi==4)){//ipv4 labeled-unicast
                            IPTagNets nlri;
                            while(ps<l-3){
                                nl=buf[pt+ps];
                                nlri.push_back(IPTagNet().assignNLRI4(nl,&buf[pt+ps+1]));
                                ps+=((nl+7)/8)+1;
                            }
                            log->withdraw_ipv4lu(this,attrs,nexthop,nlri);
                        } else
                        if((afi==1) && (safi==128)){//vpnv4 unicast
                            IPTagNets nlri;
                            while(ps<l-3){
                                nl=buf[pt+ps];
                                nlri.push_back(IPTagNet().assignNLRIRD4(nl,&buf[pt+ps+1]));
                                ps+=((nl+7)/8)+1;
                            }
                            log->withdraw_vpnv4u(this,attrs,nexthop,nlri);
                        } else
                        if((afi==2) && (safi==128)){//vpnv6 unicast
                            IPTagNets nlri;
                            while(ps<l-3){
                                nl=buf[pt+ps];
                                nlri.push_back(IPTagNet().assignNLRIRD6(nl,&buf[pt+ps+1]));
                                ps+=((nl+7)/8)+1;
                            }
                            log->withdraw_vpnv6u(this,attrs,nexthop,nlri);
                        } else
                        if((afi==25) && (safi==65)){//l2vpn
                            IPTagNets nlri;
                            while(ps<l){
                                nl=be16(&buf[pt+ps]);
                                nlri.push_back(IPTagNet().assignNLRIL2(nl,&buf[pt+ps+2]));
                                ps+=nl+2;
                            }
                            log->withdraw_l2vpn(this,attrs,nexthop,nlri);
                        } else
                            {
                            cout << "mp_unreach_unknown (" << afi << "," << safi << ")";
                        }
                        break;
                    }
                    case 16:{
                        BGPExtCommunity ec;
                        attrs.vExtCom.clear();
                        for(j=0;j<l;j+=8) {
                            ec.excType=be16(&buf[pt+j]);
                            ec.a=be16(&buf[pt+2+j]);
                            ec.b=be32(&buf[pt+4+j]);
                            attrs.vExtCom.push_back(ec);
                        }
                        /*
                        unsigned exct;
                        exct=be16(&buf[pt]);
                        cout << "EXTCOMM {";
                        switch(exct){
                            case 2:cout << "RT";break;
                            default:cout << exct;break;
                        }
                        cout  << "}" << be16(&buf[pt+2]) << ":" << be32(&buf[pt+4]) << endl;
                        */
                        break;
                    }
                    default:{
                        //cout << "UNKNOWN" << endl;
                        break;
                    }
                }
                pt+=l;
            }
            //cout << " PA block finished" << endl;
            //nlri
            IPNets nlri;
            n=1;
            while(pt<info_len){
                pl=buf[pt];
                if (bIsIPV6){
                    nlri.push_back(IPNet().assignNLRI6(pl,&buf[pt+1]));
                } else {
                    nlri.push_back(IPNet().assignNLRI4(pl,&buf[pt+1]));
                }
                pt+=((pl+7)/8)+1;
                n++;
            }
            if (bIsIPV6){
                log->withdraw_ipv6u(this,attrs,IPAddress(),withdraw);
                log->nlri_ipv6u(this,attrs,attrs.aNextHop,nlri);
            } else {
                log->withdraw(this,attrs,withdraw);
                log->nlri(this,attrs,nlri);
            }
            } catch(const BGPException &e){
                cerr << "UPDATE error: " << e.what() << endl << "Dump:" << endl;
                for(i=0;i<info_len;i++){
                    cout << (unsigned) buf[i] << " ";
                }
                cout << endl << "------------" << endl;
                throw;

            }
            break;
        };
        case 3:{
            //cout << "NOTIFICATION Message" << endl;
            MSG_Notify msg;
            log->notification(this,&msg);
            break;
        };
        case 4:{
#ifdef DEBUG    
            cout << "KEEPALIVE Message" << endl;
#endif
            log->keepalive(this);
            break;
        };
    }
    return true;
}

