#include "bgpsock.h"
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdarg.h>

namespace BGPHist {

//#define SETNONBLOCK(x) fcntl(x,F_SETFL,O_NONBLOCK)
#define SETNONBLOCK(x)
SocketException::SocketException(){
	errcode=errno;
	setStr(strerror(errcode));
}

SocketException::SocketException(const char *cmntfmt,...){
	errcode=errno;
	if (!cmntfmt) {setStr(strerror(errcode));return ;}
	const char * c=strerror(errcode);
	char cmnt[1024],szMsg[1024];
	va_list vl1;
	va_start(vl1,cmntfmt);
	vsnprintf(cmnt,sizeof(cmnt),cmntfmt,vl1);
	va_end(vl1);
	int i=(sizeof(szMsg)-1)<strlen(cmnt) ? (sizeof(szMsg)-1):strlen(cmnt);
	strncpy(szMsg,cmnt,i);szMsg[i]=0;
	int j=(sizeof(szMsg)-1-i)<strlen(c) ? (sizeof(szMsg)-1-i):strlen(c);
	strncpy(&szMsg[i],c,j);
	szMsg[i+j]=0;
	msg=szMsg;
}

SocketException::SocketException(int i){
	errcode=i;
	setStr(strerror(i));
}

SocketException::SocketException(int i,const char *s){
	errcode=i;
	setStr(s);
}

void SocketException::setStr(const char *s){
    char szMsg[1024];
	if (!s) {msg="";return ;}
	strncpy(szMsg,s,(sizeof(szMsg)-1)<strlen(s) ? (sizeof(szMsg)-1):strlen(s));
	szMsg[sizeof(szMsg)-1]=0;
	msg=szMsg;
}
const char* SocketException::what() const throw(){
    return msg.c_str();
}
SocketException::~SocketException() throw(){
}

#define SO_DEFAULT 0xffffffff
Socket::Socket():iSocket(-1),iSocketType(SOCK_STREAM),saPermanent(0),uSaLen(0),feof(0) {
    sockopts.uOpt_keepalive=SO_DEFAULT;
	sockopts.uOpt_sndbuf=SO_DEFAULT;
	sockopts.uOpt_rcvbuf=SO_DEFAULT;
}
void Socket::setsckopts(int iSocket){
    int rc=0;
    if (sockopts.uOpt_keepalive!=SO_DEFAULT)
        rc=setsockopt(iSocket,SOL_SOCKET,SO_KEEPALIVE,&sockopts.uOpt_keepalive,sizeof(sockopts.uOpt_keepalive));
    if (rc<0) perror("setsockopt(SO_KEEPALIVE)");
    if (sockopts.uOpt_sndbuf!=SO_DEFAULT)
        rc=setsockopt(iSocket,SOL_SOCKET,SO_SNDBUF,&sockopts.uOpt_sndbuf,sizeof(sockopts.uOpt_sndbuf));
    if (rc<0) perror("setsockopt(SO_SNDBUF)");
    if (sockopts.uOpt_rcvbuf!=SO_DEFAULT)
        rc=setsockopt(iSocket,SOL_SOCKET,SO_RCVBUF,&sockopts.uOpt_rcvbuf,sizeof(sockopts.uOpt_sndbuf));        
    if (rc<0) perror("setsockopt(SO_RCVBUF)");
}
Socket::Socket(const Socket &x):iSocket(x.iSocket),iSocketType(x.iSocketType),sockopts(x.sockopts),feof(x.feof){
	if (x.saPermanent){
		saPermanent=malloc(uSaLen=x.uSaLen);
		memcpy(saPermanent,x.saPermanent,uSaLen);
	}
}
Socket & Socket::operator=(const Socket &x){
	release();
	iSocket=x.iSocket;
	iSocketType=x.iSocketType;
        feof=x.feof;
	sockopts=x.sockopts;
	if (x.saPermanent){
		saPermanent=malloc(uSaLen=x.uSaLen);
		memcpy(saPermanent,x.saPermanent,uSaLen);
	}
	return *this;
}
Socket::~Socket(){close();}
void Socket::setSocket(int x){
	close();
	iSocketType=SOCK_STREAM;
	iSocket=x;
        feof=0;
    setsckopts(iSocket);
}
void Socket::close(){
        feof=1;
	if (iSocket==-1) {
		if (saPermanent) free(saPermanent);
		saPermanent=0;
		return ;
	}
	shutdown(iSocket,2);
	::close(iSocket);
	release();
}
void Socket::release() const{
	iSocket=-1;
	if (saPermanent) free(saPermanent);
	saPermanent=0;
        feof=1;
}
size_t Socket::Write(const void * buf,size_t sz){
	if (iSocket==-1) return 0;
	if (iSocketType==SOCK_STREAM){
		return write(iSocket,buf,sz);
	}
	if ((iSocketType==SOCK_DGRAM) && saPermanent){
		return sendto(iSocket,buf,sz,0,(struct sockaddr *)saPermanent,uSaLen);
	}
	return 0;
}
size_t Socket::Read(void * buf,size_t sz){
    size_t ret=0;
	if (iSocket==-1) return 0;
	if (iSocketType==SOCK_STREAM){
		ret=read(iSocket,buf,sz);
	}
	if (iSocketType==SOCK_DGRAM){
		socklen_t sl=uSaLen;
		void * vAddr=alloca(uSaLen);
		ret=recvfrom(iSocket,buf,sz,0,(struct sockaddr *)vAddr,&sl);
	}
	feof=(ret==0);
	return ret;
}
bool Socket::eof() const{
	if (iSocket==-1) return true;
	return feof!=0;
}

bool Socket::is_readready(float fTimeout){
	if (iSocket==-1) return true;
	int rt;
	fd_set fds;
	struct timeval tvl;
	tvl.tv_sec=(int)fTimeout;
	tvl.tv_usec=(fTimeout-tvl.tv_sec)*1000000;
	FD_ZERO(&fds);
	FD_SET(iSocket,&fds);
	rt=select(iSocket+1,&fds,0,0,&tvl);
	if (rt<1) return false;
	return FD_ISSET(iSocket,&fds);
}

bool Socket::is_writeready(float fTimeout){
	if (iSocket==-1) return true;
	fd_set fds;
	struct timeval tvl;
	tvl.tv_sec=(int)fTimeout;
	tvl.tv_usec=(fTimeout-tvl.tv_sec)*1000000;
	FD_ZERO(&fds);
	FD_SET(iSocket,&fds);
	return select(iSocket+1,0,&fds,0,&tvl)>0;
}

bool Socket::accept(Socket &acc){
    char cSa[1024];
    socklen_t slt;
    int sck;
    if (!valid()) return false;
    slt=sizeof(cSa);
    sck=::accept(iSocket,(struct sockaddr *)cSa,&slt);
    if (sck<0) return false;
    acc.close();
    acc.sockopts=sockopts;
    acc.iSocket=sck;
    acc.uSaLen=slt;
    acc.saPermanent=malloc(slt);
    acc.iSocketType=SOCK_STREAM;
    acc.setsckopts(sck);
    memcpy(acc.saPermanent,cSa,slt);
    return true;
}

bool Socket::openSocket(const char *suri,SocketOpenMode mode){
    if (!suri) return false;
    if (strlen(suri)<5) return false;
    close();
    if ((*((int *)suri)=='elif') && (suri[4]==':')){
        iSocket=socket(PF_UNIX,iSocketType=SOCK_STREAM,0);
        setsckopts(iSocket);
        saPermanent=malloc(uSaLen=sizeof(struct sockaddr_un));
        struct sockaddr_un * sck=(struct sockaddr_un *)saPermanent;
        memset(sck,0,sizeof(struct sockaddr_un));
        sck->sun_family=AF_UNIX;
        strncpy(sck->sun_path,suri+5,sizeof(sck->sun_path));
        if (mode==somBind){
            unlink(suri+5);
            if (bind(iSocket,(struct sockaddr *)sck,uSaLen)) throw SocketException("bind"); 
            SETNONBLOCK(iSocket);
            feof=0;
            return true;
        }
        if (mode==somConnect){
            if (connect(iSocket,(struct sockaddr *)sck,uSaLen)) throw SocketException("connect");
            SETNONBLOCK(iSocket);
            feof=0;
            return true;
        }
        return true;
    } else
#ifdef INET6
    if ( ( (*((int *)suri)=='6vpi') || (*((int *)suri)=='6pct')) && (suri[4]==':') ){
	iSocket=socket(PF_INET6,iSocketType=SOCK_STREAM,0);
    setsckopts(iSocket);
	saPermanent=malloc(uSaLen=sizeof(struct sockaddr_in6));
        struct sockaddr_in6 *sck=(struct sockaddr_in6 *)saPermanent;
        memset(sck,0,uSaLen);
        sck->sin6_family=AF_INET6;
        char szBuf[256];
        int i;
        for(i=5;suri[i] && (suri[i]!='.');i++) szBuf[i-5]=suri[i];
        szBuf[i-5]=0;
        sscanf(&suri[i+1],"%d",&i);
        sck->sin6_port=htons(i);
	struct in6_addr ul;
	if (inet_pton(AF_INET6,szBuf,&ul)!=sizeof(ul)){
            struct hostent * he=gethostbyname2(szBuf,AF_INET6);
            if (!he) return false;
	    memcpy(&ul,he->h_addr_list[0],sizeof(ul));
        }
        sck->sin6_addr=ul;
        if (mode==somBind){
                if (bind(iSocket,(struct sockaddr *)sck,uSaLen)) throw SocketException("bind"); 
                SETNONBLOCK(iSocket);
                feof=0;
                return true;
        }
        if (mode==somConnect){
                if (connect(iSocket,(struct sockaddr *)sck,uSaLen)) throw SocketException("connect"); 
		SETNONBLOCK(iSocket);
                feof=0;
                return true;
        }
	return true;
    } else
#endif    
    if ( ( (*((int *)suri)=='4vpi') || (*((int *)suri)=='4pct')) && (suri[4]==':') ){
	iSocket=socket(PF_INET,iSocketType=SOCK_STREAM,6);
    setsckopts(iSocket);
	saPermanent=malloc(uSaLen=sizeof(struct sockaddr_in));
        struct sockaddr_in *sck=(struct sockaddr_in *)saPermanent;
        memset(sck,0,sizeof(struct sockaddr_in));
        sck->sin_family=AF_INET;
        char szBuf[256];
        int i;
        for(i=5;suri[i] && (suri[i]!=':');i++) szBuf[i-5]=suri[i];
        szBuf[i-5]=0;
        sscanf(&suri[i+1],"%d",&i);
        sck->sin_port=htons(i);
        unsigned long ul=inet_addr(szBuf);
        if (ul==0xffffffff){
            struct hostent * he=gethostbyname2(szBuf,AF_INET);
            if (!he) return false;
            memcpy(&ul,he->h_addr_list[0],4);
        }
        sck->sin_addr.s_addr=ul;
        if (mode==somBind){
                if (bind(iSocket,(struct sockaddr *)sck,uSaLen)) throw SocketException("bind"); 
        	SETNONBLOCK(iSocket);
                feof=0;
                return true;
        }
        if (mode==somConnect){
                if (connect(iSocket,(struct sockaddr *)sck,uSaLen)) throw SocketException("connect"); 
        	SETNONBLOCK(iSocket);
                feof=0;
                return true;
        }
        return true;
    } else
    if ( (*((int *)suri)=='4pdu') && (suri[4]==':') ){
	iSocket=socket(PF_INET,iSocketType=SOCK_DGRAM,17);
    setsckopts(iSocket);
	saPermanent=malloc(uSaLen=sizeof(struct sockaddr_in));
	struct sockaddr_in *sck=(struct sockaddr_in *)saPermanent;
        memset(sck,0,sizeof(struct sockaddr_in));
        sck->sin_family=AF_INET;
        char szBuf[256];
        int i;
        for(i=5;suri[i] && (suri[i]!=':');i++) szBuf[i-5]=suri[i];
        szBuf[i-5]=0;
        sscanf(&suri[i+1],"%d",&i);
        sck->sin_port=htons(i);
        unsigned long ul=inet_addr(szBuf);
        if (ul==0xffffffff){
            struct hostent * he=gethostbyname2(szBuf,AF_INET);
            if (!he) return false;
	    memcpy(&ul,he->h_addr_list[0],4);
        }
        sck->sin_addr.s_addr=ul;
        if (mode==somBind){
		for(i=5;suri[i] && (suri[i]!='@');i++) ;
        	if (suri[i]=='@'){
                	sck->sin_addr.s_addr=inet_addr(suri+i+1);
                	if (sck->sin_addr.s_addr!=0xffff){
                        	unsigned u=1;
                        	if(setsockopt(iSocket, SOL_SOCKET, SO_REUSEADDR, (char *)&u, sizeof(u)))
					throw SocketException("setsockopt(SO_REUSEADDR)");
                	}
        	}
                if (bind(iSocket,(struct sockaddr *)sck,uSaLen)) throw SocketException("bind");
        if (suri[i]=='@'){
            struct ip_mreq mreq;
            memset(&mreq,0,sizeof(mreq));
            mreq.imr_multiaddr.s_addr = inet_addr(szBuf);
            mreq.imr_interface.s_addr = sck->sin_addr.s_addr;
            if (setsockopt(iSocket,IPPROTO_IP,IP_ADD_MEMBERSHIP,(char*)&mreq,sizeof(mreq))) throw SocketException("setsockopt(IP_ADD_MEMBERSHIP 0x%08x@0x%08x)",inet_addr(szBuf),sck->sin_addr.s_addr);
        }
                SETNONBLOCK(iSocket);
                feof=0;
                return true;
        }
        if (mode==somConnect){
            for(i=5;suri[i] && (suri[i]!='@');i++) ;
            if (suri[i]=='@'){
                unsigned uP=sck->sin_port;
                sck->sin_port=0;
                sck->sin_addr.s_addr=inet_addr(suri+i+1);
                bind(iSocket,(struct sockaddr *)sck,uSaLen);
                sck->sin_addr.s_addr=ul;
                sck->sin_port=uP;
            }
                SETNONBLOCK(iSocket);
                feof=0;
                return true;
        }
        return true;
    }
    return false;
}


}
