create sequence bgplogger_seq start with 1 increment by 1 nocycle order;

create table bgp_src (
 id int NOT NULL primary key,
 loggerip varchar(40) null,
 peerip varchar(40),
 peeras int,
 flags int default 0 not null
);
create or replace procedure bgplog_getsrcif(resid OUT NUMBER,peerip1 IN varchar2,peeras1 IN NUMBER)
IS tid int;
begin
 if (peerip1 is null or peeras1 is null) then
  resid:=null;
  return;
 end if;
 select min(id) into tid from bgp_src where peerip=peerip1 and peeras=peeras1;
 if tid is null then
  select bgplogger_seq.nextval into tid from dual;
  insert into bgp_src(id,peerip,peeras) values (tid,peerip1,peeras1);
  commit;
 end if;
 resid:=tid;
end;

create table bgp_commonlog (
 srcid int not null references bgp_src(id),
 dtwhen timestamp default CURRENT_TIMESTAMP,
 msgcode int null,
 msgtxt varchar2(256) null
);

create table bgp_ipv4 (
 srcid int not null references bgp_src(id),
 dtwhen timestamp default CURRENT_TIMESTAMP,
 state number(2) default 0 not null, -- 1 - ipv4u_nlri ; 2 - ipv4u_withdraw
 kind number(4) default 0 not null, -- 1 - ipv4u ; 2 - ipv4m ; 4 - ipv4lu
 netaddr number(10) default 0 not null, -- network address
 netsize number(3) default 0 not null, -- network size
 bgporigin number(1) default 0 not null, -- bgp origin code
 bgpaspath varchar2(256) null, -- bgp aspath
 bgpcommunities varchar2(256) null,
 bgpnexthop number(10) null, -- nexthop address
 bgpmed number(10) null,
 bgplocalpref number(10) null,
 bgpatomicaggregate number(10) null, -- aa address
 bgpaggregatoras number(10) null,
 bgpagregatorIP number(10) null,
 bgporiginatorIP number(10) null,
 bgpextcom varchar2(256) null,
 bgplabels varchar2(256) null,
 primary key (srcid,kind,netaddr,netsize)
);
create index bgp_ipv4_lbl on bgp_ipv4(bgplabels);
create index bgp_ipv4_ec on bgp_ipv4(bgpextcom);
create index bgp_ipv4_cm on bgp_ipv4(bgpcommunities);

create table bgp_ipv4_log (
 srcid int not null references bgp_src(id),
 dtwhen timestamp default CURRENT_TIMESTAMP,
 state number(2) default 0 not null, -- 1 - ipv4u_nlri ; 2 - ipv4u_withdraw
 kind number(4) default 0 not null, -- 1 - ipv4u ; 2 - ipv4m ; 4 - ipv4lu
 netaddr number(10) default 0 not null, -- network address
 netsize number(3) default 0 not null, -- network size
 bgporigin number(1) default 0 not null, -- bgp origin code
 bgpaspath varchar2(256) null, -- bgp aspath
 bgpcommunities varchar2(256) null,
 bgpnexthop number(10) null, -- nexthop address
 bgpmed number(10) null,
 bgplocalpref number(10) null,
 bgpatomicaggregate number(10) null, -- aa address
 bgpaggregatoras number(10) null,
 bgpagregatorIP number(10) null,
 bgporiginatorIP number(10) null,
 bgpextcom varchar2(256) null,
 bgplabels varchar2(256) null
);
create index bgp_ipv4_l_lbl on bgp_ipv4_log(bgplabels);
create index bgp_ipv4_l_ec on bgp_ipv4_log(bgpextcom);
create index bgp_ipv4_l_cm on bgp_ipv4_log(bgpcommunities);
create index bgp_ipv4_l_net on bgp_ipv4_log(srcid,kind,netaddr,netsize);
create index bgp_ipv4_l_dtwhen on bgp_ipv4_log(dtwhen);


CREATE OR REPLACE PROCEDURE wr_ipv4(i_srcid IN NUMBER,i_state IN NUMBER,i_kind IN NUMBER,i_netaddr IN NUMBER,i_netsize IN NUMBER,i_bgporigin  IN NUMBER,
 i_bgpaspath IN varchar2,i_bgpcommunities IN varchar2,i_bgpnexthop  IN NUMBER,i_bgpmed  IN NUMBER,i_bgplocalpref  IN NUMBER,
 i_bgpatomicaggregate  IN NUMBER,i_bgpaggregatoras  IN NUMBER,i_bgpagregatorIP  IN NUMBER,i_bgporiginatorIP  IN NUMBER,i_bgpextcom IN varchar2,
 i_bgplabels IN varchar2
) IS
 state1 number;
 bgporigin1 number;
 bgpaspath1 varchar2(256);
 bgpcommunities1 varchar2(256);
 bgpnexthop1 number;
 bgpmed1 number;
 bgplocalpref1 number;
 bgpatomicaggregate1 number;
 bgpaggregatoras1 number;
 bgpagregatorIP1 number;
 bgporiginatorIP1 number;
 bgpextcom1 varchar2(256);
 bgplabels1 varchar2(256);
BEGIN
 select count(*) into state1 from bgp_ipv4 where kind=i_kind and srcid=i_srcid and netaddr=i_netaddr and netsize=i_netsize;
 if state1>0 then
  select state,bgporigin,bgpaspath,bgpcommunities,bgpnexthop,bgpmed,bgplocalpref,bgpatomicaggregate,bgpaggregatoras,bgpagregatorIP,bgporiginatorIP,bgpextcom,bgplabels
  into state1,bgporigin1,bgpaspath1,bgpcommunities1,bgpnexthop1,bgpmed1,bgplocalpref1,bgpatomicaggregate1,bgpaggregatoras1,bgpagregatorIP1,bgporiginatorIP1,bgpextcom1,bgplabels1
   from bgp_ipv4 where kind=i_kind and srcid=i_srcid and netaddr=i_netaddr and netsize=i_netsize;
 end if;
 if state1 is not null and state1=coalesce(i_state,0) and bgporigin1=coalesce(i_bgporigin,0) and coalesce(bgpaspath1,'-')=coalesce(i_bgpaspath,'-') and coalesce(bgpcommunities1,'-')=coalesce(i_bgpcommunities,'-') and coalesce(bgpnexthop1,0)=coalesce(i_bgpnexthop,0) and coalesce(bgpmed1,0)=coalesce(i_bgpmed,0) and coalesce(bgplocalpref1,0)=coalesce(i_bgplocalpref,0) and coalesce(bgpatomicaggregate1,0)=coalesce(i_bgpatomicaggregate,0) and coalesce(bgpaggregatoras1,0)=coalesce(i_bgpaggregatoras,0) and coalesce(bgpagregatorIP1,0)=coalesce(i_bgpagregatorIP,0) and coalesce(bgporiginatorIP1,0)=coalesce(i_bgporiginatorIP,0) and coalesce(bgpextcom1,'-')=coalesce(i_bgpextcom,'-') then
   state1:=1;
 else
  delete from bgp_ipv4 where srcid=i_srcid and kind=i_kind and netaddr=i_netaddr and netsize=i_netsize;
  insert into bgp_ipv4 (srcid,state,kind,netaddr,netsize,bgporigin,bgpaspath,bgpcommunities,bgpnexthop,bgpmed,bgplocalpref,bgpatomicaggregate,bgpaggregatoras,bgpagregatorIP,bgporiginatorIP,bgpextcom,bgplabels) values (i_srcid,i_state,i_kind,i_netaddr,i_netsize,i_bgporigin,i_bgpaspath,i_bgpcommunities,i_bgpnexthop,i_bgpmed,i_bgplocalpref,i_bgpatomicaggregate,i_bgpaggregatoras,i_bgpagregatorIP,i_bgporiginatorIP,i_bgpextcom,i_bgplabels);
  insert into bgp_ipv4_log (srcid,state,kind,netaddr,netsize,bgporigin,bgpaspath,bgpcommunities,bgpnexthop,bgpmed,bgplocalpref,bgpatomicaggregate,bgpaggregatoras,bgpagregatorIP,bgporiginatorIP,bgpextcom,bgplabels) values (i_srcid,i_state,i_kind,i_netaddr,i_netsize,i_bgporigin,i_bgpaspath,i_bgpcommunities,i_bgpnexthop,i_bgpmed,i_bgplocalpref,i_bgpatomicaggregate,i_bgpaggregatoras,i_bgpagregatorIP,i_bgporiginatorIP,i_bgpextcom,i_bgplabels);
  commit;
 end if;
END;

create table bgp_vpnv4 (
 srcid int not null references bgp_src(id),
 dtwhen timestamp default CURRENT_TIMESTAMP,
 state number(2) default 0 not null, -- 1 - ipv4u_nlri ; 2 - ipv4u_withdraw
 kind number(4) default 0 not null, -- 1 - ipv4u ; 2 - ipv4m ; 4 - ipv4lu
 netaddrrdh number(10) default 0 not null, -- network address RD high
 netaddrrdl number(10) default 0 not null, -- network address RD low
 netaddr number(10) default 0 not null, -- network address
 netsize number(3) default 0 not null, -- network size
 bgporigin number(1) default 0 not null, -- bgp origin code
 bgpaspath varchar2(256) null, -- bgp aspath
 bgpcommunities varchar2(256) null,
 nexthoprdh number(10) null,
 nexthoprdl number(10) null,
 nexthopip number(10) null, -- nexthop address
 bgpmed number(10) null,
 bgplocalpref number(10) null,
 bgpatomicaggregate number(10) null, -- aa address
 bgpaggregatoras number(10) null,
 bgpagregatorIP number(10) null,
 bgporiginatorIP number(10) null,
 bgpextcom varchar2(256) null,
 bgplabels varchar2(256) null,
 primary key (srcid,kind,netaddrrdh,netaddrrdl,netaddr,netsize)
);
create index bgp_vpnv4_lbl on bgp_vpnv4(bgplabels);
create index bgp_vpnv4_ec on bgp_vpnv4(bgpextcom);
create index bgp_vpnv4_cm on bgp_vpnv4(bgpcommunities);


create table bgp_vpnv4_log (
 srcid int not null references bgp_src(id),
 dtwhen timestamp default CURRENT_TIMESTAMP,
 state number(2) default 0 not null, -- 1 - ipv4u_nlri ; 2 - ipv4u_withdraw
 kind number(4) default 0 not null, -- 1 - ipv4u ; 2 - ipv4m ; 4 - ipv4lu
 netaddrrdh number(10) default 0 not null, -- network address RD high
 netaddrrdl number(10) default 0 not null, -- network address RD low
 netaddr number(10) default 0 not null, -- network address
 netsize number(3) default 0 not null, -- network size
 bgporigin number(1) default 0 not null, -- bgp origin code
 bgpaspath varchar2(256) null, -- bgp aspath
 bgpcommunities varchar2(256) null,
 nexthoprdh number(10) null,
 nexthoprdl number(10) null,
 nexthopip number(10) null, -- nexthop address
 bgpmed number(10) null,
 bgplocalpref number(10) null,
 bgpatomicaggregate number(10) null, -- aa address
 bgpaggregatoras number(10) null,
 bgpagregatorIP number(10) null,
 bgporiginatorIP number(10) null,
 bgpextcom varchar2(256) null,
 bgplabels varchar2(256) null
);
create index bgp_vpnv4_l_lbl on bgp_vpnv4_log(bgplabels);
create index bgp_vpnv4_l_ec on bgp_vpnv4_log(bgpextcom);
create index bgp_vpnv4_l_cm on bgp_vpnv4_log(bgpcommunities);
create index bgp_vpnv4_l_net on bgp_vpnv4_log(srcid,kind,netaddrrdh,netaddrrdl,netaddr,netsize);
create index bgp_vpnv4_l_dtwhen on bgp_vpnv4_log(dtwhen);

CREATE OR REPLACE PROCEDURE wr_vpnv4(i_srcid IN NUMBER,i_state IN NUMBER,i_kind IN NUMBER,i_netaddrrdh IN NUMBER,i_netaddrrdl IN NUMBER,i_netaddr IN NUMBER,i_netsize IN NUMBER,i_bgporigin  IN NUMBER,
 i_bgpaspath IN varchar2,i_bgpcommunities IN varchar2,i_nexthoprdh  IN NUMBER,i_nexthoprdl  IN NUMBER,i_nexthop  IN NUMBER,i_bgpmed  IN NUMBER,i_bgplocalpref  IN NUMBER,
 i_bgpatomicaggregate  IN NUMBER,i_bgpaggregatoras  IN NUMBER,i_bgpagregatorIP  IN NUMBER,i_bgporiginatorIP  IN NUMBER,i_bgpextcom IN varchar2,
 i_bgplabels IN varchar2
) IS
 state1 number;
 bgporigin1 number;
 bgpaspath1 varchar2(256);
 bgpcommunities1 varchar2(256);
 nexthoprdh1 number;
 nexthoprdl1 number;
 nexthop1 number;
 bgpmed1 number;
 bgplocalpref1 number;
 bgpatomicaggregate1 number;
 bgpaggregatoras1 number;
 bgpagregatorIP1 number;
 bgporiginatorIP1 number;
 bgpextcom1 varchar2(256);
 bgplabels1 varchar2(256);
BEGIN
 select count(*) into state1 from bgp_vpnv4 where kind=i_kind and srcid=i_srcid and netaddrrdh=i_netaddrrdh and netaddrrdl=i_netaddrrdl and netaddr=i_netaddr and netsize=i_netsize;
 if state1>0 then
  select state,bgporigin,bgpaspath,bgpcommunities,nexthoprdh,nexthoprdl,nexthopip,bgpmed,bgplocalpref,bgpatomicaggregate,bgpaggregatoras,bgpagregatorIP,bgporiginatorIP,bgpextcom,bgplabels
  into state1,bgporigin1,bgpaspath1,bgpcommunities1,nexthoprdh1,nexthoprdl1,nexthop1,bgpmed1,bgplocalpref1,bgpatomicaggregate1,bgpaggregatoras1,bgpagregatorIP1,bgporiginatorIP1,bgpextcom1,bgplabels1
   from bgp_vpnv4 where kind=i_kind and srcid=i_srcid and netaddrrdh=i_netaddrrdh and netaddrrdl=i_netaddrrdl and netaddr=i_netaddr and netsize=i_netsize;
 end if;
 if state1 is not null and state1=coalesce(i_state,0) and bgporigin1=coalesce(i_bgporigin,0) and coalesce(bgpaspath1,'-')=coalesce(i_bgpaspath,'-') and coalesce(bgpcommunities1,'-')=coalesce(i_bgpcommunities,'-') and coalesce(nexthoprdh1,0)=coalesce(i_nexthoprdh,0) and coalesce(nexthoprdl1,0)=coalesce(i_nexthoprdl,0) and coalesce(nexthop1,0)=coalesce(i_nexthop,0) and coalesce(bgpmed1,0)=coalesce(i_bgpmed,0) and coalesce(bgplocalpref1,0)=coalesce(i_bgplocalpref,0) and coalesce(bgpatomicaggregate1,0)=coalesce(i_bgpatomicaggregate,0) and coalesce(bgpaggregatoras1,0)=coalesce(i_bgpaggregatoras,0) and coalesce(bgpagregatorIP1,0)=coalesce(i_bgpagregatorIP,0) and coalesce(bgporiginatorIP1,0)=coalesce(i_bgporiginatorIP,0) and coalesce(bgpextcom1,'-')=coalesce(i_bgpextcom,'-') then
   state1:=1;
 else
  delete from bgp_vpnv4 where srcid=i_srcid and kind=i_kind and netaddrrdh=i_netaddrrdh and netaddrrdl=i_netaddrrdl and netaddr=i_netaddr and netsize=i_netsize;
  insert into bgp_vpnv4 (srcid,state,kind,netaddrrdh,netaddrrdl,netaddr,netsize,bgporigin,bgpaspath,bgpcommunities,nexthoprdh,nexthoprdl,nexthopip,bgpmed,bgplocalpref,bgpatomicaggregate,bgpaggregatoras,bgporiginatorIP,bgpextcom,bgplabels) values (i_srcid,i_state,i_kind,i_netaddrrdh,i_netaddrrdl,i_netaddr,i_netsize,i_bgporigin,i_bgpaspath,i_bgpcommunities,i_nexthoprdh,i_nexthoprdl,i_nexthop,i_bgpmed,i_bgplocalpref,i_bgpatomicaggregate,i_bgpaggregatoras,i_bgporiginatorIP,i_bgpextcom,i_bgplabels);
  insert into bgp_vpnv4_log (srcid,state,kind,netaddrrdh,netaddrrdl,netaddr,netsize,bgporigin,bgpaspath,bgpcommunities,nexthoprdh,nexthoprdl,nexthopip,bgpmed,bgplocalpref,bgpatomicaggregate,bgpaggregatoras,bgporiginatorIP,bgpextcom,bgplabels) values (i_srcid,i_state,i_kind,i_netaddrrdh,i_netaddrrdl,i_netaddr,i_netsize,i_bgporigin,i_bgpaspath,i_bgpcommunities,i_nexthoprdh,i_nexthoprdl,i_nexthop,i_bgpmed,i_bgplocalpref,i_bgpatomicaggregate,i_bgpaggregatoras,i_bgporiginatorIP,i_bgpextcom,i_bgplabels);
  commit;
 end if;
END;


create table bgp_ipv6 (
 srcid int not null references bgp_src(id),
 dtwhen timestamp default CURRENT_TIMESTAMP,
 state number(2) default 0 not null, -- 1 - ipv4u_nlri ; 2 - ipv4u_withdraw
 kind number(4) default 0 not null, -- 1 - ipv4u ; 2 - ipv4m ; 4 - ipv4lu
 netaddrh number(20) default 0 not null, -- network address high part
 netaddrl number(20) default 0 not null, -- network address low part
 netsize number(3) default 0 not null, -- network size
 bgporigin number(1) default 0 not null, -- bgp origin code
 bgpaspath varchar2(256) null, -- bgp aspath
 bgpcommunities varchar2(256) null,
 bgpnexthoph number(20) null, -- nexthop address high part
 bgpnexthopl number(20) null, -- nexthop address low part
 bgpmed number(10) null,
 bgplocalpref number(10) null,
 bgpatomicaggregateh number(20) null, -- aa address high part
 bgpatomicaggregatel number(20) null, -- aa address low part
 bgpaggregatoras number(20) null,
 bgpagregatorIPh number(20) null,
 bgpagregatorIPl number(20) null,
 bgporiginatorIPh number(20) null,
 bgporiginatorIPl number(20) null,
 bgpextcom varchar2(256) null,
 bgplabels varchar2(256) null,
 primary key (srcid,kind,netaddrh,netaddrl,netsize)
);
create index bgp_ipv6_lbl on bgp_ipv6(bgplabels);
create index bgp_ipv6_ec on bgp_ipv6(bgpextcom);
create index bgp_ipv6_cm on bgp_ipv6(bgpcommunities);

create table bgp_ipv6_log (
 srcid int not null references bgp_src(id),
 dtwhen timestamp default CURRENT_TIMESTAMP,
 state number(2) default 0 not null, -- 1 - ipv4u_nlri ; 2 - ipv4u_withdraw
 kind number(4) default 0 not null, -- 1 - ipv4u ; 2 - ipv4m ; 4 - ipv4lu
 netaddrh number(20) default 0 not null, -- network address high part
 netaddrl number(20) default 0 not null, -- network address low part
 netsize number(3) default 0 not null, -- network size
 bgporigin number(1) default 0 not null, -- bgp origin code
 bgpaspath varchar2(256) null, -- bgp aspath
 bgpcommunities varchar2(256) null,
 bgpnexthoph number(20) null, -- nexthop address high part
 bgpnexthopl number(20) null, -- nexthop address low part
 bgpmed number(10) null,
 bgplocalpref number(10) null,
 bgpatomicaggregateh number(20) null, -- aa address high part
 bgpatomicaggregatel number(20) null, -- aa address low part
 bgpaggregatoras number(20) null,
 bgpagregatorIPh number(20) null,
 bgpagregatorIPl number(20) null,
 bgporiginatorIPh number(20) null,
 bgporiginatorIPl number(20) null,
 bgpextcom varchar2(256) null,
 bgplabels varchar2(256) null
);
create index bgp_ipv6_l_dt on bgp_ipv6_log(dtwhen);
create index bgp_ipv6_l_net on bgp_ipv6_log(srcid,kind,netaddrh,netaddrl,netsize);
create index bgp_ipv6_l_lbl on bgp_ipv6_log(bgplabels);
create index bgp_ipv6_l_ec on bgp_ipv6_log(bgpextcom);
create index bgp_ipv6_l_cm on bgp_ipv6_log(bgpcommunities);

CREATE OR REPLACE PROCEDURE wr_ipv6(i_srcid IN NUMBER,i_state IN NUMBER,i_kind IN NUMBER,i_netaddrh IN NUMBER,i_netaddrl IN NUMBER,i_netsize IN NUMBER,
 i_bgporigin  IN NUMBER,i_bgpaspath IN varchar2,i_bgpcommunities IN varchar2,i_bgpnexthoph IN NUMBER,i_bgpnexthopl IN NUMBER,i_bgpmed  IN NUMBER,i_bgplocalpref  IN NUMBER,
 i_bgpatomicaggregateh IN NUMBER,i_bgpatomicaggregatel IN NUMBER,i_bgpaggregatoras  IN NUMBER,i_bgpagregatorIPh  IN NUMBER,i_bgpagregatorIPl  IN NUMBER,
 i_bgporiginatorIPh IN NUMBER,i_bgporiginatorIPl IN NUMBER,i_bgpextcom IN varchar2,i_bgplabels IN varchar2
) IS
 state1 number;
 bgporigin1 number;
 bgpaspath1 varchar2(256);
 bgpcommunities1 varchar2(256);
 bgpnexthoph1 number(20);
 bgpnexthopl1 number(20);
 bgpmed1 number;
 bgplocalpref1 number;
 bgpatomicaggregateh1 number(20);
 bgpatomicaggregatel1 number(20);
 bgpaggregatoras1 number;
 bgpagregatorIPh1 number(20);
 bgpagregatorIPl1 number(20);
 bgporiginatorIPh1 number(20);
 bgporiginatorIPl1 number(20);
 bgpextcom1 varchar2(256);
 bgplabels1 varchar2(256);
BEGIN
 select count(*) into state1 from bgp_ipv6 where kind=i_kind and srcid=i_srcid and netaddrh=i_netaddrh and netaddrl=i_netaddrl and netsize=i_netsize;
 if state1>0 then
  select state,bgporigin,bgpaspath,bgpcommunities,bgpnexthoph,bgpnexthopl,bgpmed,bgplocalpref,bgpatomicaggregateh,bgpatomicaggregatel,bgpaggregatoras,bgpagregatorIPh,bgpagregatorIPl,bgporiginatorIPh,bgporiginatorIPl,bgpextcom,bgplabels
  into state1,bgporigin1,bgpaspath1,bgpcommunities1,bgpnexthoph1,bgpnexthopl1,bgpmed1,bgplocalpref1,bgpatomicaggregateh1,bgpatomicaggregatel1,bgpaggregatoras1,bgpagregatorIPh1,bgpagregatorIPl1,bgporiginatorIPh1,bgporiginatorIPl1,bgpextcom1,bgplabels1
   from bgp_ipv6 where kind=i_kind and srcid=i_srcid and netaddrh=i_netaddrh and netaddrl=i_netaddrl and netsize=i_netsize;
 end if;
 if state1 is not null and state1=coalesce(i_state,0) and bgporigin1=coalesce(i_bgporigin,0) and coalesce(bgpaspath1,'-')=coalesce(i_bgpaspath,'-') and coalesce(bgpcommunities1,'-')=coalesce(i_bgpcommunities,'-') and 
  coalesce(bgpnexthoph1,0)=coalesce(i_bgpnexthoph,0) and coalesce(bgpnexthopl1,0)=coalesce(i_bgpnexthopl,0) and coalesce(bgpmed1,0)=coalesce(i_bgpmed,0) and coalesce(bgplocalpref1,0)=coalesce(i_bgplocalpref,0) and 
  coalesce(bgpatomicaggregateh1,0)=coalesce(i_bgpatomicaggregateh,0) and coalesce(bgpatomicaggregatel1,0)=coalesce(i_bgpatomicaggregatel,0) and coalesce(bgpaggregatoras1,0)=coalesce(i_bgpaggregatoras,0) and 
  coalesce(bgpagregatorIPh1,0)=coalesce(i_bgpagregatorIPh,0) and coalesce(bgpagregatorIPl1,0)=coalesce(i_bgpagregatorIPl,0) and coalesce(bgporiginatorIPh1,0)=coalesce(i_bgporiginatorIPh,0) and coalesce(bgporiginatorIPl1,0)=coalesce(i_bgporiginatorIPl,0) and 
  coalesce(bgpextcom1,'-')=coalesce(i_bgpextcom,'-') then
   state1:=1;
 else
  delete from bgp_ipv6 where srcid=i_srcid and kind=i_kind and netaddrh=i_netaddrh and netaddrl=i_netaddrl and netsize=i_netsize;
  insert into bgp_ipv6 (srcid,state,kind,netaddrh,netaddrl,netsize,bgporigin,bgpaspath,bgpcommunities,bgpnexthoph,bgpnexthopl,bgpmed,bgplocalpref,bgpatomicaggregateh,bgpatomicaggregatel,bgpaggregatoras,bgporiginatorIPh,bgpagregatorIPl,bgpagregatorIPh,bgporiginatorIPl,bgpextcom,bgplabels) values
   (i_srcid,i_state,i_kind,i_netaddrh,i_netaddrl,i_netsize,i_bgporigin,i_bgpaspath,i_bgpcommunities,i_bgpnexthoph,i_bgpnexthopl,i_bgpmed,i_bgplocalpref,i_bgpatomicaggregateh,i_bgpatomicaggregatel,i_bgpaggregatoras,i_bgpagregatorIPh,i_bgpagregatorIPl,i_bgporiginatorIPh,i_bgporiginatorIPl,i_bgpextcom,i_bgplabels);
  insert into bgp_ipv6_log (srcid,state,kind,netaddrh,netaddrl,netsize,bgporigin,bgpaspath,bgpcommunities,bgpnexthoph,bgpnexthopl,bgpmed,bgplocalpref,bgpatomicaggregateh,bgpatomicaggregatel,bgpaggregatoras,bgporiginatorIPh,bgpagregatorIPl,bgpagregatorIPh,bgporiginatorIPl,bgpextcom,bgplabels) values
   (i_srcid,i_state,i_kind,i_netaddrh,i_netaddrl,i_netsize,i_bgporigin,i_bgpaspath,i_bgpcommunities,i_bgpnexthoph,i_bgpnexthopl,i_bgpmed,i_bgplocalpref,i_bgpatomicaggregateh,i_bgpatomicaggregatel,i_bgpaggregatoras,i_bgpagregatorIPh,i_bgpagregatorIPl,i_bgporiginatorIPh,i_bgporiginatorIPl,i_bgpextcom,i_bgplabels);
  commit;
 end if;
END;

create table bgp_vpnv6 (
 srcid int not null references bgp_src(id),
 dtwhen timestamp default CURRENT_TIMESTAMP,
 state number(2) default 0 not null, -- 1 - ipv4u_nlri ; 2 - ipv4u_withdraw
 kind number(4) default 0 not null, -- 1 - ipv4u ; 2 - ipv4m ; 4 - ipv4lu
 netaddrrdh number(10) default 0 not null, -- network address RD high
 netaddrrdl number(10) default 0 not null, -- network address RD low
 netaddrh number(20) default 0 not null, -- network address high part
 netaddrl number(20) default 0 not null, -- network address low part
 netsize number(3) default 0 not null, -- network size
 bgporigin number(1) default 0 not null, -- bgp origin code
 bgpaspath varchar2(256) null, -- bgp aspath
 bgpcommunities varchar2(256) null,
 nexthoprdh number(10) null,
 nexthoprdl number(10) null,
 nexthopiph number(20) null, -- nexthop address high part
 nexthopipl number(20) null, -- nexthop address low part
 bgpmed number(10) null,
 bgplocalpref number(10) null,
 bgpatomicaggregateh number(20) null, -- aa address high part
 bgpatomicaggregatel number(20) null, -- aa address low part
 bgpaggregatoras number(20) null,
 bgpagregatorIPh number(20) null,
 bgpagregatorIPl number(20) null,
 bgporiginatorIPh number(20) null,
 bgporiginatorIPl number(20) null,
 bgpextcom varchar2(256) null,
 bgplabels varchar2(256) null,
 primary key (srcid,kind,netaddrrdh,netaddrrdl,netaddrh,netaddrl,netsize)
);
create index bgp_vpnv6_lbl on bgp_vpnv6(bgplabels);
create index bgp_vpnv6_ec on bgp_vpnv6(bgpextcom);
create index bgp_vpnv6_cm on bgp_vpnv6(bgpcommunities);

create table bgp_vpnv6_log (
 srcid int not null references bgp_src(id),
 dtwhen timestamp default CURRENT_TIMESTAMP,
 state number(2) default 0 not null, -- 1 - ipv4u_nlri ; 2 - ipv4u_withdraw
 kind number(4) default 0 not null, -- 1 - ipv4u ; 2 - ipv4m ; 4 - ipv4lu
 netaddrrdh number(10) default 0 not null, -- network address RD high
 netaddrrdl number(10) default 0 not null, -- network address RD low
 netaddrh number(20) default 0 not null, -- network address high part
 netaddrl number(20) default 0 not null, -- network address low part
 netsize number(3) default 0 not null, -- network size
 bgporigin number(1) default 0 not null, -- bgp origin code
 bgpaspath varchar2(256) null, -- bgp aspath
 bgpcommunities varchar2(256) null,
 nexthoprdh number(10) null,
 nexthoprdl number(10) null,
 nexthopiph number(20) null, -- nexthop address high part
 nexthopipl number(20) null, -- nexthop address low part
 bgpmed number(10) null,
 bgplocalpref number(10) null,
 bgpatomicaggregateh number(20) null, -- aa address high part
 bgpatomicaggregatel number(20) null, -- aa address low part
 bgpaggregatoras number(20) null,
 bgpagregatorIPh number(20) null,
 bgpagregatorIPl number(20) null,
 bgporiginatorIPh number(20) null,
 bgporiginatorIPl number(20) null,
 bgpextcom varchar2(256) null,
 bgplabels varchar2(256) null
);
create index bgp_vpnv6_l_dt on bgp_vpnv6_log(dtwhen);
create index bgp_vpnv6_l_net on bgp_vpnv6_log(kind,netaddrrdh,netaddrrdl,netaddrh,netaddrl,netsize);
create index bgp_vpnv6_l_lbl on bgp_vpnv6_log(bgplabels);
create index bgp_vpnv6_l_ec on bgp_vpnv6_log(bgpextcom);
create index bgp_vpnv6_l_cm on bgp_vpnv6_log(bgpcommunities);

CREATE OR REPLACE PROCEDURE wr_vpnv6(i_srcid IN NUMBER,i_state IN NUMBER,i_kind IN NUMBER,i_netaddrrdh IN NUMBER,i_netaddrrdl IN NUMBER,i_netaddrh IN NUMBER,
 i_netaddrl IN NUMBER,i_netsize IN NUMBER,i_bgporigin  IN NUMBER,i_bgpaspath IN varchar2,i_bgpcommunities IN varchar2,
 i_nhrdh IN NUMBER,i_nhrdl In NUMBER,i_nhh IN NUMBER,i_nhl IN NUMBER,i_bgpmed  IN NUMBER,i_bgplocalpref  IN NUMBER,
 i_bgpatomicaggregateh IN NUMBER,i_bgpatomicaggregatel IN NUMBER,i_bgpaggregatoras  IN NUMBER,i_bgpagregatorIPh  IN NUMBER,i_bgpagregatorIPl  IN NUMBER,
 i_bgporiginatorIPh IN NUMBER,i_bgporiginatorIPl IN NUMBER,i_bgpextcom IN varchar2,i_bgplabels IN varchar2
) IS
 state1 number;
 bgporigin1 number;
 bgpaspath1 varchar2(256);
 bgpcommunities1 varchar2(256);
 nhrdh1 number(10);
 nhrdl1 number(10);
 nhh1 number(20);
 nhl1 number(20);
 bgpmed1 number;
 bgplocalpref1 number;
 bgpatomicaggregateh1 number(20);
 bgpatomicaggregatel1 number(20);
 bgpaggregatoras1 number;
 bgpagregatorIPh1 number(20);
 bgpagregatorIPl1 number(20);
 bgporiginatorIPh1 number(20);
 bgporiginatorIPl1 number(20);
 bgpextcom1 varchar2(256);
 bgplabels1 varchar2(256);
BEGIN
 select count(*) into state1 from bgp_vpnv6 where kind=i_kind and srcid=i_srcid and netaddrrdh=i_netaddrrdh and netaddrrdl=i_netaddrrdl and netaddrh=i_netaddrh and netaddrl=i_netaddrl and netsize=i_netsize;
 if state1>0 then
  select state,bgporigin,bgpaspath,bgpcommunities,nexthoprdh,nexthoprdl,nexthopiph,nexthopipl,bgpmed,bgplocalpref,bgpatomicaggregateh,bgpatomicaggregatel,bgpaggregatoras,bgpagregatorIPh,bgpagregatorIPl,bgporiginatorIPh,bgporiginatorIPl,bgpextcom,bgplabels
  into state1,bgporigin1,bgpaspath1,bgpcommunities1,nhrdh1,nhrdl1,nhh1,nhl1,bgpmed1,bgplocalpref1,bgpatomicaggregateh1,bgpatomicaggregatel1,bgpaggregatoras1,bgpagregatorIPh1,bgpagregatorIPl1,bgporiginatorIPh1,bgporiginatorIPl1,bgpextcom1,bgplabels1
   from bgp_vpnv6 where kind=i_kind and srcid=i_srcid and netaddrrdh=i_netaddrrdh and netaddrrdl=i_netaddrrdl and netaddrh=i_netaddrh and netaddrl=i_netaddrl and netsize=i_netsize;
 end if;
 if state1 is not null and state1=coalesce(i_state,0) and bgporigin1=coalesce(i_bgporigin,0) and coalesce(bgpaspath1,'-')=coalesce(i_bgpaspath,'-') and coalesce(bgpcommunities1,'-')=coalesce(i_bgpcommunities,'-') and 
  coalesce(nhrdh1,0)=coalesce(i_nhrdh,0) and coalesce(nhrdl1,0)=coalesce(i_nhrdl,0) and coalesce(nhh1,0)=coalesce(i_nhh,0) and coalesce(nhl1,0)=coalesce(i_nhl,0) and coalesce(bgpmed1,0)=coalesce(i_bgpmed,0) and coalesce(bgplocalpref1,0)=coalesce(i_bgplocalpref,0) and 
  coalesce(bgpatomicaggregateh1,0)=coalesce(i_bgpatomicaggregateh,0) and coalesce(bgpatomicaggregatel1,0)=coalesce(i_bgpatomicaggregatel,0) and coalesce(bgpaggregatoras1,0)=coalesce(i_bgpaggregatoras,0) and 
  coalesce(bgpagregatorIPh1,0)=coalesce(i_bgpagregatorIPh,0) and coalesce(bgpagregatorIPl1,0)=coalesce(i_bgpagregatorIPl,0) and coalesce(bgporiginatorIPh1,0)=coalesce(i_bgporiginatorIPh,0) and coalesce(bgporiginatorIPl1,0)=coalesce(i_bgporiginatorIPl,0) and 
  coalesce(bgpextcom1,'-')=coalesce(i_bgpextcom,'-') then
   state1:=1;
 else
  delete from bgp_vpnv6 where kind=i_kind and srcid=i_srcid and netaddrrdh=i_netaddrrdh and netaddrrdl=i_netaddrrdl and netaddrh=i_netaddrh and netaddrl=i_netaddrl and netsize=i_netsize;
  insert into bgp_vpnv6 (srcid,state,kind,netaddrrdh,netaddrrdl,netaddrh,netaddrl,netsize,bgporigin,bgpaspath,bgpcommunities,nexthoprdh,nexthoprdl,nexthopiph,nexthopipl,bgpmed,bgplocalpref,bgpatomicaggregateh,bgpatomicaggregatel,bgpaggregatoras,bgporiginatorIPh,bgpagregatorIPl,bgpagregatorIPh,bgporiginatorIPl,bgpextcom,bgplabels) values
   (i_srcid,i_state,i_kind,i_netaddrrdh,i_netaddrrdl,i_netaddrh,i_netaddrl,i_netsize,i_bgporigin,i_bgpaspath,i_bgpcommunities,i_nhrdh,i_nhrdl,i_nhh,i_nhl,i_bgpmed,i_bgplocalpref,i_bgpatomicaggregateh,i_bgpatomicaggregatel,i_bgpaggregatoras,i_bgpagregatorIPh,i_bgpagregatorIPl,i_bgporiginatorIPh,i_bgporiginatorIPl,i_bgpextcom,i_bgplabels);
  insert into bgp_vpnv6_log (srcid,state,kind,netaddrrdh,netaddrrdl,netaddrh,netaddrl,netsize,bgporigin,bgpaspath,bgpcommunities,nexthoprdh,nexthoprdl,nexthopiph,nexthopipl,bgpmed,bgplocalpref,bgpatomicaggregateh,bgpatomicaggregatel,bgpaggregatoras,bgporiginatorIPh,bgpagregatorIPl,bgpagregatorIPh,bgporiginatorIPl,bgpextcom,bgplabels) values
   (i_srcid,i_state,i_kind,i_netaddrrdh,i_netaddrrdl,i_netaddrh,i_netaddrl,i_netsize,i_bgporigin,i_bgpaspath,i_bgpcommunities,i_nhrdh,i_nhrdl,i_nhh,i_nhl,i_bgpmed,i_bgplocalpref,i_bgpatomicaggregateh,i_bgpatomicaggregatel,i_bgpaggregatoras,i_bgpagregatorIPh,i_bgpagregatorIPl,i_bgporiginatorIPh,i_bgporiginatorIPl,i_bgpextcom,i_bgplabels);
  commit;
 end if;
END;

create table bgp_l2vpn (
 srcid int not null references bgp_src(id),
 dtwhen timestamp default CURRENT_TIMESTAMP,
 state number(2) default 0 not null, -- 1 - ipv4u_nlri ; 2 - ipv4u_withdraw
 kind number(4) default 0 not null, -- 1 - ipv4u ; 2 - ipv4m ; 4 - ipv4lu
 netaddrrdh number(10) default 0 not null, -- network address RD high
 netaddrrdl number(10) default 0 not null, -- network address RD low
 site number(10) default 0 not null,
 offset number(10) default 0 not null,
 srange number(10) default 0 not null,
 bgporigin number(1) default 0 not null, -- bgp origin code
 bgpaspath varchar2(256) null, -- bgp aspath
 bgpcommunities varchar2(256) null,
 bgpnexthop number(10) null, -- nexthop address
 bgpmed number(10) null,
 bgplocalpref number(10) null,
 bgpatomicaggregate number(10) null, -- aa address
 bgpaggregatoras number(10) null,
 bgpagregatorIP number(10) null,
 bgporiginatorIP number(10) null,
 bgpextcom varchar2(256) null,
 bgplabels varchar2(256) null,
 primary key (srcid,kind,netaddrrdh,netaddrrdl,site,offset,srange)
);
create index bgp_l2vpn_lbl on bgp_l2vpn(bgplabels);
create index bgp_l2vpn_ec on bgp_l2vpn(bgpextcom);
create index bgp_l2vpn_cm on bgp_l2vpn(bgpcommunities);

create table bgp_l2vpn_log (
 srcid int not null references bgp_src(id),
 dtwhen timestamp default CURRENT_TIMESTAMP,
 state number(2) default 0 not null, -- 1 - ipv4u_nlri ; 2 - ipv4u_withdraw
 kind number(4) default 0 not null, -- 1 - ipv4u ; 2 - ipv4m ; 4 - ipv4lu
 netaddrrdh number(10) default 0 not null, -- network address RD high
 netaddrrdl number(10) default 0 not null, -- network address RD low
 site number(10) default 0 not null,
 offset number(10) default 0 not null,
 srange number(10) default 0 not null,
 bgporigin number(1) default 0 not null, -- bgp origin code
 bgpaspath varchar2(256) null, -- bgp aspath
 bgpcommunities varchar2(256) null,
 bgpnexthop number(10) null, -- nexthop address
 bgpmed number(10) null,
 bgplocalpref number(10) null,
 bgpatomicaggregate number(10) null, -- aa address
 bgpaggregatoras number(10) null,
 bgpagregatorIP number(10) null,
 bgporiginatorIP number(10) null,
 bgpextcom varchar2(256) null,
 bgplabels varchar2(256) null
);
create index bgp_l2vpn_l_dt on bgp_l2vpn_log(dtwhen);
create index bgp_l2vpn_l_net on bgp_l2vpn_log(srcid,kind,netaddrrdh,netaddrrdl,site,offset,srange);
create index bgp_l2vpn_l_lbl on bgp_l2vpn_log(bgplabels);
create index bgp_l2vpn_l_ec on bgp_l2vpn_log(bgpextcom);
create index bgp_l2vpn_l_cm on bgp_l2vpn_log(bgpcommunities);

CREATE OR REPLACE PROCEDURE wr_l2vpn(i_srcid IN NUMBER,i_state IN NUMBER,i_kind IN NUMBER,i_site IN NUMBER,i_offset IN NUMBER,i_range IN NUMBER,i_netsize IN NUMBER,i_bgporigin  IN NUMBER,
 i_bgpaspath IN varchar2,i_bgpcommunities IN varchar2,i_bgpnexthop  IN NUMBER,i_bgpmed  IN NUMBER,i_bgplocalpref  IN NUMBER,
 i_bgpatomicaggregate  IN NUMBER,i_bgpaggregatoras  IN NUMBER,i_bgpagregatorIP  IN NUMBER,i_bgporiginatorIP  IN NUMBER,i_bgpextcom IN varchar2,
 i_bgplabels IN varchar2
) IS
 state1 number;
 bgporigin1 number;
 bgpaspath1 varchar2(256);
 bgpcommunities1 varchar2(256);
 bgpnexthop1 number;
 bgpmed1 number;
 bgplocalpref1 number;
 bgpatomicaggregate1 number;
 bgpaggregatoras1 number;
 bgpagregatorIP1 number;
 bgporiginatorIP1 number;
 bgpextcom1 varchar2(256);
 bgplabels1 varchar2(256);
BEGIN
 select count(*) into state1 from bgp_l2vpn where kind=i_kind and srcid=i_srcid and site=i_site and offset=i_offset and srange=i_range;
 if state1>0 then
  select state,bgporigin,bgpaspath,bgpcommunities,bgpnexthop,bgpmed,bgplocalpref,bgpatomicaggregate,bgpaggregatoras,bgpagregatorIP,bgporiginatorIP,bgpextcom,bgplabels
  into state1,bgporigin1,bgpaspath1,bgpcommunities1,bgpnexthop1,bgpmed1,bgplocalpref1,bgpatomicaggregate1,bgpaggregatoras1,bgpagregatorIP1,bgporiginatorIP1,bgpextcom1,bgplabels1
   from bgp_l2vpn where kind=i_kind and srcid=i_srcid and site=i_site and offset=i_offset and srange=i_range;
 end if;
 if state1 is not null and state1=coalesce(i_state,0) and bgporigin1=coalesce(i_bgporigin,0) and coalesce(bgpaspath1,'-')=coalesce(i_bgpaspath,'-') and coalesce(bgpcommunities1,'-')=coalesce(i_bgpcommunities,'-') and coalesce(bgpnexthop1,0)=coalesce(i_bgpnexthop,0) and coalesce(bgpmed1,0)=coalesce(i_bgpmed,0) and coalesce(bgplocalpref1,0)=coalesce(i_bgplocalpref,0) and coalesce(bgpatomicaggregate1,0)=coalesce(i_bgpatomicaggregate,0) and coalesce(bgpaggregatoras1,0)=coalesce(i_bgpaggregatoras,0) and coalesce(bgpagregatorIP1,0)=coalesce(i_bgpagregatorIP,0) and coalesce(bgporiginatorIP1,0)=coalesce(i_bgporiginatorIP,0) and coalesce(bgpextcom1,'-')=coalesce(i_bgpextcom,'-') then
   state1:=1;
 else
  delete from bgp_l2vpn where kind=i_kind and srcid=i_srcid and site=i_site and offset=i_offset and srange=i_range;
  insert into bgp_l2vpn (srcid,state,kind,site,offset,srange,bgporigin,bgpaspath,bgpcommunities,bgpnexthop,bgpmed,bgplocalpref,bgpatomicaggregate,bgpaggregatoras,bgpagregatorIP,bgporiginatorIP,bgpextcom,bgplabels) values (i_srcid,i_state,i_kind,i_site,i_offset,i_range,i_bgporigin,i_bgpaspath,i_bgpcommunities,i_bgpnexthop,i_bgpmed,i_bgplocalpref,i_bgpatomicaggregate,i_bgpaggregatoras,i_bgpagregatorIP,i_bgporiginatorIP,i_bgpextcom,i_bgplabels);
  insert into bgp_l2vpn_log (srcid,state,kind,site,offset,srange,bgporigin,bgpaspath,bgpcommunities,bgpnexthop,bgpmed,bgplocalpref,bgpatomicaggregate,bgpaggregatoras,bgpagregatorIP,bgporiginatorIP,bgpextcom,bgplabels) values (i_srcid,i_state,i_kind,i_site,i_offset,i_range,i_bgporigin,i_bgpaspath,i_bgpcommunities,i_bgpnexthop,i_bgpmed,i_bgplocalpref,i_bgpatomicaggregate,i_bgpaggregatoras,i_bgpagregatorIP,i_bgporiginatorIP,i_bgpextcom,i_bgplabels);
  commit;
 end if;
END;
